#ifdef __K1__

#define CONFIGURE_MAXIMUM_POSIX_THREADS 6
//#define CONFIGURE_AMP_MAIN_STACK_SIZE 5500
#define CONFIGURE_DEFAULT_TASK_STACK_SIZE 8000 //8000
#define CONFIGURE_MINIMUM_TASK_STACK_SIZE 8000

#include <mppaipc.h>
#include <mppa/osconfig.h>
#endif

#include "version.h"
#include <getopt.h>
#include <stdio.h>
#include <iostream>
#include <unistd.h>

#include <component.h>
#include <scheduler.h>
#include <compound.h>
#include "clock.h"
#include "linux-clock.h"
#include "hidden.h"

#define RUNMODE_MULTI       1
#define RUNMODE_REALTIME    2

#ifndef ENGINE_VERSION
#define ENGINE_VERSION "UNSPECIFIED"
#endif

void printLogo() {
    std::cout << "BIP multi-threaded engine. Version: " << ENGINE_VERSION << std::endl;
}


void help(struct option *l){
  int i=0;
  printf("Help for BIP engine\n");
  while(l[i].name != 0){
    printf("  --%s", l[i].name);
    if (l[i].flag == NULL) {
      printf(" / -%c", static_cast<char>(l[i].val));
    }
    printf("\n");
    i++;
  }
}

Engine& parseCmdLine(int argc, char **argv) {
  Engine& result = *new Engine();
  static struct option long_options[] =
    {
      /* These options set a flag. */
      {(char*) "multi",      no_argument,        &result.mRunMode,          RUNMODE_MULTI},
      {(char*) "realtime",   no_argument,        &result.mRunMode,          RUNMODE_REALTIME},
      {(char*) "lazy",       no_argument,        &result.mIntLazy,    1},
      {(char*) "quiet",      no_argument,        &result.mIntQuiet,   1},
      {(char*) "help",       no_argument,        0,       'h'},
     

      /* These options don't set a flag.
          We distinguish them by their indices. */

      {(char*) "seed",         required_argument,   0, 'r'},
      {(char*) "limittime",    required_argument,   0, 'l'},
      {(char*) "maxdrift",     required_argument,   0, 'd'},
      {0,                      0,                   0,  0}
    };

  result.mIntLazy   = 0;
  result.mRunMode   = 0;
  result.mIntQuiet  = 0;
  result.mLimitTime = -1;
  result.mMaxDrift  = -1;
  result.mRandSeed  = 1;

  int option_index = 0;
  for (;;){
    int c;
    c = getopt_long (argc, argv, "s:",
                     long_options, &option_index);
    if (c == -1)
      break;
    
    switch(c){
    case 'r':
      result.mRandSeed = atoi(optarg);
      break;
    case 'l':
      result.mLimitTime = atoi(optarg);
      break;
    case 'd':
      result.mMaxDrift = atoi(optarg);
      break;
    case 'h':
      help(long_options);
      exit(0);
      break;
    case '?':
    case ':':
      printf("Unknown: '%c'\n", c);
      help(long_options);
      exit(-1);
    }
  }

  if (result.mIntQuiet > 0) { result.mQuiet = true ; printf("Running \"quiet\" mode\n"); }
  else { result.mQuiet = false; }
  if (result.mIntLazy  > 0) { result.mLazy  = true;  printf("Running \"lazy\" mode\n"); } 
  else { result.mLazy = false; }
  if (result.mRunMode == RUNMODE_REALTIME) { printf("Running \"realtime\" mode\n"); }

  return result;
}

void markPorts(Compound* system);

int main(int argc, char** argv) {

  // moved below: printLogo();
  
  Engine& engine = parseCmdLine(argc, argv);
  LogicalClock*     logicalClock;
  SimuClock         simuClock("simu_clock");
  LinuxClock        linuxClock("linux_clock");
  /* instanciates engine's clocks */
  if (engine.mRunMode == RUNMODE_REALTIME) {
    
    logicalClock = new LogicalClock("logical_clock", linuxClock);
    engine.mSystem = Compound::Create(argc, argv, logicalClock); /* real-time clock for linux */
  }
  else {
    logicalClock = new LogicalClock("logical_clock", simuClock);
    engine.mSystem = Compound::Create(argc, argv, logicalClock); /* simulation clock (transitions are instantaneous) */
  }

  //*****************************
  // not in this example markPorts(engine.mSystem);
  //*****************************

  Scheduler scheduler(engine.mSystem, (engine.mRunMode== RUNMODE_REALTIME), logicalClock, engine);
  
  
  scheduler.setup();
  linuxClock.setup(&scheduler);

  scheduler.createThreads();
  scheduler.createDefaultThreads();

  //if (!engine.mQuiet) {
    printLogo();
    std::cout <<"No. of atoms in system = " << scheduler.nAtoms() << std::endl;
    std::cout <<"No. of threads in system = " << scheduler.nThreads() << std::endl;

    if (engine.mMaxDrift>0) {
        std::cout << "Max drift = " << engine.mMaxDrift << std::endl;
    }
    if (engine.mLimitTime>0) {
        std::cout << "Limit time = " << engine.mLimitTime << std::endl;
    }
    std::cout << "====> LAUNCH: <====" << std::endl;
    std::cout << std::endl << std::flush;
    usleep(1000); // let us first print this
    //scheduler.dumpThreads();
  //}
  scheduler.launchThreads();
 
  //linuxClock.restart(); moved to run()
  scheduler.run();
  
  //delete system; // should ensure thread completion first, which we do not do yet
#ifdef __K1__
  mppa_exit(0);
#endif
  return 0;
}

/********************************************************************************************/
/********************************************************************************************/
/********************************************************************************************/

#define markMBconnectorsMaster(arg_system, arg_prod,arg_prodOut, arg_cons, arg_consIn, arg_mb) \
{\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_" arg_prodOut "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons "_ins.DOLC_read_" arg_consIn "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons "_ins.DOLC_read_return");\
}

#define markMBconnectorsSlave(arg_system, arg_mb) \
{\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.Write");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.ReadAck");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.WriteAck");\
}

#define markBBconnectorsMaster(arg_system, arg_prod,arg_prodOut, arg_cons, arg_consIn, arg_bb) \
{\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_" arg_prodOut "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.FinishBB1");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.Swap");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons "_ins.DOLC_read_" arg_consIn "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.CopyAck");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons "_ins.DOLC_read_return");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.LiveStep");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.DeadStep");\
}

#define markBBconnectorsSlave(arg_system, arg_bb) \
{\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.Write");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishPassive");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishSwap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader.Swap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.CopyAck");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.WriteAck");\
    Hidden::markSlave(arg_system, "BIP_Top." arg_bb "_ins.bb_reader.CopyAck");\
}

#define markBBconnectorsMaster2(arg_system, arg_prod,arg_prodOut, arg_cons1, arg_consIn1, arg_cons2, arg_consIn2, arg_bb) \
{\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_" arg_prodOut "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.FinishBB1");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.Swap");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons1 "_ins.DOLC_read_" arg_consIn1 "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons2 "_ins.DOLC_read_" arg_consIn2 "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.CopyAck");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons1 "_ins.DOLC_read_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons2 "_ins.DOLC_read_return");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.LiveStep");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.DeadStep");\
}

#define markBBconnectorsSlave2(arg_system, arg_bb) \
{\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.Write");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishPassive");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishSwap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader1.Swap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader2.Swap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader1.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader2.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.CopyAck");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.WriteAck");\
    Hidden::markSlave(arg_system, "BIP_Top." arg_bb "_ins.bb_reader1.CopyAck");\
    Hidden::markSlave(arg_system, "BIP_Top." arg_bb "_ins.bb_reader2.CopyAck");\
}

// This code was used in FMS application , gives example of how to hide connectors
void markPorts(Compound* system) {
    markMBconnectorsMaster (system, "sens_a2",   "p2_out",   "dp1",      "dp1_in",  "GPS_channel");
    markMBconnectorsMaster (system, "dp1",       "dp1_out",  "sens_c1",  "p52_in",  "GPS_channel_delay");
    markMBconnectorsMaster (system, "sens_c1",   "p5_out",   "dp2",      "dp2_in",  "Sensors_channel"); 
    markBBconnectorsMaster (system, "dp2",       "dp2_out",  "loc_c1",   "p91_in",  "Sensors_channel_delay");
    markBBconnectorsMaster (system, "loc_c1",    "p9_out",   "loc_c2",   "p10_in",  "HBCP_channel");
    markBBconnectorsMaster2(system, "loc_c2",    "p10_out",  "loc_c3",   "p112_in", "loc_c4", "p122_in", "LBCP_channel");
    markBBconnectorsMaster (system, "loc_c3",    "p11_out",  "loc_c1",   "p93_in",  "MagDec_channel");

    markMBconnectorsSlave (system, "GPS_channel");
    markMBconnectorsSlave (system, "GPS_channel_delay");
    markMBconnectorsSlave (system, "Sensors_channel"); 
    markBBconnectorsSlave (system, "Sensors_channel_delay");
    markBBconnectorsSlave (system, "HBCP_channel");
    markBBconnectorsSlave2(system, "LBCP_channel");
    markBBconnectorsSlave (system, "MagDec_channel");
}

