#ifdef __K1__
#define CONFIGURE_MAXIMUM_POSIX_THREADS 5
//#define CONFIGURE_AMP_MAIN_STACK_SIZE 5500
#define CONFIGURE_DEFAULT_TASK_STACK_SIZE 8000
#define CONFIGURE_MINIMUM_TASK_STACK_SIZE 8000
#include <mppaipc.h>
#include <mppa/osconfig.h>
#endif



#include "version.h"
#include <getopt.h>
#include <stdio.h>
#include <iostream>
#include <unistd.h>

#include <component.h>
#include <scheduler.h>
#include <compound.h>
#include <cstdlib>

#include "clock.h"
#include "linux-clock.h"
#include "hidden.h"

#define RUNMODE_MULTI       1
#define RUNMODE_REALTIME    2

#ifndef ENGINE_VERSION
#define ENGINE_VERSION "UNSPECIFIED"
#endif

#ifdef __sparc__
#include "my_rtems_settings.h"
#endif

void printLogo() {
    std::cout << "BIP multi-threaded engine. Version: " << ENGINE_VERSION << std::endl;
}


void help(struct option *l){
  int i=0;
  printf("Help for BIP engine\n");
  while(l[i].name != 0){
    printf("  --%s", l[i].name);
    if (l[i].flag == NULL) {
      printf(" / -%c", static_cast<char>(l[i].val));
    }
    printf("\n");
    i++;
  }
}

Engine& parseCmdLine(int argc, char **argv) {
  Engine& result = *new Engine();
  static struct option long_options[] =
    {
      /* These options set a flag. */
      {(char*) "multi",      no_argument,        &result.mRunMode,          RUNMODE_MULTI},
      {(char*) "realtime",   no_argument,        &result.mRunMode,          RUNMODE_REALTIME},
      {(char*) "lazy",       no_argument,        &result.mIntLazy,    1},
      {(char*) "quiet",      no_argument,        &result.mIntQuiet,   1},
      {(char*) "help",       no_argument,        0,       'h'},
     

      /* These options don't set a flag.
          We distinguish them by their indices. */

      {(char*) "seed",         required_argument,   0, 'r'},
      {(char*) "limittime",    required_argument,   0, 'l'},
      {(char*) "maxdrift",     required_argument,   0, 'd'},
      {0,                      0,                   0,  0}
    };

  result.mIntLazy   = 0;
  result.mRunMode   = 0;
  result.mIntQuiet  = 0;
  result.mLimitTime = -1;
  result.mMaxDrift  = -1;
  result.mRandSeed  = 1;

  int option_index = 0;
  for (;;){
    int c;
    c = getopt_long (argc, argv, "s:",
                     long_options, &option_index);
    if (c == -1)
      break;
    
    switch(c){
    case 'r':
      result.mRandSeed = atoi(optarg);
      break;
    case 'l':
      result.mLimitTime = atoi(optarg);
      break;
    case 'd':
      result.mMaxDrift = atoi(optarg);
      break;
    case 'h':
      help(long_options);
      exit(0);
      break;
    case '?':
    case ':':
      printf("Unknown: '%c'\n", c);
      help(long_options);
      exit(-1);
    }
  }

  if (result.mIntQuiet > 0) { result.mQuiet = true ; printf("Running \"quiet\" mode\n"); }
  else { result.mQuiet = false; }
  if (result.mIntLazy  > 0) { result.mLazy  = true;  printf("Running \"lazy\" mode\n"); } 
  else { result.mLazy = false; }
  if (result.mRunMode == RUNMODE_REALTIME) { printf("Running \"realtime\" mode\n"); }

  return result;
}

void markPorts(Compound* system);

void* POSIX_Init(void* argument) {

  int argc = 5;
  char* argv[] = {"binary", "--realtime", "--quiet", "--maxdrift", "10000",  NULL};
  //int argc = 4;
  //char* argv[] = {"binary", "--realtime", "--maxdrift", "100000",  NULL};
  // moved below: printLogo();
  //int argc = 2;
  //char* argv[] = {"binary", "--realtime",   0};
  
  Engine& engine = parseCmdLine(argc, argv);
  LogicalClock*     logicalClock;
  SimuClock         simuClock("simu_clock");
  LinuxClock        linuxClock("linux_clock");
  /* instanciates engine's clocks */
  if (engine.mRunMode == RUNMODE_REALTIME) {
    
    logicalClock = new LogicalClock("logical_clock", linuxClock);
    engine.mSystem = Compound::Create(argc, argv, logicalClock); /* real-time clock for linux */
  }
  else {
    logicalClock = new LogicalClock("logical_clock", simuClock);
    engine.mSystem = Compound::Create(argc, argv, logicalClock); /* simulation clock (transitions are instantaneous) */
  }

  //*****************************
  //printf("DEBUG> Not marking the ports...\n");
  markPorts(engine.mSystem);
  //*****************************

  Scheduler scheduler(engine.mSystem, (engine.mRunMode== RUNMODE_REALTIME), logicalClock, engine);
  
  
  scheduler.setup();
  linuxClock.setup(&scheduler);

  scheduler.createThreads();
  scheduler.createDefaultThreads();

  //if (!engine.mQuiet) {
    printLogo();
    std::cout <<"No. of atoms in system = " << scheduler.nAtoms() << std::endl;
    std::cout <<"No. of threads in system = " << scheduler.nThreads() << std::endl;

    if (engine.mMaxDrift>0) {
        std::cout << "Max drift = " << engine.mMaxDrift << std::endl;
    }
    if (engine.mLimitTime>0) {
        std::cout << "Limit time = " << engine.mLimitTime << std::endl;
    }
    std::cout << "====> LAUNCH: <====" << std::endl;
    std::cout << std::endl << std::flush;
    usleep(1000); // let us first print this
    //scheduler.dumpThreads();
  //}
  scheduler.launchThreads();
 
  //linuxClock.restart(); moved to run()
  scheduler.run();
  
  //delete system; // should ensure thread completion first, which we do not do yet
#ifdef __K1__
  mppa_exit(0);
#endif
  return 0;
}

/********************************************************************************************/
/********************************************************************************************/
/********************************************************************************************/

#define markMBconnectorsMaster(arg_system, arg_prod,arg_prodOut, arg_cons, arg_consIn, arg_mb) \
{\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_" arg_prodOut "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons "_ins.DOLC_read_" arg_consIn "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons "_ins.DOLC_read_return");\
}

#define markMBconnectorsSlave(arg_system, arg_mb) \
{\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.Write");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.ReadAck");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_mb "_ins.WriteAck");\
}

#define markBBconnectorsMaster(arg_system,  arg_prod,  arg_prodOut, arg_BB_index,arg_cons, arg_consIn, arg_bb) \
{\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_" arg_prodOut "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.FinishBB" arg_BB_index);\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.Swap");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons "_ins.DOLC_read_" arg_consIn "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.CopyAck");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons "_ins.DOLC_read_return");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.LiveStep");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.DeadStep");\
}

#define markBBconnectorsSlave(arg_system, arg_bb) \
{\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.Write");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishPassive");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishSwap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader.Swap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.CopyAck");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.WriteAck");\
    Hidden::markSlave(arg_system, "BIP_Top." arg_bb "_ins.bb_reader.CopyAck");\
}

#define markBBconnectorsMaster2(arg_system, arg_prod,  arg_prodOut, arg_BB_index,arg_cons1, arg_consIn1, arg_cons2, arg_consIn2, arg_bb) \
{\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_" arg_prodOut "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.FinishBB" arg_BB_index);\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.Swap");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons1 "_ins.DOLC_read_" arg_consIn1 "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_cons2 "_ins.DOLC_read_" arg_consIn2 "_call");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.CopyAck");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_prod "_ins.DOLC_write_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons1 "_ins.DOLC_read_return");\
    Hidden::markMaster(arg_system,  "BIP_Top." arg_cons2 "_ins.DOLC_read_return");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.LiveStep");\
    Hidden::markMaster(arg_system, "BIP_Top." arg_bb "_ins.bb_writer.DeadStep");\
}

#define markBBconnectorsSlave2(arg_system, arg_bb) \
{\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.Write");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishPassive");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.FinishSwap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader1.Swap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader2.Swap");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader1.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_reader2.Read");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.CopyAck");\
    Hidden::markSlave(arg_system,  "BIP_Top." arg_bb "_ins.bb_writer.WriteAck");\
    Hidden::markSlave(arg_system, "BIP_Top." arg_bb "_ins.bb_reader1.CopyAck");\
    Hidden::markSlave(arg_system, "BIP_Top." arg_bb "_ins.bb_reader2.CopyAck");\
}

void markPorts(Compound*) {}

/** parallelproc
void markPorts(Compound* system) {
    markMBconnectorsMaster (system, "split",   "p_PROC_XIF_Write2",       "processa",      "p_PROC_XIF_Read2",  "processa");
    markMBconnectorsMaster (system, "split",   "p_PROC_XIF_Write",        "processb",      "p_PROC_XIF_Read",   "processb");
    markMBconnectorsSlave (system, "xmailbox");
    markMBconnectorsSlave (system, "xmailbox2");

}*/

/** GNC
void markPorts(Compound* system) {
    markMBconnectorsMaster (system, "data_input_dispatcher",   "p_PROC_Navigation_Input_Data_Buffer_Get",       "guidance_navigation_task",      "p_PROC_Navigation_Output_Data_Buffer_Get",  "navigation_input_data_get_mb");
    markBBconnectorsMaster (system, "data_input_dispatcher",   "p_PROC_Control_Input_Data_Buffer_Get",     "1", "control_fm_task",               "p_PROC_Control_Output_Data_Buffer_Get",     "control_input_data_get_bb");

    markBBconnectorsMaster (system, "control_fm_task",         "p_PROC_Control_Input_Data_Buffer_Set_Output",    "1", "control_output_task",       "p_PROC_Control_Output_Data_Buffer_Get_Output",  "control_input_data_buffer_get_output_bb");
    markBBconnectorsMaster (system, "control_fm_task",         "p_PROC_Navigation_Input_Data_Buffer_Set_Control","2", "guidance_navigation_task",  "p_PROC_Navigation_Output_Data_Buffer_Set_Control",  "navigation_set_control_bb");

    markBBconnectorsMaster (system, "guidance_navigation_task","p_PROC_Set_Navigation_In", "1", "control_fm_task",  "p_PROC_Set_Navigation_Out",  "set_navigation_bb");

    markMBconnectorsSlave (system, "navigation_input_data_get_mb");
    markBBconnectorsSlave (system, "control_input_data_get_bb");
    markBBconnectorsSlave (system, "control_input_data_buffer_get_output_bb"); 
    markBBconnectorsSlave (system, "navigation_set_control_bb");
    markBBconnectorsSlave (system, "set_navigation_bb");
}

*/
