#ifndef ATOM_H_
#define ATOM_H_

#include "component.h"
#include <vector>
#include <set>
#include <assert.h>
#include <pthread.h>
#include <stdio.h>
#include <bip-time.h>
class CtrlState;
class Port;
class Connector;
class Priority;
class Scheduler;
class Clock;
class LogicalClock;
class TimeConstraint;
class Compound;
class SThread;
/***********************************
 * Class: Atom
 * Description: Atomic component
 ***********************************/
class Atom : public Component {

  /*** Public data members ***/
  public:
    enum Status {BUSY = 0, READY = 1};

    typedef std::vector<Connector*> connectors_type;
    typedef std::vector<Priority*> priorities_type;
    typedef std::vector<Clock*>  clocks_type;
    typedef std::vector<TimeConstraint*>  constraints_type;
	typedef std::vector<Port*>  ports_type;

    connectors_type mLinks;      /* Atom links (connectors) */
    priorities_type mConditions;  /* atom conditions */
    clocks_type mClocks;
    constraints_type mConstraints;
    ports_type     mPorts;       /* Atom ports */
    ports_type mNextPorts;

    /* Almost-lazy mode: only indicate whether we have next (exported and hence instanteneous) ports
     * or not (and hence only internal-continuous ports or a deadlock) */
    bool mHaveNextPortsFlag; /* flag set  by the generated component */
    bool mCachedNextPortsFlag; /* cached flag used by the engine */

    Port*      mElectedSpecialPort;
    Port*      mLastExecutedPort;
    ports_type  mNotifiedPorts;
    ports_type  mPreviouslyNotifiedPorts;
    CtrlState*             BIP_STATE;   /* current control-state */
    Scheduler*             mScheduler;  /* back pointer to scheduler */
    Status                 mStatus;     /* status wrt schduler */
    bool                   mSelected;   /* selected for execution */

  protected:
    SThread*                mThreadHolder;
    SThread*                mExecutorThread;

  /*** Constructoris & Dectructors ***/
  public:
    Atom(const char*, const char*, Compound*,  const LogicalClock*);
    virtual ~Atom();

  /*** Public Data-Access Methods ***/
  public:
    connectors_type& links()
    {
      return mLinks;
    }
    ports_type& nextPorts()
       {
         return mNextPorts;
       }
    ports_type& ports()
    {
    	return mPorts;
    }
    ports_type& notifiedPorts()
    {
    	return mNotifiedPorts;
    }
    Port* electedSpecialPort() {
        return mElectedSpecialPort;
    }
    void setElectedSpecialPort(Port* port);
    inline Status status() const { return mStatus; }
    inline void setStatus(const Status status) { mStatus = status; }

    inline SThread* threadHolder()  { return mThreadHolder; }
    void   setThreadHolder(SThread* thread) { mThreadHolder = thread; }

    inline SThread* executorThread()  { return mExecutorThread; }
    void   setExecutorThread(SThread* thread) { mExecutorThread = thread; }

    inline bool selected() const    { return mSelected; }
    inline void setSelected(bool b) { mSelected = b; }

    inline Port* lastExecutedPort() const { return mLastExecutedPort; }

    void postProcessNotifiedPorts();


  /*** Public Utility Methods ***/
  public:
    virtual bool       isAtom() const     { return true; }
    virtual bool       isCompound() const { return false; }
            void       resetLink();      /* mark links for re-evaluation */
            void       addPort(Port *p); /*      */


  /*** Public Methods visible from appln. threads or scheduler ***/
  public:
    virtual void getBusyAtoms(Scheduler*);


    virtual void setup(Scheduler*);
    virtual void createDefaultThreads();

    virtual void goBusy();
    virtual void goReady();

            void notifyAllPorts(); // calls appln's "notifyPorts()" for the atom and for slaves of just executed master (if any)
            void notify(const int, ...);  // called from notifyPorts()

           void executeAllActions(Port* ch); // calls appln's "execute" for "ch"  and his slaves (if any)

protected:
    virtual void notifyPorts() = 0;     // application code
    virtual void execute(Port* ch) = 0; // application code

public:
            void updateNextApp(Port*);
            void initializeLUU(BIP::Time& );
            bool inNextPortList(Port*);
    virtual void dump(FILE* file) const;


    // from application
    void setNextPortsFlag(bool value) { mHaveNextPortsFlag = value; } // by component (almost lazy mode)

    // from scheduler
    void putNextPortsFlagInCache() { mCachedNextPortsFlag = mHaveNextPortsFlag; } // by the engine (almost lazy mode)
    bool getCachedNextPortsFlag()  { return mCachedNextPortsFlag; }  // by the engine (almost lazy mode)

    void resetLinks_notifiedPorts();
    void setLinksFalse_PreviousPorts();
    void updatePreviousPorts();
};

class CtrlState {
  /*** Private data members ***/
  private:
    char*   mName;
  /*** Constructoris & Dectructors ***/
  public:
    CtrlState(const char* = NULL);
    ~CtrlState();

    inline char* name() { return mName; }
};

#endif /* ATOM_H_*/
