/*******************************************************************************
 *                            C++ Verimag
 *    
 *
 *------------------------------------------------------------------------------
 * 	            Distributed and Complex Systems - Gires
 *------------------------------------------------------------------------------
 * bip-time.h
 *------------------------------------------------------------------------------
 *  Time for BIP2 Engine implementation
 *  + class Time        : time management
 *------------------------------------------------------------------------------
 * Created by Jacques COMBAZ on Wed Jan 13 2010
 *
 ******************************************************************************/

#ifndef BIP_TIME_H
#define BIP_TIME_H

#include <iostream>
#include <sstream>
#include <cassert>

namespace BIP { 
  /* class for computing time values, expressed in nanosecond by default */
  class Time {
    /*** Protected data members ***/
    protected:
      long long mTime;

    /*** public Constructors & Dectructors ***/
    public:
      inline Time()                              { assert(MAX>1000000L && MIN<-1000000L);}
      inline Time(const Time&      t)            { mTime = t.mTime; assert(MAX>1000000L && MIN<-1000000L);}

      inline Time(const long long& t, Time unit) { mTime =             t  * unit.mTime; }
      inline Time(      long       t, Time unit) { mTime = (long long)(t) * unit.mTime; }
      inline Time(      int        t, Time unit) { mTime = (long long)(t) * unit.mTime; }
      inline Time(const double&    t, Time unit) { mTime = (long long)(t) * unit.mTime; }

    /*** protected Constructors & Dectructors ***/
    protected:
      inline Time(const long long& t)            { mTime = t; }
      inline Time(      int        t)            { mTime = (long long)(t); }
      inline Time(      long       t)            { mTime = (long long)(t); }
      inline Time(const double&    t)            { mTime = (long long)(t); }


    /*** Public data members ***/
    public:
      inline Time& operator= (const Time& t) { if (this != &t) { mTime = t.mTime; } return *this; }

      inline const Time operator-() const {
        Time r;
             if (*this == MAX) r = MIN;
        else if (*this == MIN) r = MAX;
        else   r.mTime = -mTime;
        return r;
      }
      inline Time& operator+=(const Time& t) {
        if (*this != MAX && *this != MIN) {
          if (t == MAX || t == MIN) *this = t;
          else mTime += t.mTime;
        }
        return *this;
      }
      inline Time& operator-=(const Time& t) {
        if (*this != MAX && *this != MIN) {
               if (t == MAX) *this = MIN;
          else if (t == MIN) *this = MAX;
          else mTime -= t.mTime;
        }
        return *this;
      }
      inline const Time operator+(const Time& t)      const { Time r = *this; r += t; return r; }
      inline const Time operator-(const Time& t)      const { Time r = *this; r -= t; return r; }

      inline const long long operator/(const Time& t) const { long long r = this->mTime / t.mTime; return r; }

      inline bool  operator< (const Time& t) const { return mTime <  t.mTime; }
      inline bool  operator<=(const Time& t) const { return mTime <= t.mTime; }
      inline bool  operator> (const Time& t) const { return t <  *this; }
      inline bool  operator>=(const Time& t) const { return t <= *this; }
      inline bool  operator==(const Time& t) const { return mTime == t.mTime; }
      inline bool  operator!=(const Time& t) const { return !(*this == t); }

      friend std::ostream&  operator<<(std::ostream &, const Time&);
      friend std::istream&  operator>>(std::istream &,       Time&);

      inline void scaleBy(long long PPHack_slow_factor) { mTime = mTime*PPHack_slow_factor; }
      inline void descaleBy(long long PPHack_slow_factor) { mTime = mTime/PPHack_slow_factor; }

    /*** Predefined Constant Values ***/
    public:
      static const Time ZERO;
      static const Time GRANULARITY;
      static const Time MIN;
      static const Time MAX;

      static const Time MICROSECOND;
      static const Time MILLISECOND;
      static const Time SECOND;

      static const Time DEFAULT;
  };

  inline Time min(Time& t, Time& u) {
    if (t <= u) return t;
    else        return u;
  }
  inline Time max(Time& t, Time& u) {
    if (t >= u) return t;
    else        return u;
  }
}

#endif

