#ifndef CONNECTOR_H
#define CONNECTOR_H

#include "structure.h"
#include "bip-time.h"
#include "clock.h"
#include "constraint.h"
#include "priority.h"
#include "scheduler.h"
#include <vector>
#include <unordered_map>
#include <cstdio>
#include <cstdarg>
#include <cassert>

class Connector;
class Compound;
class Port;
class Atom;

namespace std {
template <> 
struct hash<BitVector>
{
    size_t operator()(const BitVector &val) const;
};

inline size_t hash<BitVector>::operator()(const BitVector &val) const {
  size_t res(0);

//   std::cerr << "start: " << res << std::endl;
  for(BitVector::const_iterator it = val.begin(); it != val.end(); ++it) {
    res <<= 1;
    res += *it;
//     std::cerr << res << std::endl;
  }
//   std::cerr << "end: " << res << std::endl;

  return res;
}

}

/*********************************
 * Class: Interaction
 * Description: Interaction class
 *********************************/
class Interaction {
  /*** Protected data members ***/
  public:
    enum Status {UNKNOWN = 0, PENDING = 1, INHIBITED = 2, MAXIMAL = 3};

    typedef std::vector<Priority*> priorities_type;
    typedef std::vector<Interaction*> interactions_type;

  /*** Protected data members ***/
  protected:
    Connector* mConnector;  /* the parent Connector */
    BitVector mSynced;     /* the synced ports     */
    priorities_type mInhibitors; /* potential inhibitors */
    interactions_type mSuper;   /* super interactions   */
    Status                mStatus;
    Intervals    mConstraint;          /* time constraint  */
   
  /*** Constructors & Dectructors ***/
  public:
    Interaction(Connector*, const BitVector&);
    Interaction(const Interaction&);
    virtual ~Interaction();

  /*** Public Data-Access Methods ***/
  public:
    Connector* connector() const
    {
      return mConnector;
    }

    const BitVector& synced() const
    {
      return mSynced;
    }

    priorities_type& inhibitors()
    {
      return mInhibitors;
    }

    interactions_type& super()
    {
      return mSuper;
    }

    Status status() const               { return mStatus; }
    void setStatus(const Status s)      { mStatus = s; }
    
    //by ahlem
    inline        Intervals& constraint()                               { return mConstraint; }
    inline  const Intervals& constraint()                         const { return mConstraint; }
  /*** Public Utility Methods ***/
  public:
    void   setup();
    
    bool   equal(const Interaction*) const;
    bool   isSubset(const Interaction*, bool&) const;
    void   execute();
    Status evaluate();
    void   evaluateTC(Interaction*);
    static Interaction* Create(Connector*, const BitVector&);
    static Interaction* Create(Connector*, int size, ...);
    static Interaction* CreateRealTime(Connector*, const BitVector&);
    void print(FILE* file) const;
    void dump(FILE* file) const;
    BIP::Time     deadline(const BIP::Time& time) const { return constraint().deadline(time); }
    BIP::Time     next(const BIP::Time& time)     const { return constraint().next(time); }
    Interaction*  copy() const;
    void          dominate(Interaction&, const BIP::Time& = BIP::Time::ZERO);
    void          dominate(Interaction&, const Intervals&, const BIP::Time& = BIP::Time::ZERO);

    friend class Scheduler;
};

/*************************************
 * Connector info for hidden interactions
 *************************************/
namespace Hidden {
    struct ConnInfo {
        struct Impl;

        Impl* impl;
        ConnInfo() { impl = NULL; }
        ~ConnInfo() { if (impl) delete impl;}
        
        struct Impl {
            Port* master;
            std::set<Port*>* allSlaves;
            std::set<Port*>* syncedSlaves;
            
            Impl()  { master = NULL; allSlaves = NULL; syncedSlaves = NULL; }
            ~Impl() { 
                if (allSlaves)    delete allSlaves; 
                if (syncedSlaves) delete syncedSlaves;
            }
            void createRootInfo() { 
                 assert(!allSlaves && !syncedSlaves);
                   allSlaves = new std::set<Port*>;
                syncedSlaves = new std::set<Port*>;                    
            }
        };
    };
}

/*******************************
 * Class: Connector
 * Description: Connector class
 *******************************/
class Connector {
  /*** Public data members ***/
  public:
    enum Status {UNSET = 0, UNKNOWN = 1, ENABLED = 2, DISABLED = 3};

    typedef std::vector<Port*> ports_type;
    typedef std::vector<Connector*> connectors_type;
    typedef std::vector<Interaction*> interactions_type;
    typedef std::vector<Priority*> priorities_type;
    typedef std::unordered_map<BitVector, Interaction> interactionsHashMap_type;

  /*** Private data members ***/
  private:
    static char LABEL[1024];
    char              mInfo[512];   /* various infos buffer                */
    
  protected: 
    BIP::Time  mLowerApp  ;   /*approximated lower bound of time constraint*/
    BIP::Time  mUpperApp ;   /*approximated Upper bound of time constraint*/
    Urgency    mUrgencyApp;   /*approximated Urgency of time constraint*/
    BIP::Time  mDeadlineApp ;
    bool       mIsConstantApp;
    bool mApproximated;
    int mAll;
  /*** Protected data members ***/
  protected:
    char*              mName;          /* name of connector */
    Compound*          mHolder;        /* holder compound */
    ports_type mPorts;         /* port array */
    Port*              mxPort;         /* exported port */
    bool               first; 
    bool               mLegal;         /* legal i.e. complete or maximal */
    bool               mDominated;     /* dominated by priorities */
    Interaction*       mFeasibleInter; /* feasible interaction */
    bool               mSubconn;       /* used as a sub-connector
                                          in hierarcical connector */
    bool               mEval;          /* flag marking evaluated */
    connectors_type    mLinks;    /* back pointer to higher links */
    Status             mStatus;        /* status wrt the scheduler */
    int                mAtomsNotified;
    Scheduler*         mScheduler;  /* back pointer to scheduler */
    bool               mEnabledNextConnector;
    interactionsHashMap_type mStoredInteractions; /* store the interactions created through this connector */
    priorities_type  mDominatedPriorities; /* list of priorities in which the connector is dominated */
    priorities_type  mDominatorPriorities; /* list of priorities in which the connector is dominator */
    int                     mPriority;       /* static priority level */
    bool                    mSubconnSameHolder; /* used as a sub-connector
					          in hierarcical connector*/
	bool                    mRecomputexPort; /* recomputation of xPort() needed */

    Hidden::ConnInfo  mHiddenInfo;

  /*** Constructors & Dectructors ***/
  public:
    Connector(const char*, Compound*, const bool, const int, ...);
    Connector(const char*, Compound*, const bool); 
    virtual ~Connector();

  /*** Public Data-Access Methods ***/
  public:
    char* name() const
    {
      return mName;
    }

    Compound* holder() const
    {
      return mHolder;
    }

    ports_type& ports()
    {
      return mPorts;
    }

    const ports_type& ports() const
    {
      return mPorts;
    }

    Port*     xPort()     const { return mxPort; }
    void      setxPort(Port* p) { mxPort = p; }
    bool      legal()     const { return mLegal; }
    bool      dominated() const { return mDominated; }
    void      setDominated (bool dominated) { mDominated = dominated; }
    Interaction* feasibleInter() const { return mFeasibleInter; }
    void      setFeasibleInter(Interaction* i) { mFeasibleInter = i; }
    bool      isSubconn() const { return mSubconn; }
    bool      evaluated() const { return mEval; } 
    void      setEvaluated(bool b)   { mEval = b; }
    inline bool                isSubconnSameHolder() const { return mSubconnSameHolder; }
    inline void                setSubconnSameHolder(bool b) { mSubconnSameHolder = b; }
    inline        BIP::Time lowerApp()                                 { return mLowerApp; }
    inline        BIP::Time upperApp()                                 { return mUpperApp; }
    inline        Urgency    urgencyApp()                               { return mUrgencyApp; }
    inline        bool    isConstantApp()                               { return mIsConstantApp; }

    inline        void       setLowerApp( BIP::Time& L )                {  mLowerApp= L; }
    inline        void       setUpperApp( BIP::Time& U )                {  mUpperApp= U; }
    inline        void       setUrgencyApp( Urgency Ur)                 {  mUrgencyApp= Ur; }
    inline        void       setIsConstantApp( bool isCons)             {  mIsConstantApp= isCons; }

    inline        BIP::Time    deadlineApp()                           { return mDeadlineApp; }
    inline        void          setDeadlineApp(BIP::Time& D)            {mDeadlineApp= D; }
    inline        bool          enabledNextConnector()                  { return mEnabledNextConnector; }
    inline        int           atomsNotified()                         { return mAtomsNotified; }
    inline   priorities_type& dominatedPriorities()  { return mDominatedPriorities; }
    inline   priorities_type& dominatorPriorities()  { return mDominatorPriorities; }
    inline int                 getPriority() const { return mPriority; }
    inline void                setPriority(int p)  { mPriority = p; }
    inline bool                noPrio()    const   
                                     { return (mPriority == -9999); }
    inline bool                recomputexPort() const { return mRecomputexPort; }
    inline void                setRecomputexPort(const bool b) { mRecomputexPort = b; }
    connectors_type& links()
      {
        return mLinks;
      }

    Status    status()    const { return mStatus; }
    void setStatus(const Status status)  { mStatus = status; }

    //--- Hidden connector ----------
    inline bool  markedHidden() { return mHiddenInfo.impl!=NULL; }
    inline void  createHiddenInfo() { 
        assert(mHiddenInfo.impl==NULL); 
        mHiddenInfo.impl = new Hidden::ConnInfo::Impl;
    }
    inline Hidden::ConnInfo::Impl* hidden() { 
        assert(mHiddenInfo.impl!=NULL); 
        return mHiddenInfo.impl;
    } 

   /*** Public Utility Methods ***/
   public:
     char*        label();    
     Interaction* evaluate(); 
     void         execute();
     void         createFeasibleInter();
     void         resetLink();
     void         setLinkFalse();
     void         resetDependentLinks();
     bool         synced()   const;
     virtual bool complete() const;
     virtual bool guard();
     bool         maximal()  const;
     virtual const Intervals&       constraint(const BitVector&);
     virtual void up();
     virtual void dn();
     void         setVarPorts(const int, va_list);

     virtual void setup(Scheduler*);
     virtual bool isBroadcast() const;
     virtual void setupBroadcast(const int);

     virtual bool allAtomsReady() const;
     virtual bool disabled() const;
     virtual bool unknown() const;

     bool         holderReady();
     void         holderGoBusy(std::list<Atom*>& new_busy_atoms);
     void         addLink(Connector*);
     virtual void dump(FILE* file) const;
     Interaction* getInteraction(const BitVector &syncedPorts);
     
     
     void         computexPort();
     void         initializeLUUConnector(BIP::Time&);
     //void         updateLUUConnector(BIP::Time&, BIP::Time&, bool);
     void         resetEnabledConnector();
     void         initializeLUUConnectorDown(BIP::Time&, Port* fromDownPort);
     void         initializeLUUConnectorUp(BIP::Time& , Port* fromDownPort);
     int          atomsBusyNumber();
     bool         isNextConnector();
     void         updateLUU(BIP::Time&);
     void         add(Priority*);
     void         updateConnectorAppFromPort(Port*, bool cached);
    

   /*** Public Methods invoked from appln. ***/
   public:
     virtual void define() {}
};

#endif







