#ifndef PORT_H
#define PORT_H

#include "constraint.h"
#include <list>
#include <set>
#include <cassert>

class Atom;
class Connector;
class Port;
class SThread;

/*************************************
 * Port info for hidden interactions
 *************************************/

namespace Hidden {
    struct PortInfo {
        struct Impl;

        Impl* impl;
        PortInfo() { impl = NULL; }
        ~PortInfo() { if (impl) delete impl;}

        
        struct Impl {
            bool isMaster;
            bool isSlave;
            std::set<Connector*> rootConnectors;       
            Connector* choice; // for master
            std::vector<Port*>* conotified_ports; // for slaves
            Impl() { isMaster=false; isSlave = false; choice = NULL; conotified_ports = NULL;} 
            ~Impl() { 
                if (conotified_ports) delete conotified_ports;
           }
            void createConotifiedPorts() {
                assert(!conotified_ports);
                conotified_ports = new std::vector<Port*>;
            }
        };
    };
}

/***************************
 * Class: Port
 * Description: port class
 **************************/
class Port {
	
  public:
	typedef std::vector<Connector*> connectors_type;
	
	connectors_type mLinks; /* all connectors connected to this port */
	
	
	 /*** Protected data members ***/
  protected:
    char*      mName;       /* Name of the port */
    Atom*      mAtomHolder; /* holder for atomic port */
    Connector* mConnHolder; /* holer for conn port */
    bool       mSynced;     /* the holder syncs it     */
  
    bool       mElected;    /* the scheduler elects it */
    bool       mConnected;  /* i.e. connected, else internal label */
    bool       mInternal;   /* if internal port */
    
    Intervals  mConstraint;  /* time constraint  */
    
    Urgency    mPortUrgency;
//    bool       mGuardApp;
  
     BIP::Time  mLowerApp  ;   /*approximated lower bound of time constraint*/
     BIP::Time  mUpperApp ;   /*approximated Upper bound of time constraint*/
     Urgency    mUrgencyApp;   /*approximated Urgency of time constraint*/
     bool       mIsConstantApp;

     // To avoid race conditions, we need cached version of the above to be accessed by 
     // the scheduler when the atom holder is busy
     BIP::Time  mCachedLowerApp  ;   /*approximated lower bound of time constraint*/
     BIP::Time  mCachedUpperApp ;   /*approximated Upper bound of time constraint*/
     Urgency    mCachedUrgencyApp;   /*approximated Urgency of time constraint*/
     bool       mCachedIsConstantApp;

     Hidden::PortInfo mHiddenInfo;

  /*** Constructors & Dectructors ***/
  public:
    Port(const char*, 
         Atom*);
    Port(const char*,
         Connector*);
    virtual ~Port();

  /*** Public Data-Access Methods ***/
  public:
    inline char*      name()        const      { return mName; }
    inline Atom*      atomHolder()  const      { return mAtomHolder; }
           SThread*    threadHolder() const;
    inline Connector* connHolder()  const      { return mConnHolder; } 
    inline bool       synced()      const      { return mSynced; }


    inline void       setSynced(bool synced)   { mSynced = synced; }
   
    inline bool       elected()     const      { return mElected; }
    inline void       setElected(bool elected) { mElected = elected; }
    inline bool       isConnected() const      { return mConnected; }
    inline void       setConnected()           { mConnected = true; }
    inline bool       isInternal()  const      { return mInternal; }
    inline void       setInternal()            { mInternal = true; }
    inline        Intervals& constraint()                               { return mConstraint; }
    inline  const Intervals& constraint()                         const { return mConstraint; }
    inline        void       setConstraint(Intervals&  constraint)      {mConstraint = constraint;}
 //  inline        bool       guardApp()                                 { return mGuardApp; }
 //   inline  const bool       guardApp()                           const { return mGuardApp; }
 //   inline        void    setGuardApp(bool b)                           { mGuardApp = b;}
    inline        BIP::Time lowerApp(bool cached)                                 { return cached ? mCachedLowerApp : mLowerApp; }
    inline        BIP::Time upperApp(bool cached)                                 { return cached ? mCachedUpperApp : mUpperApp; }
    inline        Urgency    urgencyApp(bool cached)                              { return cached ? mCachedUrgencyApp : mUrgencyApp; }
    inline        bool      isConstantApp(bool cached)                            { return cached ? mCachedIsConstantApp : mIsConstantApp;}
    inline        void       setLowerApp( BIP::Time L )                {  mLowerApp= L; }
    inline        void       setUpperApp( BIP::Time U )                {  mUpperApp= U; }
    inline        void       setUrgencyApp( Urgency Ur)                 {  mUrgencyApp= Ur; }
    inline        void       setIsConstantApp( bool isCons)             {  mIsConstantApp= isCons; }
                  void       copyAppToCache();

  
    connectors_type& links()
      {
        return mLinks;
      }
  /*** Public Utility Methods ***/
  public:
    void execute();
    void addLink(Connector*);
    bool holderReady();
    void holderGoBusy(std::list<Atom*>&);
    void triggerAtom();
    //void updateApp();
    void initializeLUU(BIP::Time& );
    void initializeLUUDown(BIP::Time& );
    //void initializeAppFromConstraint();
    //bool isNext();
    void initializeNextPorts(BIP::Time& );
    void resetLink();
    void setLinkFalse();

    //--- Hidden port ----------
    inline bool  markedHidden() { return mHiddenInfo.impl!=NULL; }
    inline void  createHiddenInfo() { 
        assert(mHiddenInfo.impl==NULL); 
        mHiddenInfo.impl = new Hidden::PortInfo::Impl;
    }
    inline Hidden::PortInfo::Impl* hidden() { 
        assert(mHiddenInfo.impl!=NULL); 
        return mHiddenInfo.impl;
    } 
  /*** Public Methods invoked from appln. ***/
  public:
    bool complete() const;
    bool guard();

};

#endif
