#ifndef SCHEDULER_H
#define SCHEDULER_H

#include "bip-time.h"
#include "constraint.h"
#include "clock.h"
#include "structure.h"
#include <list>
#include <map>
#include <string>
#include <pthread.h>
#include <map>
#include <vector>
#include <set>

class Compound;
class Interaction;
class Atom;
class Port;
class Connector;
class SThread;
class BipThread;
struct Engine;


class Scheduler {
  public:

    typedef std::list<Interaction*> interactions_type;
    typedef std::list<Atom*> atoms_type;
    typedef std::list<Port*> ports_type;
    typedef std::list<Connector*> connectors_type;
    typedef std::list<SThread*> threads_type;
    typedef std::map<std::string,SThread*> threadmap_type;
    typedef std::pair<std::string,SThread*> threadmap_pair;
    typedef std::map<Atom*,std::vector<Port*>*> nextports_map_type;    

    

  	 
  private:
    Compound*        mSystem;      /* the system under test                */
    atoms_type mReady;       /* snapshot (list) of ready atoms       */
    std::set<Atom*> mExceptionallyBusy;
    interactions_type mPending;     /* list of pending interactions         */
    
    interactions_type mMaxInteractions; /* list of maximal interactions */
    int               mAtoms;       /* total number of atoms in the system  */
 //   int               mReadyAtoms;  /* number of ready atoms                */
    char              mInfo[512];   /* various infos buffer                 */
    atoms_type        mBusyAtomList;
    std::set<Connector*>  mNextConn;
    ports_type        mNextInternalPorts;
    Interaction*       mChoice;
    LogicalClock*      mLogicalClock;
    PlatformClock*     mPlatformClock;
    threads_type       mThreads;
    threadmap_type     mThreadMap;
    BipThread*         mBipThread;

    nextports_map_type mNextPortsMap; /* for busy atoms: next ports */  

    atoms_type         mPhysReadyAtoms; // atoms that should go ready after clock update
 
      
    pthread_cond_t    mResume;      /* block thread on a condition variable */
    pthread_mutex_t   mMutex;       /* mutex variable                       */
    
    /* execution parameters */
  
    bool              mQuiet;       /* 1 => dont show scheduled interactions */

  	bool              mLazy;        /* flag marking a lazy oracle */
    bool              mAlmostLazy;  /* almost lazy oracle (uses "have next ports" flag) */
  	int               mRandSeed;    /* seed for randomized execution         */
    FILE*             mOutFile;     /* contains simulation output trace      */
    BIP::Time         mMaxDrift;
    BIP::Time         mLimitTime;
    bool              mRealTime;

  public:
     pthread_cond_t condition_event; //= PTHREAD_COND_INITIALIZER;
     pthread_mutex_t mutex_event; //= PTHREAD_MUTEX_INITIALIZER;

       



 
#ifdef PERF
    BIP_Clock         comp_clk;     /* clock measuring computation time      */
    int               n_running_atoms; /* no. of running atoms at any moment */
#endif

  public:
    Scheduler(Compound*, bool, LogicalClock*, const Engine& engineParam); 
    virtual ~Scheduler();

    interactions_type& pending() { return mPending; } 
    atoms_type&        busyAtomList()  {return mBusyAtomList; } 
    std::set<Connector*>&   nextConn()  {return mNextConn; } 
    ports_type&        nextInternalPorts()  {return mNextInternalPorts; } 
    interactions_type& maxinteractions() { return mMaxInteractions; }
    bool isPending(Connector*);

    inline pthread_cond_t* resume()  { return &mResume; }
    inline pthread_mutex_t* mutex()  { return  &mMutex; }
   
    void putNextPortsInMap(Atom* atom);
    bool containedInNextPortsMap(Port* port);
    static void visitUpperConnectors( Connector* connector, std::set<Connector*>& visited);
    void extendNextConnectors(BIP::Time time, atoms_type& new_busy_atoms);
 
    void setup();
    void createThreads();  // implemented externally by bip compiler
    void createDefaultThreads();
    SThread *obtainThread(const char* name);
    BipThread *bipThread();
    void dumpThreads();
    void launchThreads();
    int threadCount() { return mThreads.size(); }

    void run();
    void selectEnabled(atoms_type&);
    void putReady(Atom*);
    bool haveReady();
    void getReady(atoms_type&, bool assumeUptodateClock);
    void getReadyAtoms(atoms_type&, bool assumeUptodateClock);
    void waitAllAtomsReady(atoms_type&);
    bool waitImportantAtomsReady(atoms_type&, bool assumeUptodateClock);
    bool actualizeNewReadyAtoms(atoms_type& newReadyAtoms, bool assumeUptodateClock);
    bool condTimedWait(timespec& absTime);
    bool checkAndPutReadyBipThreadAtoms(); 
    void updateLinks(atoms_type& newAtomList);
    void putExceptionallyBusy(Atom* atom);
    void eraseExceptionallyBusy(Atom* atom);
    void eraseExceptionallyBusy(std::list<Atom*>& atoms_list);
    void execute(Connector*, Interaction*, BIP::Time& );
    //void reEvaluate(atoms_type&);
    void incrAtomCount() { mAtoms++; }
    inline int nAtoms() { return mAtoms; }
    inline int nThreads() { return mThreads.size(); }

    void removeAtom(Atom*);    
    void loginfo(FILE*) const;

       inline LogicalClock*  logicalClock()    const { return mLogicalClock; }
       inline PlatformClock* platformClock()   const { return mPlatformClock; }
       inline Interaction*   choice()            const { return mChoice; }
       inline void           setChoice(Interaction* choice)  { mChoice =  choice; }

       inline BIP::Time      maxDrift()               const { return mMaxDrift; }
    inline bool quiet() { return mQuiet; } 

};
struct  Engine {
  Compound*         mSystem;      /* the system under test                */
  int               mIntLazy;
  bool              mLazy;        /* flag marking a lazy oracle */
  int               mRandSeed;    /* seed for randomized execution         */
  int               mRunMode; 
  int               mIntQuiet;
  bool              mQuiet;       /* 1 => dont show scheduled interactions */
  int               mMaxDrift;
  int               mLimitTime;

  Engine(); 
};

#endif
