/*******************************************************************************
 *                            C++ Verimag
 *    
 *
 *------------------------------------------------------------------------------
 * 	            Distributed and Complex Systems - Gires
 *------------------------------------------------------------------------------
 * constraint.h
 *------------------------------------------------------------------------------
 *  Time constraints implementation for BIP2 Engine
 *  + class Interval    : time interval with urgency type
 *  + class Constraint  : list for time intervals
 *------------------------------------------------------------------------------
 * Created by Jacques COMBAZ on Fri Nov 21 2008
 *
 ******************************************************************************/

#ifndef CONSTRAINT_H
#define CONSTRAINT_H

#include "bip-time.h"

#include <vector>
#include <iostream>

class Interval;
class BasicIntervals;
class Intervals;
class SingleClockConstraint;
class SubtractionClockConstraint;
class TimeConstraint;

class LogicalClock;
class Clock;

/* defines the types of urgency used in BIP, ordered by priority level */
typedef enum { LAZY = 0, DELAYABLE = 1, EAGER = 2 } Urgency;


/**************************************
 * Class: TypedInterval
 * Description:
 *   - time interval
 ****************************************/
class Interval {
  /*** Protected data members ***/
  protected:
    BIP::Time    mLow;         /* lower bound of the interval */
    BIP::Time    mHigh;        /* upper bound of the interval */
    bool         mDominated;   /* dominated in a priority rule */
    Urgency      mUrgency;     /* urgency type */
    

  /*** Constructors & Dectructors ***/
  public:
    Interval(Urgency);
    Interval(Urgency, bool);
    Interval(const BIP::Time&, const BIP::Time&, Urgency = LAZY, bool = false);
    Interval(const Interval &);
    Interval(const BIP::Time&, const BIP::Time&, const Interval &);
    Interval(const BIP::Time&, const BIP::Time&, const Interval &, bool);
    Interval(const Interval &, bool, Urgency = LAZY);
    virtual ~Interval();


  /*** Public Data-Access Methods ***/
  public:
    inline BIP::Time    low()                                    const { return mLow; }
    inline BIP::Time    high()                                   const { return mHigh; }
    inline void         setLow(const BIP::Time& low)                   { mLow  = low; }
    inline void         setHigh(const BIP::Time& high)                 { mHigh = high; }
           void         applyDelay(const BIP::Time& delay);
    inline bool         dominated()                              const { return mDominated; }
    inline void         setDominated(bool dominated)                   { mDominated = dominated; }
    inline Urgency      urgency()                                const { return mUrgency; }
    inline void         setUrgency(Urgency urgency)                    { mUrgency = urgency; }
    inline void         setMaxUrgency(Urgency urgency)                 { if (urgency > mUrgency) mUrgency = urgency; }
 

  /*** Public Utility Methods ***/
  public:
    inline bool      in(const BIP::Time& time)             const { return (low() <= time) && (time <= high()); }
    inline bool      inside(const BIP::Time& time)         const { return (low() <  time) && (time <  high()); }
    inline bool      in(const Interval& interval)          const { return (low() <= interval.low()) && (interval.high() <= high()); }
    inline bool      inside(const Interval& interval)      const { return (low() <  interval.low()) && (interval.high() <  high()); }
    inline void      setEmpty()                                  { setLow(BIP::Time::MAX); setHigh(BIP::Time::MIN); }
    inline bool      empty()                               const { return high() < low(); }
    inline bool      singleton()                           const { return low() == high(); }
 
           Interval& operator=(const Interval&);
           Interval& operator&=(const Interval&); /* intersection */
           Interval& operator-=(const Interval&); /* minus (set)  */

    inline       Interval& operator- ()                                { BIP::Time tmp = low(); setLow(-high()); setHigh(-tmp); return *this; }
    inline       Interval& operator+=(const BIP::Time& delay)          { setLow(low() + delay); setHigh(high() + delay); return *this; }  /* time translation */
    inline       Interval& operator-=(const BIP::Time& delay)          { setLow(low() - delay); setHigh(high() - delay); return *this; }  /* time translation */
    inline const Interval  operator+ (const BIP::Time& delay)    const { Interval interval = *this; interval += delay; return interval; } /* time translation */
    inline const Interval  operator- (const BIP::Time& delay)    const { Interval interval = *this; interval -= delay; return interval; } /* time translation */

    inline bool      operator< (const Interval& interval) const { return (interval.low() <  low()) && (high() <  interval.high()); } /* strict inclusion */
    inline bool      operator<=(const Interval& interval) const { return (interval.low() <= low()) && (high() <= interval.high()); } /*        inclusion */
    inline bool      operator> (const Interval& interval) const { return (low() <  interval.low()) && (interval.high() <  high()); }
    inline bool      operator>=(const Interval& interval) const { return (low() <= interval.low()) && (interval.high() <= high()); }
    inline bool      operator==(const Interval& interval) const { return (low() == interval.low()) && (interval.high() == high()); }
    inline bool      operator!=(const Interval& interval) const { return (low() != interval.low()) || (interval.high() != high()); }

    friend std::ostream&  operator<<(std::ostream &, const Interval&);
};


class BasicIntervals {
  /*** Private data members ***/
  private:
    mutable const BasicIntervals* mCopy;
            const bool            mConstant;
    mutable std::vector<Interval> mIntervals;

  /*** Constructors & Dectructors ***/
  private:
    BasicIntervals(bool);
    BasicIntervals(const Interval&, bool);

  public:
    BasicIntervals();
    BasicIntervals(const Interval&);
    BasicIntervals(const BasicIntervals&);
    virtual ~BasicIntervals();

  /*** Private Utility Methods ***/
  private:
    inline bool                  isConstant()                             const { return mConstant; }
    inline bool                  isCopy()                                 const { return mCopy != NULL; }
    inline const BasicIntervals& copy()                                   const { return *mCopy; }
    inline void                  setCopy(const BasicIntervals& intervals) const { mCopy = &intervals; }
    inline void                  unCopy()                                 const { if (isCopy()) { mIntervals = copy().mIntervals; mCopy = NULL; } }

  /*** Protected Utility Methods ***/
  protected:
    inline void erase(int low, int high) {
      unCopy();
      mIntervals.erase(mIntervals.begin() + low,
                       mIntervals.begin() + high + 1);
    }

    inline void insert(int index, const Interval& interval) {
      if (!interval.empty()) {
        unCopy();
        mIntervals.insert(mIntervals.begin() + index, interval);
      }
    }

    inline const Interval& operator[](unsigned int index) const {
      if (!isCopy()) {
        return mIntervals[index];
      }
      else {
	return copy().mIntervals[index];
      }
    }

    inline Interval& operator[](unsigned int index) {
      unCopy();
      return mIntervals[index];
    }

    inline bool empty() const {
      bool ret;
      if (isFalse()) {
        ret = true;
      }
      else if (isTrue()) {
        ret = false;
      }
      else if (isCopy()) {
        ret = copy().mIntervals.empty();
      }
      else {
        ret = mIntervals.empty();
        if (ret) { setCopy(FALSE()); }
      }
      return ret;
    }

    inline int size() const {
      if (isFalse()) {
        return 0;
      }
      else if (isTrue()) {
        return 1;
      }
      else if (isCopy()) {
        return copy().mIntervals.size();
      }
      else {
        return mIntervals.size();
      }
    }

  /*** Private Constants ***/
  private:
    static BasicIntervals mFalse;
    static BasicIntervals mTrueLazy;
    static BasicIntervals mTrueLazyDominated;
    static BasicIntervals mTrueDelayable;
    static BasicIntervals mTrueEager;

  /*** Public Utility Methods ***/
  public:
    inline bool isFalse()             const { if (isCopy()) return mCopy == &FALSE(); else return this == &FALSE() || mIntervals.empty(); }
    inline bool isTrue()              const { return isTrueLazy() || isTrueLazyDominated() || isTrueDelayable() || isTrueEager(); }
    inline bool isTrueLazy()          const { return mCopy == &TRUE_LAZY()           || this == &TRUE_LAZY(); }
    inline bool isTrueLazyDominated() const { return mCopy == &TRUE_LAZY_DOMINATED() || this == &TRUE_LAZY_DOMINATED(); }
    inline bool isTrueDelayable()     const { return mCopy == &TRUE_DELAYABLE()      || this == &TRUE_DELAYABLE(); }
    inline bool isTrueEager()         const { return mCopy == &TRUE_EAGER()          || this == &TRUE_EAGER(); }

    static void initializeConstants();

    inline static const BasicIntervals& FALSE()               { return mFalse; }
    inline static const BasicIntervals& TRUE_LAZY()           { return mTrueLazy; }
    inline static const BasicIntervals& TRUE_LAZY_DOMINATED() { return mTrueLazyDominated; }
    inline static const BasicIntervals& TRUE_DELAYABLE()      { return mTrueDelayable; }
    inline static const BasicIntervals& TRUE_EAGER()          { return mTrueEager; }

           BasicIntervals& operator= (const BasicIntervals&);
    inline bool            operator==(const BasicIntervals& intervals) const;
    inline bool            operator!=(const BasicIntervals& intervals) const;

           void applyDomination();
    inline void clear() { setCopy(FALSE()); }
           void reset(Urgency = LAZY);
           void clean();
};


/***********************************************
 * Class: Intervals
 * Description:
 *   - disjunction of intervals w.r.t.
 *     one (x in ...) or two clocks (x-y in ...) 
 ***********************************************/
class Intervals : public BasicIntervals {
  /*** Constructors & Dectructors ***/
  public:
    Intervals();
    Intervals(const Interval&);
    Intervals(const BasicIntervals&);
    Intervals(const Intervals&);
    Intervals(const BIP::Time&, const BIP::Time&, Urgency);
    virtual ~Intervals();

  /*** Protected Utility Methods ***/
  protected:
           int       before(const BIP::Time&) const;
           int       after(const BIP::Time&) const;
           int       position(const BIP::Time&, int&) const;
           int       position(const BIP::Time&) const;
           bool      intersect(const Interval&, int&, int&) const;
           bool      erase(int, int);
           bool      setDominated(int, int, bool);
           void      setAllDominated(bool);
           bool      setUrgency(int, int, Urgency);
           bool      setMaxUrgency(int, int, Urgency);

  /*** Public Utility Methods ***/
  public:
    inline bool      in(const BIP::Time& time) const { return position(time) >= 0; }
           bool      in(const Interval &) const;
           bool      urgent(const BIP::Time&) const;                        /* return true iff the time constraint is urgent at time  */
           bool      existUrgency(int, int, Urgency) const;
    inline BIP::Time first() const { if (size() > 0) return (*this)[0].low();           else return BIP::Time::MAX; }
    inline BIP::Time last()  const { if (size() > 0) return (*this)[size() - 1].high(); else return BIP::Time::MIN; }
           BIP::Time deadline(const BIP::Time&) const;                      /* return the next time urgencyPredicate() is true after time  */
           BIP::Time next(const BIP::Time&) const;                          /* return the next time the guard is true after time           */
           bool      check() const;                                         /* debug function */

           void      minus(const Interval&, bool);
           void      minus(const Intervals&, const BIP::Time&, bool);

                 Intervals& operator-=(const Interval&);     /* minus (set)  operator */
                 Intervals& operator-=(const Intervals&);    /* minus (set)  operator */
                 Intervals& operator&=(const Interval&);     /* intersection operator */
                 Intervals& operator&=(const Intervals&);    /* intersection operator */
                 Intervals& operator|=(const Interval&);     /* union        operator */
                 Intervals& operator|=(const Intervals&);    /* union        operator */

                 void              min(const Intervals&);    /* min operator */
                 void              max(const Intervals&);    /* max operator */

    inline const Intervals  operator& (const Interval& i)  const { Intervals r = *this; r &= i; return r; }
    inline const Intervals  operator& (const Intervals& c) const { Intervals r = *this; r &= c; return r; }
    inline const Intervals  operator- (const Interval& i)  const { Intervals r = *this; r -= i; return r; }
    inline const Intervals  operator- (const Intervals& c) const { Intervals r = *this; r -= c; return r; }
    inline const Intervals  operator| (const Interval& i)  const { Intervals r = *this; r |= i; return r; }
    inline const Intervals  operator| (const Intervals& c) const { Intervals r = *this; r |= c; return r; }
                 Intervals& operator= (const Intervals&);

                 Intervals& operator+=(const BIP::Time&);    /* time translation operator */
                 Intervals& operator-=(const BIP::Time&);    /* time translation operator */


  /*** Friend Operators ***/
  public:
    friend std::ostream&  operator<<(std::ostream &, const Intervals&); /* output operator */
    friend std::ostream&  operator<<(std::ostream &, const Intervals*); /* output operator */
};


/***********************************************
 * Class: SingleClockConstraint
 * Description:
 *   constraint of the form x in [L,U]
 ***********************************************/

class SingleClockConstraint {

  /*** Protected data members ***/
  protected:
    const Clock* mClock;      /* first clock x */
    Interval     mInterval;   /* interval */

  /*** Constructors & Dectructors ***/
  public:
    SingleClockConstraint(const Clock&, const BIP::Time&, const BIP::Time&);
    SingleClockConstraint(const SingleClockConstraint&);
    virtual ~SingleClockConstraint();

  /*** Public Data-Access Methods ***/
  public:
    const inline Clock&    clock()       const { return *mClock; }
    const inline Interval& interval()    const { return mInterval; }

  /*** Public Utility Methods ***/
  public:
    SingleClockConstraint& operator=(const SingleClockConstraint&);
};


/***********************************************
 * Class: SubtractionClockConstraint
 * Description:
 *   constraint of the form x-y in [L,U]
 ***********************************************/

class SubtractionClockConstraint {

  /*** Protected data members ***/
  protected:
    const Clock* mClock;      /* first clock x */
    const Clock* mMinusClock; /* second clock y */
    Interval     mInterval;   /* interval */

  /*** Constructors & Dectructors ***/
  public:
    SubtractionClockConstraint(const Clock&, const Clock&, const BIP::Time&, const BIP::Time&);
    SubtractionClockConstraint(const SubtractionClockConstraint&);
    virtual ~SubtractionClockConstraint();

  /*** Public Data-Access Methods ***/
  public:
    const inline Clock&    clock()       const { return *mClock; }
    const inline Clock&    minusClock()  const { return *mMinusClock; }
    const inline Interval& interval()    const { return mInterval; }

  /*** Public Utility Methods ***/
  public:
    SubtractionClockConstraint& operator=(const SubtractionClockConstraint&);
};


/****************************************
 * Class: TimeConstraint
 * Description:
 *   - conjunction of clock constraints
 ****************************************/
class TimeConstraint {

  /*** Protected data members ***/
  protected:
                Urgency                     mUrgency;                /* global urgency */
    std::vector<SingleClockConstraint>      mSingleConstraints;      /* list of single clock constraints */
    std::vector<SubtractionClockConstraint> mSubtractionConstraints; /* list of subtraction clock constraint */
    mutable     bool                        mEvaluated;              /* true: need recomputation of the conjunction */
    const       LogicalClock&               mClock;                  /* logical clock used for the computation of the conjunction */
    mutable     Intervals                   mConstraint;             /* used for dynamic computation of conjunction of mConstraints */

  /*** Constructors & Dectructors ***/
  public:
    TimeConstraint(const LogicalClock&, Urgency = LAZY);
    virtual ~TimeConstraint();

  /*** Protected Data-Access Methods ***/
  protected:
    inline       bool             evaluated()          const { return mEvaluated; }
    inline       void             setEvaluated(bool b) const { mEvaluated = b; }

  /*** Public Data-Access Methods ***/
  public:
    inline const Intervals&       constraint()         const { if (!evaluated()) computeConstraint(); return mConstraint; }
    inline const LogicalClock&    clock()              const { return mClock; }


  /*** Protected Utility Methods ***/
  protected:
           void computeConstraint() const;


  /*** Public Utility Methods ***/
  public:
    inline Urgency urgency()             const { return mUrgency; }
    inline void    setUrgency(Urgency urgency) { mUrgency = urgency; }                  // set the urgency

           void    add(const Clock&,               const BIP::Time&, const BIP::Time&); // add a single clock constraint
           void    add(const Clock&, const Clock&, const BIP::Time&, const BIP::Time&); // add a subtraction clock constraint

    inline void    reset() const { setEvaluated(false); } // force recomputation of the conjonction of clock constraints
           void    clear();                               // reinitialize the time constraint

};

#endif
