# ##########################################################################################################
# Title: Taste-IV 2 FPPN-DOLC modules
# Author: Fotios Gioulekas
# Date: 2015/10/12
# Project:MoSaTT-CMP
# Name: fppnxml.py
# Version: 8.0
# Abstract:
# This is a python module that contains the functions which are called by the TASTE-IV 2 FPPN-DOLC generator
# for creating the FPPN-DOLC.xml file
############################################################################################################
# import the GenModules module
import genmodules
import logging
import os


###################################################################################33
# The prettify function is used for the generation of a readable XML file
# Return a pretty-printed XML string for the Element elem
# inputs: (1) elem: XML element
#         (2) ET:ET API
#         (3) minidom: minidom API
# output: beautify XML tree that is user readable
#######################################################################################

def prettify(elem, ET, minidom):
    rough_string = ET.tostring(elem)  # It should be checked how to include encoding ='utf-8')
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="  ")


# The createDolcSubElements function performs as follows:
# For each one of the process element of the FPPN-DOLC XML model
# insert the child element source with the attribute location defining its associated C code
# named label.C, where the label is taken from attributes of the FunctionObjects within
# the corresponding InterfaceView.xml. It also creates the process elements,
# their ports and associated controllers. Additionally, it extracts the channel elements.
# function name: createDolcSubElements
# inputs: (1)app: element root of the FPPNDOLC.xml file
#         (2)functionlist:list with function objects and their characteristics
#         (3)ET: ET API
#         (4)childrenlist: dictionary of children dictionaries
#         (5)xmlroot: the xmlroot of the interfaceview.xml file
#         (6)path: it is the path where the taste project is located
#         (7)pathforfppnsrc: it is the path where the taste2dolc generator will place the generated c files
# outputs:(1) app: the updated XML tree in DOLC format
#         (2) pervals = {label1:{period : #, deadline: #, fpriority: #, WCET : #, criticality: #}, label2{...} ...}
#             it is for the periodic processes for now.
#         (3) cleanchains: the precedence chain strings



def createDolcSubElements(app, functionlist, ET, childrenlist, xmlroot, path, pathforfppnsrc):
    # create the FPPN-DOLC processes basic elements
    # initialize the dictionary with periodic processes characteristics
    pervals = {}
    # important keys for dictionary with process characteristics (ptype = periodic or imperiodic, m=ceiling (Tuser/Tsporadic)
    # when type = periodic is not taken into account
    charkeys = ["period","deadline", "WCET", "fpriority", "criticality", "ptype", "m"]

    for item in functionlist:

        # print(item)
        # get the FPPN class of the corresponding function object
        value = item.get('TASTE_IV_Properties-FPPNClass')
        # get the source code language of the corresponding function object
        source_code = item.get('Source_Language')
        # get the id of the corresponding function object
        idf = item.get('id')
        # get the label of the corresponding function object
        label = item.get('label')
        # find the children components of the function object
        childcomponents = childrenlist[label]

        # print("features\n")
        # print(childcomponents)
        numOfFeatures = childrenlist[label]['numofFEATUREelements']
        # print("\nnum of Features")
        # print(numOfFeatures)
        # print("label")
        # print(label)
        # print("child components")
        # print(childcomponents)

        # find the children of the function objects with element tag FEATURE
        # The FEATURE child corresponds to the interfaces of the function object (i.e. process or channel or
        # sporadic protocol)
        childcomponentsfeat = childrenlist[label]['FEATURE']
        # print("\n child features")
        # print(childcomponentsfeat)

        # call the findProvidedIfs function
        providedIFlist = genmodules.findProvidedIfs(xmlroot)

        # print("\nprovided ifs")
        # print(providedIFlist)
        # find if the ids of the FEATURES of the Function Units that are included as values in the provided interfaces
        # dictionaries
        # initialize the list of dictionaries for storing provided interfaces that are children (features) of the
        # Function units in listpprovin list
        listpprovin = []
        for childFeature in childcomponentsfeat:
            for itemProvIf in providedIFlist:
                # find if the childFeature is included as a value in the dictionary itemProvIf
                if childFeature in itemProvIf.values():
                    # if yes append this to the listpprovin
                    listpprovin.append(itemProvIf)
        # for debugging purposes
        # print("\n list of provided IFs that have the ids:")
        # for debugging purposes
        # print(listpprovin)
        # print(float(listpprovin[0]['Deadline']) / 1000)
        # print(len(listpprovin))
        # print('The id of the FunctionObject is:' + item.get('id'))
        # print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])

        if source_code == 'C':
            # =====================
            # PERIODIC PROCESS A
            # =====================

            if value == 'periodic_process_A':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the periodic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the periodic process = ' + item.get('label'))
                com1 = ET.Comment("Periodic Process: " + item.get('label').lower())
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='A')
                app.append(process)

                # source = ET.SubElement(process, 'source', location=item.get('label') + '.c')
                ET.SubElement(process, 'source', location=str.lower(item.get('label')) + '.c')
                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3

                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:
                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1


                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1

                                    else:

                                        pass
                                    index +=index



                    # create the associated controller
                    com2 = ET.Comment("Periodic Controller of process: " + item.get('label').lower())
                    app.append(com2)
                    if 'Deadline' not in listpprovin[0] or 'RCMperiod' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values')
                        logging.debug(
                            'The id of the ProvidedInterfaceObject in TASTE XML file is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))
                        # Create child activation for the controller
                        # cross check validation for periodic provided if
                        if listpprovin[0]['RCMoperationKind'] == 'cyclic':
                            activation = ET.SubElement(controller, 'activation', type='periodic')
                        else:
                            print(
                                '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic\n')
                            print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            print(
                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                            logging.debug(
                                'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic')
                            logging.debug(
                                'The id of the ProvidedInterfaceObject in TASTE XML file is:' + listpprovin[0]['id'])
                            exit(-1)
                        # Create child parameter for the activation tag
                        # Note: the RCMperiod is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        ET.SubElement(activation, 'parameter', name='period',
                                      value=str(float(listpprovin[0]['RCMperiod']) / 1000))

                        # Create port child in controller of type output to the associated process
                        port = ET.SubElement(controller, 'port', type='out_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_start')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='start')
                        # Create port child in controller of type input from the associated process
                        port = ET.SubElement(controller, 'port', type='in_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_finish')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='finish')

                        app.append(controller)
                    # create the control channels for the control connections between process to controller and
                    # vice-versa
                    com3 = ET.Comment("Control-channel from process " + item.get('label').lower() + ' to controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment("Control-channel from controller to process " + item.get('label').lower())
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)
                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                ########################
                #  C 2 c generation    #
                ########################

                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])

                if 'Compute_Execution_Time' not in listpprovin[0] :
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                elif item.get("TASTE_IV_Properties-Fpriority") is None:
                    print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                    exit(-1)
                else:

                    datavalues = [float(listpprovin[0]['RCMperiod']),float(listpprovin[0]['Deadline']),
                              float(listpprovin[0]['Compute_Execution_Time']),item.get("TASTE_IV_Properties-Fpriority"),
                              'A','periodic',0]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []



            # =====================
            #  PERIODIC PROCESS B
            # =====================

            elif value == 'periodic_process_B':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the periodic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the periodic process = ' + item.get('label'))
                com1 = ET.Comment("Periodic Process: " + item.get('label').lower())
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='B')
                app.append(process)
                # source = ET.SubElement(process, 'source', location=item.get('label') + '.c')
                ET.SubElement(process, 'source', location=str.lower(item.get('label')) + '.c')
                # app.insert(2, source)
                # app.append(ET.SubElement(process,'source',location=item.get('label').lower()+'.c'))

                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    com2 = ET.Comment("Periodic Controller of process: " + item.get('label').lower())
                    app.append(com2)
                    if 'Deadline' not in listpprovin[0] or 'RCMperiod' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values')
                        logging.debug(
                            'The id of the ProvidedInterfaceObject in TASTE XML file is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))
                        # Create child activation for the controller
                        # cross check validation for periodic provided if
                        if listpprovin[0]['RCMoperationKind'] == 'cyclic':
                            activation = ET.SubElement(controller, 'activation', type='periodic')
                        else:
                            print(
                                '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic\n')
                            print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            print(
                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                            logging.debug(
                                'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic')
                            logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            exit(-1)
                        # Create child parameter for the activation tag
                        # Note: the RCMperiod is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        ET.SubElement(activation, 'parameter', name='period',
                                      value=str(float(listpprovin[0]['RCMperiod']) / 1000))

                        # Create port child in controller of type output to the associated process
                        port = ET.SubElement(controller, 'port', type='out_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_start')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='start')
                        # Create port child in controller of type input from the associated process
                        port = ET.SubElement(controller, 'port', type='in_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_finish')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='finish')

                        app.append(controller)

                    # create the control channels for the control connections between process to controller and
                    # vice-versa
                    com3 = ET.Comment("Control-channel from process " + item.get('label').lower() + ' to controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment("Control-channel from controller to process " + item.get('label').lower())
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)


                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                    # #######################
                    #  c 2 c generation
                    #####################

                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])

                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                elif item.get("TASTE_IV_Properties-Fpriority") is None:
                    print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                    exit(-1)
                else:
                    datavalues = [float(listpprovin[0]['RCMperiod']),float(listpprovin[0]['Deadline']),
                              float(listpprovin[0]['Compute_Execution_Time']),item.get("TASTE_IV_Properties-Fpriority"),
                              'B', 'periodic',0]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []

            # =====================
            # PERIODIC PROCESS C
            # =====================

            elif value == 'periodic_process_C':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the periodic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the periodic process = ' + item.get('label'))
                com1 = ET.Comment("Periodic Process: " + item.get('label'))
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='C')
                app.append(process)
                # source = ET.SubElement(process, 'source', location=item.get('label') + '.c')
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # app.insert(2, source)
                # app.append(ET.SubElement(process,'source',location=item.get('label')+'.c'))

                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index


                    com2 = ET.Comment("Periodic Controller of process: " + item.get('label'))
                    app.append(com2)
                    if 'Deadline' not in listpprovin[0] or 'RCMperiod' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values')
                        logging.debug(
                            'The id of the ProvidedInterfaceObject in TASTE XML file is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))
                        # Create child activation for the controller
                        # cross check validation for periodic provided if
                        if listpprovin[0]['RCMoperationKind'] == 'cyclic':
                            activation = ET.SubElement(controller, 'activation', type='periodic')
                        else:
                            print(
                                '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic\n')
                            print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            print(
                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                            logging.debug(
                                'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic')
                            logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            exit(-1)
                        # Create child parameter for the activation tag
                        # Note: the RCMperiod is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        ET.SubElement(activation, 'parameter', name='period',
                                      value=str(float(listpprovin[0]['RCMperiod']) / 1000))

                        # Create port child in controller of type output to the associated process
                        port = ET.SubElement(controller, 'port', type='out_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_start')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='start')
                        # Create port child in controller of type input from the associated process
                        port = ET.SubElement(controller, 'port', type='in_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_finish')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='finish')

                        app.append(controller)

                    # create the control channels for the control connections between process to controller and
                    # vice-versa
                    com3 = ET.Comment("Control-channel from process " + item.get('label') + ' to controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment("Control-channel from controller to process " + item.get('label'))
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)


                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################
                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])

                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                elif item.get("TASTE_IV_Properties-Fpriority") is None:
                    print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                    exit(-1)
                else:
                    datavalues = [float(listpprovin[0]['RCMperiod']),float(listpprovin[0]['Deadline']),
                              float(listpprovin[0]['Compute_Execution_Time']),item.get("TASTE_IV_Properties-Fpriority"),
                              'C', 'periodic',0]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []

            # =====================
            # PERIODIC PROCESS D
            # =====================

            elif value == 'periodic_process_D':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the periodic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the periodic process = ' + item.get('label'))
                com1 = ET.Comment("Periodic Process: " + item.get('label'))
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='D')
                app.append(process)
                # source = ET.SubElement(process, 'source', location=item.get('label') + '.c')
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # app.insert(2, source)
                # app.append(ET.SubElement(process,'source',location=item.get('label')+'.c'))

                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    com2 = ET.Comment("Periodic Controller of process: " + item.get('label').lower())
                    app.append(com2)
                    #print('PRINT LISTOFPROVIN {}\n'.format(listpprovin[0]))
                    if 'Deadline' not in listpprovin[0] or 'RCMperiod' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values')
                        logging.debug(
                            'The id of the ProvidedInterfaceObject in TASTE XML file is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))
                        # Create child activation for the controller
                        # cross check validation for periodic provided if
                        if listpprovin[0]['RCMoperationKind'] == 'cyclic':
                            activation = ET.SubElement(controller, 'activation', type='periodic')
                        else:
                            print(
                                '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic\n')
                            print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            print(
                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                            logging.debug(
                                'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic')
                            logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            exit(-1)
                        # Create child parameter for the activation tag
                        # Note: the RCMperiod is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        ET.SubElement(activation, 'parameter', name='period',
                                      value=str(float(listpprovin[0]['RCMperiod']) / 1000))

                        # Create port child in controller of type output to the associated process
                        port = ET.SubElement(controller, 'port', type='out_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_start')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='start')
                        # Create port child in controller of type input from the associated process
                        port = ET.SubElement(controller, 'port', type='in_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_finish')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='finish')

                        app.append(controller)

                    # create the control channels for the control connections between process to controller and
                    # vice-versa
                    com3 = ET.Comment("Control-channel from process " + item.get('label').lower() + ' to controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment("Control-channel from controller to process " + item.get('label'))
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)


                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################
                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])


                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                elif item.get("TASTE_IV_Properties-Fpriority") is None:
                    print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                    exit(-1)
                else:
                    datavalues = [float(listpprovin[0]['RCMperiod']),float(listpprovin[0]['Deadline']),
                              float(listpprovin[0]['Compute_Execution_Time']),item.get("TASTE_IV_Properties-Fpriority"),
                              'D', 'periodic',0]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []

            # =====================
            # PERIODIC PROCESS E
            # =====================

            elif value == 'periodic_process_E':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the periodic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the periodic process = ' + item.get('label'))
                com1 = ET.Comment("Periodic Process: " + item.get('label'))
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='E')
                app.append(process)
                # source = ET.SubElement(process, 'source', location=item.get('label') + '.c')
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # app.insert(2, source)
                # app.append(ET.SubElement(process,'source',location=item.get('label')+'.c'))

                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    com2 = ET.Comment("Periodic Controller of process: " + item.get('label').lower())
                    app.append(com2)
                    #print('PRINT LISTOFPROVIN {}\n'.format(listofprovin[0]))
                    if 'Deadline' not in listpprovin[0] or 'RCMperiod' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline or Period values')
                        logging.debug(
                            'The id of the ProvidedInterfaceObject in TASTE XML file is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))
                        # Create child activation for the controller
                        # cross check validation for periodic provided if
                        if listpprovin[0]['RCMoperationKind'] == 'cyclic':
                            activation = ET.SubElement(controller, 'activation', type='periodic')
                        else:
                            print(
                                '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic\n')
                            print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            print(
                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                            logging.debug(
                                'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic')
                            logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                            exit(-1)
                        # Create child parameter for the activation tag
                        # Note: the RCMperiod is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        ET.SubElement(activation, 'parameter', name='period',
                                      value=str(float(listpprovin[0]['RCMperiod']) / 1000))

                        # Create port child in controller of type output to the associated process
                        port = ET.SubElement(controller, 'port', type='out_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_start')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='start')
                        # Create port child in controller of type input from the associated process
                        port = ET.SubElement(controller, 'port', type='in_event',
                                             name='p_CTRL_' + item.get('label').lower() + '_finish')
                        # Create event child in port tag
                        ET.SubElement(port, 'event', name='finish')

                        app.append(controller)

                    # create the control channels for the control connections between process to controller and
                    # vice-versa
                    com3 = ET.Comment("Control-channel from process " + item.get('label').lower() + ' to controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment("Control-channel from controller to process " + item.get('label'))
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)


                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################
                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])


                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                elif item.get("TASTE_IV_Properties-Fpriority") is None:
                    print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                    exit(-1)
                else:
                    datavalues = [float(listpprovin[0]['RCMperiod']),float(listpprovin[0]['Deadline']),
                              float(listpprovin[0]['Compute_Execution_Time']),item.get("TASTE_IV_Properties-Fpriority"),
                              'E', 'periodic',0]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []


            # =====================
            # SPORADIC PROCESS A
            # =====================

            elif value == 'sporadic_process_A':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the sporadic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the sporadic process = ' + item.get('label'))
                com1 = ET.Comment("Sporadic Process: " + item.get('label'))
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='A')
                app.append(process)
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    # create the associated controller
                    com2 = ET.Comment("Aperiodic Controller of aperiodic process: " + item.get('label'))
                    app.append(com2)
                    if 'Deadline' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline value\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline value')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))

                        # call the object id of the sporadic process function
                        sporprocid = item.get('id')

                        # Create child activation for the controller
                        # cross check validation for aperiodic provided interface

                    # call the ctrlaperiodic function for the definition of the activation

                    mmax, interval, protocolperiod, protocolsource, fpuser = genmodules.ctrlaperiodic(xmlroot, idf,
                                                                                              childrenlist, label)
                    if listpprovin[0]['RCMoperationKind'] == 'sporadic':
                        activation = ET.SubElement(controller, 'activation', type='aperiodic')
                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    # Create child parameters for the activation tag

                    ET.SubElement(activation, 'parameter', name='m_max', value=str(mmax))
                    ET.SubElement(activation, 'parameter', name='interval', value=str(interval))
                    ET.SubElement(activation, 'parameter', name='protocol_period', value=str(protocolperiod))
                    ET.SubElement(activation, 'parameter', name='protocol_source', value=str(protocolsource))

                    # Create port child in controller of type output to the associated process
                    port = ET.SubElement(controller, 'port', type='out_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in controller of type input from the associated process
                    port = ET.SubElement(controller, 'port', type='in_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')

                    # Append aperiodic controller to xml file
                    app.append(controller)

                    # Control channels between aperiodic controller and aperiodic process
                    com3 = ET.Comment(
                        "Control-channel from aperiodic process " + item.get(
                            'label').lower() + ' to aperiodic controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment(
                        "Control-channel from aperiodic controller to aperiodic process " + item.get('label'))
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)
                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################
                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])


                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                # elif item.get("TASTE_IV_Properties-Fpriority") is None:
                #     print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                #     logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                #     exit(-1)

                elif float(listpprovin[0]['Deadline']) < float(interval*1000): # the deadlline of the sporadic process is lower than the period of the user in ms
                    print("\n---ERROR! The Deadline of the sporadic process '{}' is lower than the period of the user process! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("The Deadline of the sporadic process '{}' is lower than the period of the user process".format(item.get("label")))
                    exit(-1)

                else:
                    fspriority = 0
                    datavalues = [float(interval*1000),float(listpprovin[0]['Deadline'])-float(interval*1000),
                              float(listpprovin[0]['Compute_Execution_Time']),int(fspriority),
                              'A','imperiodic', mmax]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []


            # =====================
            # SPORADIC PROCESS B
            # =====================

            elif value == 'sporadic_process_B':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the sporadic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the sporadic process = ' + item.get('label'))
                com1 = ET.Comment("Sporadic Process: " + item.get('label'))
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='B')
                app.append(process)
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    # create the associated controller
                    com2 = ET.Comment("Aperiodic Controller of aperiodic process: " + item.get('label').lower())
                    app.append(com2)

                    if 'Deadline' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline value\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline value')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))

                        # call the object id of the sporadic process function
                        sporprocid = item.get('id')

                        # Create child activation for the controller
                        # cross check validation for aperiodic provided interface

                    # call the ctrlaperiodic function for the definition of the activation

                    mmax, interval, protocolperiod, protocolsource, fpuser = genmodules.ctrlaperiodic(xmlroot, idf,
                                                                                              childrenlist, label)
                    print('\n####\n')
                    print('nmax {}, interval {}, properiod {}, source {}\n'.format(mmax, interval, protocolperiod,
                                                                                   protocolsource))
                    if listpprovin[0]['RCMoperationKind'] == 'sporadic':
                        activation = ET.SubElement(controller, 'activation', type='aperiodic')
                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    # Create child parameters for the activation tag

                    ET.SubElement(activation, 'parameter', name='m_max', value=str(mmax))
                    ET.SubElement(activation, 'parameter', name='interval', value=str(interval))
                    ET.SubElement(activation, 'parameter', name='protocol_period', value=str(protocolperiod))
                    ET.SubElement(activation, 'parameter', name='protocol_source', value=str(protocolsource))

                    # Create port child in controller of type output to the associated process
                    port = ET.SubElement(controller, 'port', type='out_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in controller of type input from the associated process
                    port = ET.SubElement(controller, 'port', type='in_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')

                    # Append aperiodic controller to xml file
                    app.append(controller)

                    # Control channels between aperiodic controller and aperiodic process
                    com3 = ET.Comment(
                        "Control-channel from aperiodic process " + item.get(
                            'label').lower() + ' to aperiodic controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment(
                        "Control-channel from aperiodic controller to aperiodic process " + item.get('label'))
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)
                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################
                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])



                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                # elif item.get("TASTE_IV_Properties-Fpriority") is None:
                #     print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                #     logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                #     exit(-1)

                elif float(listpprovin[0]['Deadline']) < float(interval*1000): # the deadline of the sporadic process is less than the period of the user in ms
                    print("\n---ERROR! The Deadline of the sporadic process '{}' is lower than the period of the user process! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("The Deadline of the sporadic process '{}' is lower than the period of the user process".format(item.get("label")))
                    exit(-1)

                else:
                    fspriority = 0
                    datavalues = [float(interval*1000),float(listpprovin[0]['Deadline'])-float(interval*1000),
                              float(listpprovin[0]['Compute_Execution_Time']),int(fspriority),
                              'B', 'imperiodic', mmax]

                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []


            # =====================
            # SPORADIC PROCESS C
            # =====================

            elif value == 'sporadic_process_C':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the sporadic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the sporadic process = ' + item.get('label'))
                com1 = ET.Comment("Sporadic Process: " + item.get('label'))
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='C')
                app.append(process)
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    # create the associated controller
                    com2 = ET.Comment("Aperiodic Controller of aperiodic process: " + item.get('label').lower())
                    app.append(com2)
                    if 'Deadline' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline value\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline value')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))

                        # call the object id of the sporadic process function
                        sporprocid = item.get('id')

                        # Create child activation for the controller
                        # cross check validation for aperiodic provided interface

                    # call the ctrlaperiodic function for the definition of the activation

                    mmax, interval, protocolperiod, protocolsource, fpuser = genmodules.ctrlaperiodic(xmlroot, idf,
                                                                                              childrenlist, label)
                    if listpprovin[0]['RCMoperationKind'] == 'sporadic':
                        activation = ET.SubElement(controller, 'activation', type='aperiodic')
                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    # Create child parameters for the activation tag

                    ET.SubElement(activation, 'parameter', name='m_max', value=str(mmax))
                    ET.SubElement(activation, 'parameter', name='interval', value=str(interval))
                    ET.SubElement(activation, 'parameter', name='protocol_period', value=str(protocolperiod))
                    ET.SubElement(activation, 'parameter', name='protocol_source', value=str(protocolsource))

                    # Create port child in controller of type output to the associated process
                    port = ET.SubElement(controller, 'port', type='out_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in controller of type input from the associated process
                    port = ET.SubElement(controller, 'port', type='in_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')

                    # Append aperiodic controller to xml file
                    app.append(controller)

                    # Control channels between aperiodic controller and aperiodic process
                    com3 = ET.Comment(
                        "Control-channel from aperiodic process " + item.get(
                            'label').lower() + ' to aperiodic controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment(
                        "Control-channel from aperiodic controller to aperiodic process " + item.get('label').lower())
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)
                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################

                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])


                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                # elif item.get("TASTE_IV_Properties-Fpriority") is None:
                #     print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                #     logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                #     exit(-1)

                elif float(listpprovin[0]['Deadline']) < float(interval*1000): # the Deadline of the sporadic process is lower than the period of the user in ms
                    print("\n---ERROR! The Deadline of the sporadic process '{}' is lower than the period of the user process! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("The Deadline of the sporadic process '{}' is lower than the period of the user process".format(item.get("label")))
                    exit(-1)

                else:
                    fspriority = 0
                    datavalues = [float(interval*1000),float(listpprovin[0]['Deadline'])-float(interval*1000),
                              float(listpprovin[0]['Compute_Execution_Time']),int(fspriority),
                              'C', 'imperiodic', mmax]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []


            # =====================
            # SPORADIC PROCESS D
            # =====================

            elif value == 'sporadic_process_D':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the sporadic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the sporadic process = ' + item.get('label'))
                com1 = ET.Comment("Sporadic Process: " + item.get('label').lower())
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='D')
                app.append(process)
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    # create the associated controller
                    com2 = ET.Comment("Aperiodic Controller of aperiodic process: " + item.get('label').lower())
                    app.append(com2)
                    if 'Deadline' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline value\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline value')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))

                        # call the object id of the sporadic process function
                        sporprocid = item.get('id')

                        # Create child activation for the controller
                        # cross check validation for aperiodic provided interface

                    # call the ctrlaperiodic function for the definition of the activation

                    mmax, interval, protocolperiod, protocolsource, fpuser = genmodules.ctrlaperiodic(xmlroot, idf,
                                                                                              childrenlist, label)
                    if listpprovin[0]['RCMoperationKind'] == 'sporadic':
                        activation = ET.SubElement(controller, 'activation', type='aperiodic')
                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    # Create child parameters for the activation tag

                    ET.SubElement(activation, 'parameter', name='m_max', value=str(mmax))
                    ET.SubElement(activation, 'parameter', name='interval', value=str(interval))
                    ET.SubElement(activation, 'parameter', name='protocol_period', value=str(protocolperiod))
                    ET.SubElement(activation, 'parameter', name='protocol_source', value=str(protocolsource))

                    # Create port child in controller of type output to the associated process
                    port = ET.SubElement(controller, 'port', type='out_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in controller of type input from the associated process
                    port = ET.SubElement(controller, 'port', type='in_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')

                    # Append aperiodic controller to xml file
                    app.append(controller)

                    # Control channels between aperiodic controller and aperiodic process
                    com3 = ET.Comment(
                        "Control-channel from aperiodic process " + item.get(
                            'label').lower() + ' to aperiodic controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label') + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment(
                        "Control-channel from aperiodic controller to aperiodic process " + item.get('label'))
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)
                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################

                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])



                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                # elif item.get("TASTE_IV_Properties-Fpriority") is None:
                #     print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                #     logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                #     exit(-1)

                elif float(listpprovin[0]['Deadline']) < float(interval*1000): # the deadline of the sporadic process is lower than the period of the user in ms
                    print("\n---ERROR! The Deadline of the sporadic process '{}' is lower than the period of the user process! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("The Deadline of the sporadic process '{}' is lower than the period of the user process".format(item.get("label")))
                    exit(-1)

                else:
                    fspriority = 0
                    datavalues = [float(interval*1000),float(listpprovin[0]['Deadline'])-float(interval*1000),
                              float(listpprovin[0]['Compute_Execution_Time']),int(fspriority),
                              'D', 'imperiodic', mmax]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []


            # =====================
            # SPORADIC PROCESS E
            # =====================

            elif value == 'sporadic_process_E':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the sporadic process = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the sporadic process = ' + item.get('label'))
                com1 = ET.Comment("Sporadic Process: " + item.get('label').lower())
                app.append(com1)
                process = ET.Element("process", name=item.get('label').lower(), criticality='E')
                app.append(process)
                ET.SubElement(process, 'source', location=str.lower(item.get('label').lower()) + '.c')
                # call the provided interface function in order to determine
                # the port-types of the process that are associated with the
                # controller, define the associated controller, its port-types and characteristics.

                # A periodic process has only a single provided interface therefore
                # check if it is identified that only one provided interface corresponds to this
                # function unit that will be translated to a periodic process
                if len(listpprovin) == 1:
                    # Create port child in process of type input from the associated periodic controller
                    port = ET.SubElement(process, 'port', type='in_event',
                                         name='p_PROC_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in process type output to the associated periodic controller
                    port = ET.SubElement(process, 'port', type='out_event',
                                         name='p_PROC_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')
                    # call required interfaces
                    reqiflist = genmodules.findRequiredIfs(xmlroot)
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # print("\n REQIFLIST {}".format(reqiflist))
                    # find if the ids of the FEATURES of the Function Units that are included as values in the required interfaces
                    # dictionaries
                    # initialize the list of dictionaries for storing required interfaces that are children (features) of the

                    # call the channel function
                    # call findChannel to identify the port types and the associated interfaces
                    listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,childrenlist,providedIFlist)
                    # call findConnections in order to retrieve from the connorderedpairs the port-type
                    connections, connorderedpairs = genmodules.findConnections(xmlroot)
                    index = 3
                    for chanitem in listchannels:
                        types = dictchanports[chanitem['id']]

                        for childFeature in childcomponentsfeat:

                            for itemReqIf in reqiflist:


                                # find if the childFeature is included as a value in the dictionary itemProvIf

                                if childFeature in itemReqIf['id']:#values():
                                    # if yes find the

                                    idrif = itemReqIf['id']
                                    labelrif = itemReqIf['label']
                                    # call the channel function
                                    # call findChannel to identify the port types and the associated interfaces
                                    # listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot,
                                    #                                                                childrenlist,
                                    #                                                                providedIFlist)
                                    #
                                    # # call findConnections in order to retrieve from the connorderedpairs the port-type
                                    # connections, connorderedpairs = genmodules.findConnections(xmlroot)

                                    #for chanitem in listchannels:


                                    # reverse dictionary of connorderedpairs
                                    invconnorderedpairs = {v: k for k, v in connorderedpairs.items()}
                                    typeid = invconnorderedpairs[idrif]

                                    if idrif in connorderedpairs.values() and typeid in types.keys():

                                        port_type = types[typeid]

                                        if port_type == 'in_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='out_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_write')
                                            index += 1

                                        elif port_type == 'out_data':
                                            process.insert(index, ET.Comment("data port connected to channel " + chanitem.get('label').lower()))
                                            ET.SubElement(process, 'port', type='in_data',name='p_PROC_' + labelrif)  # 'p_PROC_' + chanitem.get('label') + '_read')
                                            index += 1
                                    else:
                                        pass
                                    index +=index

                    # create the associated controller
                    com2 = ET.Comment("Aperiodic Controller of aperiodic process: " + item.get('label').lower())
                    app.append(com2)
                    if 'Deadline' not in listpprovin[0]:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject has not a valid Deadline value\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug('The identified interface ProvidedInterfaceObject has not a valid Deadline value')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    else:
                        # Note: the Deadline is in msec in interfaceview.xml and it is necessary to convert it to seconds
                        controller = ET.Element("controller", name='C' + item.get('label').lower(),
                                                deadline=str(float(listpprovin[0]['Deadline']) / 1000))

                        # call the object id of the sporadic process function
                        sporprocid = item.get('id')

                        # Create child activation for the controller
                        # cross check validation for aperiodic provided interface

                    # call the ctrlaperiodic function for the definition of the activation

                    mmax, interval, protocolperiod, protocolsource, fpuser = genmodules.ctrlaperiodic(xmlroot, idf,
                                                                                              childrenlist, label)
                    if listpprovin[0]['RCMoperationKind'] == 'sporadic':
                        activation = ET.SubElement(controller, 'activation', type='aperiodic')
                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic\n')
                        print('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=aperiodic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + listpprovin[0]['id'])
                        exit(-1)
                    # Create child parameters for the activation tag

                    ET.SubElement(activation, 'parameter', name='m_max', value=str(mmax))
                    ET.SubElement(activation, 'parameter', name='interval', value=str(interval))
                    ET.SubElement(activation, 'parameter', name='protocol_period', value=str(protocolperiod))
                    ET.SubElement(activation, 'parameter', name='protocol_source', value=str(protocolsource))

                    # Create port child in controller of type output to the associated process
                    port = ET.SubElement(controller, 'port', type='out_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_start')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='start')
                    # Create port child in controller of type input from the associated process
                    port = ET.SubElement(controller, 'port', type='in_event',
                                         name='p_CTRL_' + item.get('label').lower() + '_finish')
                    # Create event child in port tag
                    ET.SubElement(port, 'event', name='finish')

                    # Append aperiodic controller to xml file
                    app.append(controller)

                    # Control channels between aperiodic controller and aperiodic process
                    com3 = ET.Comment(
                        "Control-channel from aperiodic process " + item.get(
                            'label').lower() + ' to aperiodic controller')
                    app.append(com3)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_p2c')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_finish')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label') + '_finish')
                    app.append(control_channel)
                    com4 = ET.Comment(
                        "Control-channel from aperiodic controller to aperiodic process " + item.get('label'))
                    app.append(com4)
                    control_channel = ET.Element("control_channel", name='Ctrl_' + item.get('label').lower() + '_c2p')
                    ET.SubElement(control_channel, 'port', name='p_CTRL_' + item.get('label').lower() + '_start')
                    ET.SubElement(control_channel, 'port', name='p_PROC_' + item.get('label').lower() + '_start')
                    app.append(control_channel)
                else:
                    print(
                        '\nError! The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    print('The id of the FunctionObject is:' + item.get('id'))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug(
                        'The generator has not identified provided interfaces (or the number of them is greater than 1) for the corresponding periodic process')
                    logging.debug('The id of the FunctionObject is:' + item.get('id'))
                    exit(-1)

                #####################
                #  c 2 c generation
                #####################

                # create the c code files based on path and pathforffpn

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cproccodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                        listpprovin[0]['label'])



                if 'Compute_Execution_Time' not in listpprovin[0]:
                    print("\n---ERROR! WCET is not set for the process '{}' and interface '{}'! \n".format(item.get("label"),listpprovin[0]['id']))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("WCET is not set for the process '{}' and interface '{}'!".format(item.get("label"),listpprovin[0]['id']))
                    exit(-1)
                # elif item.get("TASTE_IV_Properties-Fpriority") is None:
                #     print("\n---ERROR! Fpriority is not defined for the process '{}'! \n".format(item.get("label")))
                #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                #     logging.debug("Fpriority is not defined for the process '{}'".format(item.get("label")))
                #     exit(-1)

                elif float(listpprovin[0]['Deadline']) < float(interval*1000): # the Deadline of the sporadic process is lower than the period of the user in ms
                    print("\n---ERROR! The Deadline of the sporadic process '{}' is lower than the period of the user process! \n".format(item.get("label")))
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug("The Deadline of the sporadic process '{}' is lower than the period of the user process".format(item.get("label")))
                    exit(-1)

                else:
                    fspriority = 0
                    datavalues = [float(interval*1000),float(listpprovin[0]['Deadline'])-float(interval*1000),
                              float(listpprovin[0]['Compute_Execution_Time']),int(fspriority),
                              'E', 'imperiodic', mmax]
                #update dictionary pervals
                pervals[item.get("label").lower()]=dict(zip(charkeys, datavalues))
                datavalues = []



            # =====================
            # BLACKBOARD CHANNEL
            # =====================

            elif value == 'blackboard':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the blackboard = {}'.format(item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the blackboard = ' + item.get('label'))
                com1 = ET.Comment("Blackboard data-channel: " + item.get('label').lower())
                app.append(com1)
                if item.get('TASTE_IV_Properties-DataChannelSize') == None:
                    print("\n --- ERROR: the backboard channel with name {} has not a defined size or length--\n".format(item.get('label')))
                    logging.error("the Blackboard channel with name {} has not a defined size or length--\n".format(item.get('label')))
                    exit(-1)
                data_channel = ET.Element("data_channel", name=item.get('label').lower(),
                                          type=item.get('TASTE_IV_Properties-FPPNClass'),
                                          size=item.get('TASTE_IV_Properties-DataChannelSize'), lifespan='32768')

                # call findChannel to identify the port types and the associated interfaces
                listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot, childrenlist,
                                                                               providedIFlist)

                #print('The identified channels are: {}\n'.format(listchannels))
                logging.info('The identified channels are: {}'.format(listchannels))
                for channelitem in listchannels:
                    # get the label of the channel item
                    chlabel = channelitem['label']
                    # get the children of the channelitem by using the label
                    chitemchildren = childrenlist[chlabel]['FEATURE']

                    logging.info('The channel with label "{}" has the following children'.format(chlabel))
                    logging.info('... Children ids: {}'.format(chitemchildren))

                types = dictchanports[idf]
                listofprovin = chprovin[item['id']]
                for pintitem in listofprovin:
                    port_type = types[pintitem['id']]
                    # print("\nport_type{}".format(port_type))
                    # print("\nlistofprovin {}".format(listofprovin))
                    labelpint = pintitem['label']
                    if port_type == 'in_data':
                        # item.get('label') + '_write',
                        ET.SubElement(data_channel, 'port', name='p_CHAN_' + labelpint,
                                      type=port_type)
                    elif port_type == 'out_data':
                        # item.get('label') + '_read'
                        ET.SubElement(data_channel, 'port', name='p_CHAN_' + labelpint,
                                      type=port_type)
                    else:
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        print(
                            '\nError! The generator has  provided interfaces for the channel without in or out type')
                        print('The id of the FunctionObject is:' + item.get('id'))
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.debug('The xml IV has  provided interfaces for the channel without in or out type')
                        logging.debug('The id of the FunctionObject is:' + item.get('id'))
                        exit(-1)
                app.append(data_channel)

            # =====================
            # SPORADIC PROTOCOL
            # =====================

            elif value == 'sporadic_protocol':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the sporadic protocol = {}'.format(
                        item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the sporadic protocol = ' + item.get('label'))
                com1 = ET.Comment("There is a Sporadic Protocol with label: " + item.get('label').lower())
                app.append(com1)

                #####################s
                #  c 2 c generation
                #####################

                # Find all the required interfaces
                riforspor = genmodules.findRequiredIfs(xmlroot)
                rifcounter = 0
                for chFtr in childcomponentsfeat:
                    for itmRI in riforspor:
                        # find if the childFeature is included as a value in the dictionary itemProvIf
                        if chFtr in itmRI.values():
                            # if yes find the
                            idrf = itmRI['id']
                            rilabel = itmRI['label']
                            rifcounter = + 1
                        else:
                            pass

                if rifcounter > 1 or rifcounter == 0:
                    logging.error('The Sporadic Protocol with id = ' + item.get('id') + ' and label = ' + item.get(
                        'label') + ' has either more than one Required Interfaces or zero')
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    print('\nError! The Sporadic Protocol with id = ' + item.get('id') + ' and label = ' + item.get(
                        'label') + ' has either more than one Required Interfaces or zero')
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    exit(-1)
                else:
                    pass

                # Find all the ordered connections for the connected interfaces
                connecs, conordpairs = genmodules.findConnections(xmlroot)

                # check if the required interface of the sporadic protocol exists in the ordered pairs of connections
                if idrf in conordpairs.values():
                    # find the associated key from the dictionary of the ordered pairs of connections
                    rikey = [rikey for rikey, value in conordpairs.iteritems() if value == idrf][0]
                    # check if the rikey is included in the childrenlist and retrieve the associated process
                    for itchild in childrenlist:
                        if rikey in childrenlist[itchild]['FEATURE']:
                            apeprocesslabel = itchild
                        else:
                            pass
                elif idrf in conordpairs.keys():
                    rikey = conordpairs[idrf]
                    # check if the rikey is included in the childrenlist and retrieve the associated process
                    for itchild in childrenlist:
                        if rikey in childrenlist[itchild]['FEATURE']:
                            apeprocesslabel = itchild
                        else:
                            pass
                else:
                    logging.error('The Sporadic Protocol with id = ' + item.get('id') + ' and label = ' + item.get(
                        'label') + ' is not connected with a sporadic process')
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    print('\nError! The Sporadic Protocol with id = ' + item.get('id') + ' and label = ' + item.get(
                        'label') + ' is not connected with a sporadic process')
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    exit(-1)

                # check if the associated with the sporadic protocol process with label apeprocesslabel is of type sporadic
                countsporpr = 0
                for itf in functionlist:
                    if apeprocesslabel in itf['label'] and 'sporadic_process' in itf['TASTE_IV_Properties-FPPNClass']:
                        countsporpr += 1
                    else:
                        pass
                if countsporpr != 1:
                    logging.error('The Sporadic Protocol with id = ' + item.get('id') + ' and label = ' + item.get(
                        'label') + ' is connected with a process with label = ' + apeprocesslabel + ' that is not of type sporadic')
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    print('\nError! The Sporadic Protocol with id = ' + item.get('id') + ' and label = ' + item.get(
                        'label') + ' is connected with a process with label = ' + apeprocesslabel + ' that is not of type sporadic')
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    exit(-1)
                else:
                    pass

                tfilec = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.c', 'r')
                tfileh = open(path + '/' + item.get('label').lower() + '/' + item.get('label').lower() + '.h', 'r')
                dolcfc = open(pathforfppnsrc + '/' + item.get('label').lower() + '.c', 'w+')
                dolcfh = open(pathforfppnsrc + '/' + item.get('label').lower() + '.h', 'w')

                genmodules.cprotocodegen(tfilec, tfileh, item.get('label').lower(), dolcfc, dolcfh,
                                         listpprovin[0]['label'],
                                         apeprocesslabel, rilabel)

            # =====================
            # MAILBOX CHANNEL
            # =====================

            elif value == 'mailbox':
                print(
                    '\nTaste-IV 2 FPPN-DOLC Generator is processing the mailbox = {}'.format(item.get('label')))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator is processing the mailbox = ' + item.get('label'))
                com1 = ET.Comment("Mailbox data-channel: " + item.get('label').lower())
                app.append(com1)
                if item.get('TASTE_IV_Properties-DataChannelSize') == None or item.get('TASTE_IV_Properties-DataChannelLength') == None :
                    print("\n --- ERROR: the mailbox channel with name {} has not a defined size or length--\n".format(item.get('label')))
                    logging.error("the mailbox channel with name {} has not a defined size or length--\n".format(item.get('label')))
                    exit(-1)
                data_channel = ET.Element("data_channel", type=item.get('TASTE_IV_Properties-FPPNClass'),
                                          name=item.get('label').lower(),
                                          size=item.get('TASTE_IV_Properties-DataChannelSize'),
                                          length=item.get('TASTE_IV_Properties-DataChannelLength'))
                # call findChannel to identify the port types and the associated interfaces
                listchannels, dictchanports, chprovin = genmodules.findChannel(functionlist, xmlroot, childrenlist,
                                                                               providedIFlist)

                logging.info('The identified channels are: {}'.format(listchannels))
                for channelitem in listchannels:
                    # get the label of the channel item
                    chlabel = channelitem['label']
                    # get the children of the channelitem by using the label
                    chitemchildren = childrenlist[chlabel]['FEATURE']
                    print(
                    '\n --- The channel with label "{}" has the children with ids {}\n'.format(chlabel, chitemchildren))
                    logging.info('The channel with label "{}" has the following children'.format(chlabel))
                    logging.info('... Children ids: {}'.format(chitemchildren))

                types = dictchanports[idf]
                listofprovin = chprovin[item['id']]
                for pintitem in listofprovin:
                    port_type = types[pintitem['id']]
                    # print("\nport_type{}".format(port_type))
                    # print("\nlistpprovin {}".format(listofprovin))
                    labelpint = pintitem['label']
                    if port_type == 'in_data':
                        ET.SubElement(data_channel, 'port', name='p_CHAN_' + labelpint,
                                      type=port_type)
                    elif port_type == 'out_data':
                        ET.SubElement(data_channel, 'port', name='p_CHAN_' + labelpint,
                                      type=port_type)
                    else:
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        print(
                            '\nError! The generator The generator has  provided interfaces for the channel without in or out type')
                        print('The id of the FunctionObject is:' + item.get('id'))
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.debug('The xml IV has  provided interfaces for the channel without in or out type')
                        logging.debug('The id of the FunctionObject is:' + item.get('id'))
                        exit(-1)
                app.append(data_channel)
                # source = ET.SubElement(data_channel, 'source', location=item.get('label') + '.c')
            else:
                print('\nError! The Taste Function attribute FPPNClass is not defined\n')
                print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                logging.debug('The Taste Function attribute FPPNClass is not defined')
                logging.debug('The id of the Function Object in TASTE xml is:' + idf)
                exit(0)
        else:
            print('\nError! The Taste Function attribute Source Language = C is not defined\n')
            print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
            logging.debug('The Taste Function attribute Source Language = C is not defined')
            logging.debug('The id of the Function Object in TASTE xml is:' + idf)
            exit(-1)

    # create process to channel connections
    # call channel function
    listchannels, dictchanports, pprovin = genmodules.findChannel(functionlist, xmlroot, childrenlist, providedIFlist)
    # call findConnections in order to retrieve from the connorderedpairs the port-type
    connections, connorderedpairs = genmodules.findConnections(xmlroot)

    # for each channel
    for chanitem in listchannels:
        # find the port-types of each channels provided-interfaces
        porttypes = dictchanports[chanitem['id']]
        # find the direction for each provided interface
        listofprovin = pprovin[chanitem['id']]
        for provinitem in listofprovin:
            # call find required interfaces
            rinterf = genmodules.findRequiredIfs(xmlroot)
            for itemrinterf in rinterf:
                if connorderedpairs[provinitem['id']] == itemrinterf['id']:
                    labelreqif = itemrinterf['label']
                else:
                    pass
            direction = porttypes[provinitem['id']]
            if direction == 'in_data':
                comdatacon = ET.Comment(
                    "Data connection between channel with provided interface " + provinitem['label'] + " & id = " +
                    provinitem['id'] + "  and process"
                                       " with required interface " + labelreqif + " and id = " + connorderedpairs[
                        provinitem['id']])
                print("\n label of req interface: {}".format(labelreqif))
                app.append(comdatacon)
                connection = ET.Element("connection", name=chanitem.get('label').lower() + '_channel_IN')
                # ET.SubElement(connection, 'port', name='p_PROC_' + chanitem.get('label') + '_write')
                ET.SubElement(connection, 'port', name='p_PROC_' + labelreqif)
                # ET.SubElement(connection, 'port', name='p_CHAN_' + chanitem.get('label') + '_write')
                ET.SubElement(connection, 'port', name='p_CHAN_' + labelreqif)
                app.append(connection)
                print('\n ---Data connection between a process with required interface {} and a channel with label {} '
                      'and provided interface with id {} has been placed\n'.format(
                    connorderedpairs[provinitem['id']],
                    chanitem.get('label'),
                    provinitem['id']))
                logging.info('Data connection between a process with required interface {} and a channel with label {} '
                             'and provided interface with id {} has been placed '.format(
                    connorderedpairs[provinitem['id']],
                    chanitem.get('label'),
                    provinitem['id']))
            elif direction == 'out_data':
                comdatacon = ET.Comment(
                    "Data connection between channel with provided interface " + provinitem['label'] + " & id = " +
                    provinitem['id'] + "  and process"
                                       " with required interface " + labelreqif + " and id = " + connorderedpairs[
                        provinitem['id']])
                print("\n label of req interface: {}".format(labelreqif))
                app.append(comdatacon)
                connection = ET.Element("connection", name=chanitem.get('label').lower() + '_channel_OUT')
                # ET.SubElement(connection, 'port', name='p_PROC_' + chanitem.get('label') + '_read')
                ET.SubElement(connection, 'port', name='p_PROC_' + labelreqif)
                # ET.SubElement(connection, 'port', name='p_CHAN_' + chanitem.get('label') + '_read')
                ET.SubElement(connection, 'port', name='p_CHAN_' + labelreqif)
                app.append(connection)
                print('\n ---Data connection between a process with required interface {} and a channel with label {} '
                      'and provided interface with id {} has been placed\n'.format(
                    connorderedpairs[provinitem['id']],
                    chanitem.get('label'),
                    provinitem['id']))
                logging.info('Data connection between a process with required interface {} and a channel with label {} '
                             'and provided interface with id {} has been placed'.format(
                    connorderedpairs[provinitem['id']],
                    chanitem.get('label'),
                    provinitem['id']))
            else:
                logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                print('\nError! The xml IV has  provided interfaces for the channel without in or out type')
                print('The id of the FunctionObject is:' + chanitem.get('id'))
                print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                logging.debug('The generator has  provided interfaces for the channel without in or out type')
                logging.debug('The id of the FunctionObject is:' + chanitem.get('id'))
                exit(-1)



    # create global requirements and precedence chain
    commentglobal = ET.Comment("Global requirements")
    app.append(commentglobal)
    globaldata = ET.Element("global", name='globalCycle')
    #   call precedence function
    prec = genmodules.precedence(xmlroot)
    # call channel function
    listchannels, dictchanports, channelprovint = genmodules.findChannel(functionlist, xmlroot, childrenlist,
                                                                         providedIFlist)

    #print('listachn = {}\n'.format(listchannels))
    #print('chanports = {}\n'.format(dictchanports))

    chains = [] # this list is used to remove dublicates when chain strings are created!

    for channel in listchannels:
        dictpair = prec[channel['id']]
        #print('dictparir = {}\n'.format(dictpair))
        value1 = dictpair.values()[0]
        value2 = dictpair.values()[1]

        # the lower the functional priority the higher the precedence
        if value1[1] > value2[1]:
            precstr = '' + value2[0] + ',' + value1[0] + ''
            #ET.SubElement(globaldata, 'precedence', chain=precstr.lower())
            chains.append(precstr.lower())
            print('\n --- The precedence pair for the channel with label {} is: {} '
                  .format(channel['label'], precstr))
            logging.info('The precedence pair for the channel with label {} is: {} '
                         .format(channel['label'], precstr))
        elif value1[1] < value2[1]:
            precstr = '' + value1[0] + ',' + value2[0] + ''
            #ET.SubElement(globaldata, 'precedence', chain=precstr.lower())
            chains.append(precstr.lower())
            print('\n --- The precedence pair for the channel with label {} is: {} '
                  .format(channel['label'], precstr))
            logging.info('The precedence pair for the channel with label {} is: {} '
                         .format(channel['label'], precstr))
        else:
            print(
                '\nError! The channel with label {} and id {} is connected to processes that have equal functional Priority'.format(
                    channel['label'], channel['id']))
            print('The id of the FunctionObject is:' + chanitem.get('id'))
            print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
            logging.debug(
                'The channel with label {} and id {} is connected to processes that have equal functional Priority'.format(
                    channel['label'], channel['id']))
            exit(-1)

    print('\n\nImportant NOTE: In case that generator identifies dublicates in precendece chains, it removes them!\n')
    print('This is due to the fact that the DOLC2BIP raises error when identifies dublicates in precedence chains even if they refer to different channels\n')
    logging.info('Important NOTE: In case that generator identifies dublicates in precendece chains, it removes them!'
                 'This is due to the fact that the DOLC2BIP raises error when identifies dublicates in precedence chains even if they refer to different channels')

    cleanchains = [] # list that contains data from chains list when duplicates are removed

    #remove dubplicates from the chains list and place them in cleanchains list
    [cleanchains.append(xzy) for xzy in chains if xzy not in cleanchains]
    for  chainstrings in cleanchains:
        ET.SubElement(globaldata, 'precedence', chain=chainstrings)
        print('\n --- The precedence pair = {} has been placed!'.format(chainstrings))
        logging.info('The precedence pair = {} has been placed!'.format(chainstrings))





    app.append(globaldata)

    # ------------------
    # Create the file temp_asn1.h
    # ------------------

    # IMPORTANT! Check if the temp_asn1.h exists.
    # If exists do not do anything else do create it

    # pathforfppn temp_asn1.h
    dolctempasn = pathforfppnsrc + '/temp_asn1.h'
    if os.path.isfile(dolctempasn):
        pass
    else:
        dolctempasnf = open(dolctempasn, 'w')
        genmodules.tempasnone(dolctempasnf)

    return app, pervals, cleanchains

