##########################################################################################################
# Title: Taste-IV 2 FPPN-DOLC modules
# Author: Fotios Gioulekas
# Date: 2015/10/12
# Project:MoSaTT-CMP
# Name: GenModules
# Version: 8.0
# Abstract:
# This is a python module that contains the functions which are called by the TASTE-IV 2 FPPN-DOLC generator
############################################################################################################
import logging

import math
############################################################################################################
# function:functionObject
# This function identifies the children element with tag <FunctionObject>, their
# attributes, and subchildren elements with tags FEATURE and CHILD  and the associated reference attributes
# input: XMLroot element
# outputs: (1)numoffunctionobjects: number of FunctionObject elements contained in the TASTE-IV xml file
#          (2)fattlist:list of function object attributes
############################################################################################################
def functionObject(xmlroot):
    functionobjects = xmlroot.findall('FunctionObject')
    # Debugging purposes
    # print(functionobjects)

    # Number of elements tagged FunctionObjects in the TASTE-IV xml file
    numoffunctionobjects = len(functionobjects)

    # Retrieve the attributes for each one of the function objects and stored them in the function list of dictionaries
    fattlist = []  # functionObject list of attributes

    for fobject in functionobjects:
        # print(fobject.attrib) # Debugging purposes
        fattlist.append(fobject.attrib)  # append each dictionary of function object attributes to the flist

    return numoffunctionobjects, fattlist

############################################################################################################
# functioname: functionchildren
# inputs: (1) XMLroot element
#         (2) fattlist:list of function object attributes
# output: fchildren: dictionary of children dictionaries
############################################################################################################
def functionchildren(xmlroot, fattlist):
    functionobjects = xmlroot.findall('FunctionObject')
    # Retrieve the children of each one of the FunctionObject elements and store them in the fchildren dictionary
    # dictionaries

    fchildren = {}  # format = {label:'{{CHILD:[objx, objy, ..],numofchildref:lenoflistchild['CHILD')}',
    # {FEATURE:objx, objy, ..],numofFeatureef:lenoflistfeature['FEATURE')}}


    # for folist in fattlist:
    for fobject in functionobjects:
        childdict = {}
        featuredict = {}
        # lists
        childlist = []
        featurelist = []
        labelkey = fobject.attrib.get('label')  # , "empty")
        # print("/n LABEL KEY:{}".format(labelkey))
        if len(labelkey) == 0:
            print("\nError: Function in Taste without Label\n")
            print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
            logging.error('The FunctionObject with id {} has not a label'.format(fobject.attrib.get('id')))
            exit(-1)
        else:
            # Debugging purposes
            # print(labelkey)
            for fchild in fobject.getchildren():

                # GRAPHIC child is not required for the processing
                if fchild.tag != 'GRAPHIC':
                    x = fchild.tag
                    if x == 'CHILD':
                        # print('\n')
                        # print(x) # Debugging purposes
                        y = fchild.attrib
                        # retrieve the reference value from the y dictionary
                        q = y['reference']
                        childlist.append(q)
                        # print('\n')
                        # print('childlist')
                        # print(childlist)
                        # print(q)
                        # print(y) # Debugging purposes
                        # print('\n')
                    elif x == 'FEATURE':
                        # print('\n')
                        # print(x) # Debugging purposes
                        yy = fchild.attrib
                        # retrieve the reference value from the y dictionary
                        qq = yy['reference']
                        featurelist.append(qq)
                        # print('\n')
                        # print(yy) # Debugging purposes
                        # print('\n')
                        # print(qq) # Debugging purposes
                    else:
                        pass

        # childdict=dict.fromkeys('CHILD', childlist)
        childdict.update({'numofCHILDelements': len(childlist)})
        childdict['CHILD'] = childlist
        # childdict.Add({'CHILD':childlist})

        # childdict['numofCHILDelements']=len(childlist)
        featuredict.update({'numofFEATUREelements': len(featurelist)})
        featuredict['FEATURE'] = featurelist

        # merge childdict with featuredict
        childdict.update(featuredict)
        fchildren[labelkey] = childdict

    return fchildren

############################################################################################################
# the findProvidedIfs function identifies the provided interfaces in the InterfaceView.xml file
# and returns a list with their attributes
# input: xmlroot element
# output: provIFlist (list of provided interfaces with their attributes)
############################################################################################################
def findProvidedIfs(xmlroot):
    # find for each process its provided interface, add the port types
    # and create its associated controller with their port-types
    provIntObj = xmlroot.findall('ProvidedInterfaceObject')

    # Retrieve the attributes for each one of the provided interfaces objects and stored them in
    # the provided ifs list of dictionaries
    provIFlist = []  # provided interface list of attributes
    for ifobject in provIntObj:
        # print(ifobject.attrib) # Debugging purposes
        provIFlist.append(ifobject.attrib)
    return provIFlist

############################################################################################################
# the findRequiredIfs function identifies the required interfaces in the InterfaceView.xml file
# and returns a list with their attributes
# input: (1) xmlroot element
# output: (1) reqIFlist (list of required interfaces with their attributes)
############################################################################################################

def findRequiredIfs(xmlroot):
    # find for each process its provided interface, add the port types
    # and create its associated controller with their port-types
    reqIntObj = xmlroot.findall('RequiredInterfaceObject')
    # Retrieve the attributes for each one of the required interfaces objects and stored
    # them in the provided ifs list of dictionaries
    reqIFlist = []  # provided interface list of attributes
    for ifobject in reqIntObj:
        # print(ifobject.attrib) # Debugging purposes
        reqIFlist.append(ifobject.attrib)
    return reqIFlist

#######################################################################################
# the findConnections function identifies the GenericConnectionObject in IV.xml file
# and returns a dictionary with the name of the connection and their children with their
# associated attributes
# input: (1) xmlroot: the xml root element
# output:(1) connections
#        (2) connOrderedPairs
#######################################################################################

def findConnections(xmlroot):
    # Retrieve the attributes for each one of the required interfaces objects and stored
    genConnObj = xmlroot.findall('GenericConnectionObject')
    connections = {}  # format = {label{CHILD:{reference:id,..}}, label2:{CHILD:{reference:id,..}}},
    connorderedpairs = {}  # format = {child_id:child_id, child_id:child_id,....} matching process id with channel id
    listcon = []  # list with connection ids
    for itemCon in genConnObj:
        labelcon = itemCon.attrib.get('label')
        if len(labelcon) == 0:
            print("\nError: GenericConnectionObject Function in Taste without Label\n")
            print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
            logging.error('The GenericConnectionObject with id {} has not a label'.format(genConnObj.attrib.get('id')))
            exit(-1)
            # Debugging purposes
            # print(labelcon)
        else:
            conChildDiction = {}  # attributes of each children and childrenname

            for conchild in itemCon.getchildren():

                # try to get the child name and its attributes and store them in the dictionary
                # GRAPHIC child is not required for the processing
                if conchild.tag != 'GRAPHIC':
                    childname = conchild.tag
                    conchild.attrib
                    conChildDiction.update({childname: conchild.attrib})
                    if len(itemCon.getchildren()) != 3:
                        print("\nError: GenericConnectionObject Function in Taste without 2 children\n")
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('The GenericConnectionObject with id {} has not 2 children objects'.format(
                            genConnObj.attrib.get('id')))
                        exit(-1)
                    else:
                        y = conchild.attrib.get('reference')
                        # print(y)
                        listcon.append(y)
                        # print("\n listcon {}".format(listcon))

                else:
                    pass

        connections.update({labelcon: conChildDiction})

    connorderedpairs = dict(zip(listcon[::2], listcon[1::2]))

    # debug purposes
    #print("\n CONNECTIONS = {}".format(connections))
    return connections, connorderedpairs

################################################################################################
# the findChannel function identifies the channel functions and their children
# and returns a list with their attributes
# input: (1) fattlist: list of function object attributes
#        (2) xmlroot: the xml root element
#        (3 fchildren: dictionary of children dictionaries
#        (4)providedIFlist: it is the list with the provided interfaces
# output:(1)listchannels: a list with dictionaries of each channel attributes
#        (2)dictchanports: a dictionary that identifies the type of port per each channel interface
#        (3)dictprovin: a dictionary of the provided interfaces per each channel
##################################################################################################

def findChannel(fattlist, xmlroot, fchildren, providedIFlist):
    # list that returns only the identified channels with their attributes in dictionary
    listchannels = []
    # dict that contains dictionaries with channel id and the association of provided interface and type of port e.g. in or out
    dictchanports = {}  # e.g. {'obj154': {'obj169': 'in_data', 'obj206': 'out_data'}, ...}
    # dictionary with channel and its interfaces
    dictprovin = {}
    for item in fattlist:
        # print("\n item {}".format(item))
        # find the channel functions:
        fppnclass = item['TASTE_IV_Properties-FPPNClass']
        # print("\n fppnClass {}".format(fppnclass))
        if (fppnclass == 'blackboard') or (fppnclass == 'mailbox'):
            # if yes append them to the listchannels list
            listchannels.append(item)
            # else:
            #     print('\nError! No channel Function is defined in TASTE-IV\n')
            #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
            #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error! '
            #                   'No channel Function is defined in TASTE-IV!')
            #     exit(-1)
            # debugging purposes
    #print('\n The identified channels are: {}\n'.format(listchannels))
    logging.info('The identified channels are: {}'.format(listchannels))

    for channelitem in listchannels:
        # get the label of the channel item
        chlabel = channelitem['label']

        # get the children of the channelitem by using the label
        chitemchildren = fchildren[chlabel]['FEATURE']

        #print('len of children = {}\n'.format(len(chitemchildren)))
        # find the provided interfaces for this channel
        #print('\n providedIFlist: {}'.format(providedIFlist))

        # list that returns its associated provided interfaces
        listpprovin = []

        # dictionary with port types and provided interfaces ids
        porttypes = {}

        for childFeature in chitemchildren:


            for itemProvIf in providedIFlist:
                # find if the childFeature is included as a value in the dictionary itemProvIf
                if childFeature == itemProvIf['id']:
                    # if yes, append this to the listpprovin
                    #print('\n ChildFeature: {}'.format(childFeature))
                    #listpprovin.append(itemProvIf)

                    if itemProvIf['RCMoperationKind'] != 'protected':
                        print("\nError: the Channel with id '{}' has a provided interface with id '{}'"
                              " that is not of type protected\n".format(channelitem['id'], itemProvIf['id']))
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error("The Channel with id '{}' has a provided interface with id '{}' "
                                      "that is not of type protected".format(channelitem['id'], itemProvIf['id']))
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        exit(-1)
                    else:
                        #pass
                        listpprovin.append(itemProvIf)


                    operation = itemProvIf['operation']
                    # debugging purposes
                    #print('\n operation {}'.format(operation))
                    # build a string that will be used to identify the OperationObject in Taste IV xml file based on the id
                    opstr = ".//OperationObject[@id='" + operation + "']"
                    # debugging purposes
                    #print("\n text {}".format(opstr))
                    operationobject = xmlroot.findall(opstr)
                    # debugging purposes
                    #print('\n operation object {}'.format(operationobject))
                    # get children from this XML element
                    #print('LEN OF opobjects = {}\n'.format(len(operationobject)))
                    for opitem in operationobject:

                        opchildren = opitem.getchildren()
                        #print('opitem children = {}\n'.format(opchildren))
                        #print('LEN of opitems = {}\n'.format(len(opitem)))

                        countIN = 0  # counter for IN direction port types
                        countBoolean = 0  # In case of in direction count Boolean data types
                        countOUT = 0  # counter for OUT direction port types
                        countthreefeatures = 0  # It should be 4 or 5 in order to have a correct operation parameter
                        # for the provided interface of the channel with direction in
                        counttwofeatures = 0  # It should be 3 or 4 in order to have a correct operation parameter for
                        # the provided interface of the channel with direction out
                        if len(opchildren) == 3:
                            for opchild in opchildren:
                                # debugging purposes
                                opchildat = opchild.attrib
                                # print('\n opchildren {}'.format(opchildat))

                                # build a string that will be used to identify the ParameterObject in Taste IV xml file based on the id
                                parstr = ".//ParameterObject[@id='" + opchildat['reference'] + "']"
                                # debugging purposes
                                # print("\n text {}".format(parstr))
                                paramobject = xmlroot.findall(parstr)
                                # debugging purposes
                                # print('\n parameter object {}'.format(paramobject))
                                for itemparam in paramobject:
                                    paramattr = itemparam.attrib
                                    # debugging purposes
                                    # print('\n parameter item {}'.format(paramattr))
                                    if paramattr['direction'] == 'IN':
                                        countIN += 1
                                        if paramattr['parameterType'] == 'DataView::T_Boolean':
                                            countBoolean += 1
                                        elif len(paramattr['parameterType']) == 0:
                                            print(
                                                "\nError: the Channel with id '{}' has a provided interface with id '{}' that "
                                                " its parameter type for IN is not properly defined\n".format(
                                                channelitem['id'], itemProvIf['id']))
                                            print(
                                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                                            logging.error(
                                                "The Channel with id '{}' has a provided interface with id '{}' that "
                                                "its parameter type for IN is not properly defined".
                                                format(channelitem['id'], itemProvIf['id']))

                                    elif paramattr['direction'] == 'OUT' and paramattr[
                                        'parameterType'] == 'DataView::T_Boolean':
                                        countOUT += 1
                                    else:
                                        print(
                                            "\nError: the Channel with id '{}' has a provided interface with id '{}' that "
                                            " its operation does not exhibit IN or OUT directions\n".format(
                                                channelitem['id'], itemProvIf['id']))
                                        print(
                                            "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                                        logging.error(
                                            "The Channel with id '{}' has a provided interface with id '{}' that "
                                            "its operation does not exhibit IN or OUT directions".
                                            format(channelitem['id'], itemProvIf['id']))
                                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                                        exit(-1)

                            countthreefeatures = countIN + countBoolean + countOUT
                            if countthreefeatures == 4 or 5:
                                porttypes[itemProvIf['id']] = 'in_data'
                            else:
                                print("\nError: the Channel with id '{}' has a provided interface with id '{}' that "
                                      "its IN directions is not well defined\n".format(channelitem['id'],
                                                                                       itemProvIf['id']))
                                print(
                                    "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                                logging.error("The Channel with id '{}' has a provided interface with id '{}' that "
                                              "its IN directions is not well defined".
                                              format(channelitem['id'], itemProvIf['id']))
                                logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                                exit(-1)
                                # debugging purposes
                                # print("\n countthreefeatures: {}".format(countthreefeatures))
                        elif len(opchildren) == 2:
                            for opchild in opchildren:
                                # debugging purposes
                                opchildat = opchild.attrib
                                # print('\n opchildren {}'.format(opchildat))

                                # build a string that will be used to identify the ParameterObject in Taste IV xml file based on the id
                                parstr = ".//ParameterObject[@id='" + opchildat['reference'] + "']"
                                # debugging purposes
                                # print("\n text {}".format(parstr))
                                paramobject = xmlroot.findall(parstr)
                                # debugging purposes
                                # print('\n parameter object {}'.format(paramobject))
                                for itemparam in paramobject:
                                    paramattr = itemparam.attrib
                                    # debugging purposes
                                    # print('\n parameter item {}'.format(paramattr))
                                    if paramattr['direction'] == 'OUT':
                                        countOUT += 1
                                        if paramattr['parameterType'] == 'DataView::T_Boolean':
                                            countBoolean += 1
                                        elif len(paramattr['parameterType']) == 0:
                                            print(
                                                "\nError: the Channel with id '{}' has a provided interface with id '{}' that "
                                                " its parameter type for OUT is not properly defined\n".format(
                                                    channelitem['id'],
                                                    itemProvIf['id']))
                                            print(
                                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                                            logging.error(
                                                "The Channel with id '{}' has a provided interface with id '{}' that "
                                                "its parameter type for OUT is not properly defined".
                                                    format(channelitem['id'], itemProvIf['id']))
                                            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                                    else:
                                        print(
                                            "\nError: the Channel with id '{}' has a provided interface with id '{}' that "
                                            " its operation exhibits not OUT directions for both of its 2 operation parameters\n"
                                                .format(channelitem['id'], itemProvIf['id']))
                                        print(
                                            "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                                        logging.error(
                                            "The Channel with id '{}' has a provided interface with id '{}' that "
                                            "its operation exhibits not OUT directions for both of its 2 operation parameters".
                                                format(channelitem['id'], itemProvIf['id']))
                                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                                        exit(-1)

                            counttwofeatures = countBoolean + countOUT
                            # debugging purposes
                            # print("\n counttwofeatures: {}".format(counttwofeatures))
                            if counttwofeatures == 3 or 5:
                                porttypes[itemProvIf['id']] = 'out_data'
                            else:
                                print("\nError: the Channel with id '{}' has a provided interface with id '{}' that "
                                      "its OUT directions is not well defined\n".format(channelitem['id'],
                                                                                        itemProvIf['id']))
                                print(
                                    "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                                logging.error("The Channel with id '{}' has a provided interface with id '{}' that "
                                              "its OUT directions is not well defined".
                                              format(channelitem['id'], itemProvIf['id']))
                                logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                                exit(-1)
                        else:
                            print("\nError: the Channel with id '{}' has a provided interface with id '{}' that "
                                  " its operation does not exhibit 2 neither 3 parameters\n".format(channelitem['id'],
                                                                                                    itemProvIf['id']))
                            print(
                                "############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                            logging.error("The Channel with id '{}' has a provided interface with id '{}' that "
                                          "its operation does not exhibit 2 neither 3 parameters".
                                          format(channelitem['id'], itemProvIf['id']))
                            logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                            exit(-1)


                else:
                    pass




        dictchanports[channelitem['id']] = porttypes


        dictprovin[channelitem['id']] = listpprovin




        #dictprovin[channelitem['id']] = listpprovin
        #print('DICTPROVIN = {}\n'.format(dictprovin[channelitem['id']]))
        #dictchanports[channelitem['id']] = porttypes
    # debugging purposes
    # print('\n Channel: "{}" has the children: {}'.format(chlabel, chitemchildren))
    # logging.info('The channel with label "{}" has the following children'.format(chlabel))
    # logging.info('... Children ids: {}'.format(chitemchildren))

    #print("\n Channel dictprovin {}:".format(dictprovin))
    #print('PROV IFs = {}\n'.format(len(dictprovin)))
    #print("\n Channel list= {}".format(listchannels))
    #print('CHANNELS # = {}\n'.format(len(listchannels)))
    #print("\n Channel dictchanports {}:".format(dictchanports))
    #print('Number of ports = {}\n'.format(len(dictchanports)))

    return listchannels, dictchanports, dictprovin

##########################################################################################################
# the precedence function identifies the pair of processes that are connected to each channel and returns
# a dictionary with channel lables FPPN-Priority and process labels
# input: (1) xmlroot: the xml root element
#
# output:(1)predict: dictionary with channel and connected pair of processes with the following format
#          {channelid:{process1id:[ label, FPPN-priority],process2id:[label, FPPN-priority]},...}
#
############################################################################################################

def precedence(xmlroot):  # , listchannels, dictchanports, chpprovin):
    # call find functionobjects
    numOfFunctionObjects, functionlist = functionObject(xmlroot)
    # print("\nFuncctionlist{}".format(functionlist))
    # call find function children
    childrenlist = functionchildren(xmlroot, functionlist)
    # call find provided interfaces
    providedIFlist = findProvidedIfs(xmlroot)
    # call channel function
    listchannels, dictchanports, chpprovin = findChannel(functionlist, xmlroot, childrenlist, providedIFlist)
    # call findConnections in order to retrieve from the connorderedpairs the port-type
    connections, connorderedpairs = findConnections(xmlroot)


    #print('list channels = {}\n'.format(listchannels))
    #print('dict chanports = {}\n'.format(dictchanports))
    #for chphan in chpprovin:
        #print('channels provin = {}\n'.format(chpprovin[chphan] ))
        #print('number of provin = {}\n'.format(len(chpprovin[chphan])))
    #print('chp provin = {}\n'.format(chpprovin))

    precdict = {}  # dictionary with channel and connected pair of processes data


    # find the provided interfaces for each channel
    for channelitem in listchannels:
        process = {}
        provinterfaceslist = chpprovin[channelitem['id']]
        for provifch in provinterfaceslist:
            idofpif = provifch['id']
            # retrieve the id of the associated required if of the connected process
            procid = connorderedpairs[idofpif]
            processlist = []
            for itemfunction in functionlist:
                childdict = childrenlist[itemfunction['label']]
                if procid in childdict['FEATURE']:
                    processlist.append(itemfunction['label'])
                    processlist.append(itemfunction['TASTE_IV_Properties-Fpriority'])
                    process[itemfunction['id']] = processlist
                else:
                    pass
        precdict[channelitem['id']] = process

    return precdict

#######################################################################################################################################################
# the ctrlaperiodic function is called when a sporadic process is identified
# it returns the activation parameters used in the corresponding controller
# input: (1) xmlroot: the xml root element
#        (2) objectid: the object id of the sporadic process Taste function
#        (3) childrenlist: dictionary of children dictionaries
#        (4) label:  label of the sporadic protocol
#
# output:(1) mmax: math.ceil( Period of the user (ms) which is retrieved from its cyclic provided interface/min inter-arrival time (ms) from sporadic process sporadic provided interface)
#        (2) interval: period of the user in sec which is retrieved from its provided cyclic interface
#        (3) protocolperiod: it is the period which is retrieved from the cyclic provided interface of a the sporadic protocol function that is connected to the sporadic process
#        (4) protocolsource: it is the name of the c code file in lowercase that corresponds to the sporadic protocol
#        (5) fpuser: it is the functional priority of the user
############################################################################################################################################

def ctrlaperiodic(xmlroot, objectid, childrenlist, label):
    precdict = precedence(xmlroot)

    # print("\n predict: {}".format(precdict))
    for itemprecdict in precdict:
        # print("\n item : {}".format(itemprecdict))
        if objectid in precdict[itemprecdict]:
            yy = list(precdict[itemprecdict])
            for ii in yy:
                if ii != objectid:
                    # find and store the id of the associated user of the sporadic process
                    userid = ii
                    userlabel = precdict[itemprecdict][ii][0]
                    fpuser = precdict[itemprecdict][ii][1]
                    #print("\n fpuser =  {}\n".format(precdict[itemprecdict][ii][1]))
                else:
                    pass
        else:
            pass


    #print("\n userid =  {}".format(userid))
    #print("\n userlabel =  {}".format(userlabel))
    # find the provided interface of the process
    # call the findProvidedIfs function
    providedIFlist = findProvidedIfs(xmlroot)

    # find the children of the sporadic process function object with element tag FEATURE
    # It corresponds to the provided interface
    childcomponentsfeat = childrenlist[label]['FEATURE']

    # print(providedIFlist)
    counter = 0
    for childFeature in childcomponentsfeat:

        for itemProvIf in providedIFlist:
            # find if the childFeature is included as a value in the dictionary itemProvIf

            if childFeature in itemProvIf.values():
                if counter > 1:
                    print('\nError! The sporadic process has more than one provided interfaces\n')
                    print('The id of the sporadic process is:' + objectid)
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug('The sporadic process has more than one provided interfaces')
                    logging.debug('The id of the sporadic process is:' + objectid)
                    exit(-1)

                else:
                    if itemProvIf['RCMoperationKind'] == 'sporadic':
                        min_inter_arrival = itemProvIf['RCMperiod']
                        providsporadicprocess = itemProvIf['id']
                        # print("\n min: {}".format(min_inter_arrival))
                        counter = counter + 1

                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=sporadic\n')
                        print('The id of the ProvidedInterfaceObject is:' + childFeature)
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=sporadic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + childFeature)
                        logging.debug('The id of the sporadic process is:' + objectid)
                        exit(-1)

    # find the children of the periodic user process function object with element tag FEATURE
    # It corresponds to the provided interface
    childcomponentsfe = childrenlist[userlabel]['FEATURE']

    counteruser = 0
    for childFeat in childcomponentsfe:

        for itemProvI in providedIFlist:
            # find if the childFeature is included as a value in the dictionary itemProvI

            if childFeat in itemProvI.values():

                if counteruser > 1:
                    print('\nError! The periodic user process has more than one provided interfaces\n')
                    print('The id of the sporadic process is:' + userid)
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug('The sporadic process has more than one provided interfaces')
                    logging.debug('The id of the periodic user process is:' + userid)
                    exit(-1)

                else:
                    if itemProvI['RCMoperationKind'] == 'cyclic':
                        Tuser = itemProvI['RCMperiod']
                        # print("\n min: {}".format(min_inter_arrival))
                        counteruser = counteruser + 1

                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic\n')
                        print('The id of the ProvidedInterfaceObject is:' + childFeat)
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=cyclic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + childFeat)
                        logging.debug('The id of the Process is:' + userid)
                        exit(-1)

    mmax = int(math.ceil(float(Tuser) / float(min_inter_arrival)))

    interval = float(Tuser) / 1000

    connections, connorderedpairs = findConnections(xmlroot)
    # print("\n connections: {} and conorder: {}".format(connections, connorderedpairs))

    # find the required interface of the sporadic protocol that it connects to the provided interface of the sporadic process
    reqintidofsporadicprotocol = connorderedpairs[providsporadicprocess]


    # for each dictionary key which is the function object label check if the id of required interface of the sporadic protocol is included
    # in its children
    for chiditem in childrenlist:
        # print(chiditem)
        if reqintidofsporadicprotocol in childrenlist[chiditem]['FEATURE']:
            sporadic_protocol_label = chiditem
        else:
            pass
    if sporadic_protocol_label is None:
        print('\nError! The sporadic process is not connected to a sporadic protocol\n')
        print('The id of the sporadic process is:' + objectid)
        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
        logging.debug('The sporadic process is not connected to a sporadic protocol')
        logging.debug('The id of the sporadic process is:' + objectid)
        exit(-1)
    else:
        pass

    numoffunctionobjects, fattlist = functionObject(xmlroot)

    for fobject in fattlist:
        if fobject['label'] == sporadic_protocol_label and fobject[
            'TASTE_IV_Properties-FPPNClass'] == 'sporadic_protocol' and fobject['Source_Language'] == 'C':
            protocolsource = str.lower(sporadic_protocol_label) + '.c'
            idspor = fobject['id']
        else:
            pass
    if protocolsource is None:
        print('\nError! The sporadic process is not connected to a sporadic protocol\n')
        print('The id of the sporadic process is:' + objectid)
        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
        logging.debug('The sporadic process is not connected to a sporadic protocol')
        logging.debug('The id of the sporadic process is:' + objectid)
        exit(-1)
    else:
        pass


    # find the children of the sporadic protocol function object with element tag FEATURE
    # It corresponds to the provided interface
    childcomponents = childrenlist[sporadic_protocol_label]['FEATURE']

    # print(providedIFlist)
    counterspor = 0
    for childFe in childcomponents:

        for itemProv in providedIFlist:
            # find if the childFeature is included as a value in the dictionary itemProvIf

            if childFe in itemProv.values():
                if counterspor > 1:
                    print('\nError! The sporadic protocol has more than one provided interfaces\n')
                    print('The id of the sporadic protocol is:' + idspor)
                    print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                    logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                    logging.debug('The sporadic protocol has more than one provided interfaces')
                    logging.debug('The id of the sporadic protocol is:' + idspor)
                    exit(-1)

                else:
                    if itemProv['RCMoperationKind'] == 'cyclic':
                        protocolperiod = float(itemProv['RCMperiod']) / 1000
                        # print("\n min: {}".format(min_inter_arrival))
                        counterspor = counterspor + 1

                    else:
                        print(
                            '\nError! The identified interface ProvidedInterfaceObject of the sporadic protocol is not of RCMoperationKind=cyclic\n')
                        print('The id of the ProvidedInterfaceObject is:' + childFe)
                        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
                        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
                        logging.debug(
                            'The identified interface ProvidedInterfaceObject is not of RCMoperationKind=sporadic')
                        logging.debug('The id of the ProvidedInterfaceObject is:' + childFe)
                        logging.debug('The id of the sporadic protocol is:' + idspor)
                        exit(-1)

    return mmax, interval, protocolperiod, protocolsource, fpuser

################################################################################################################################
# the cproccodegen function is called when processes are present and it generates the associated  label.c and label.h files
# input: (1) tfilec: it is the taste process label.c file to be read
#        (2) tfileh: it is the taste process label.h file to be read
#        (3) processlabel: it is the process label from Taste xml file
#        (4) dolcfc: it is the dolc file label.c to be created an writen
#        (5) dolcfh: it is the dolc file label.h to be created and writen
#        (6) pilabel: it is the label of the provided interface of the process. Please note that currently each
#            process has a single provided interface (PI)
#################################################################################################################################

def cproccodegen(tfilec, tfileh, processlabel, dolcfc, dolcfh, pilabel):
    print("\n ------- Taste-IV 2 FPPN-DOLC Generator is processing the TASTE C files ---------\n")
    print("\n ------- files: " + processlabel + ".c " + processlabel + ".h  ---------\n")
    logging.info('Taste-IV 2 FPPN-DOLC Generator is processing the TASTE C files')
    logging.info('files: ' + processlabel + ".c " + processlabel + ".h")
    # 1st generate the default include statements in the dolcfh file
    dolcfh.write("#ifndef " + processlabel.upper() + "_H\n")
    dolcfh.write("#define " + processlabel.upper() + "_H\n")
    dolcfh.write("\n")
    dolcfh.write("#include <dolc.h>\n")
    # dolcfh.write("#include <string.h>\n") # TO BE UNCOMMENTED IF THIS IS NECESSARY TO BE INCLUDED
    dolcfh.write('#include "temp_asn1.h"\n')

    print(
        "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc header file '{}' ---------\n".format(
            processlabel.lower() + ".h"))
    logging.info(
        'Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc header file ' + processlabel.upper() + ".h")


    # 2nd tranfer all include statements from tfilec to the dolcfh appart from the
    # #include "processlabel.lower().h"

    # read lines from tfilec and store them

    # flag that counts the presence of #include statements in tfilec
    flag1 = 0
    pos = None

    # temporary list for storing the lines to be created in dolcfc file
    dolctempc = []
    for num, tfcline in enumerate(tfilec):
        if "#include" in tfcline and '#include "' + processlabel.lower() + '.h"' not in tfcline:
            flag1 += 1

            # print(tfcline)
            # write the include statement in dolcfh file
            dolcfh.write(tfcline)
            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator has copied the include process declarations into dolc header file '{}' ---------\n".format(
                    processlabel.lower() + ".h"))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator has copied the include process declarations into dolc header file ' + processlabel.upper() + ".h")

        elif '#include "' + processlabel.lower() + '.h"' in tfcline:
            # write the 1st line of the DOLC processlabel.lower().c file
            dolctempc.append('#include "' + processlabel.lower() + '.h"\n')
            # dolcfc.write('#include "'+processlabel.lower()+'.h"\n')
            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc c file '{}' ---------\n".format(
                    processlabel.lower() + ".c"))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc c file ' + processlabel.upper() + ".c")

        else:
            # --------------------------------------
            # GENERATE the DOLC process label.c file
            # --------------------------------------
            # search for 'state.' string and replace with the dolc pointer
            if 'state.' in tfcline:
                su = tfcline.replace('state.', '(*p->local).')
                # copy the tfilec to dolcfc without the #include statements
                dolctempc.append(su)
                # dolcfc.write(su)
            # search for processlabel.lower()+'_startup() string and replace with the dolc init function
            elif processlabel.lower() + '_startup()' in tfcline:
                startup = tfcline.replace(processlabel.lower() + '_startup()',
                                          processlabel.lower() + '_init(DOLCProcess *p)')
                dolctempc.append(startup)
                # dolcfc.write(startup)
                print(
                    "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the startup function in dolc c file '{}' ---------\n".format(
                        processlabel.lower() + ".c"))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator has created the startup function in dolc c file ' + processlabel.upper() + ".c")

            # search for processlabel.lower()+'_PI_pilabel() string and replace with the dolc fire function
            elif processlabel.lower() + '_PI_' + pilabel + '()' in tfcline:

                pifire = tfcline.replace(processlabel.lower() + '_PI_' + pilabel + '()',
                                         processlabel.lower() + '_fire(DOLCProcess *p, int mode' + processlabel.lower() + ')')
                dolctempc.append(pifire)
                # dolcfc.write(pifire)
                print(
                    "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the PI_fire function in dolc c file '{}' ---------\n".format(
                        processlabel.lower() + ".c"))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator has created the startup PI_fire in dolc c file ' + processlabel.upper() + ".c")

            # try to find the line position that contains the struct definition of process's state
            elif processlabel.lower() + '_state_t' in tfcline and 'struct' in tfcline and 'static' not in tfcline:
                pos = num
                postmp = num
                dolctempc.append(tfcline)
                # dolcfc.write(tfcline)
                print(
                    "\n ------- Taste-IV 2 FPPN-DOLC Generator has identified the state stucture definition in TASTE c file '{}' at line '{}' ---------\n".format(
                        processlabel.lower() + ".c", pos))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator has identified the state stucture definition in TASTE c file ' + processlabel.upper() + ".c at line " + str(
                        pos))

            else:
                dolctempc.append(tfcline)
                # dolcfc.write(tfcline)
    if pos is None:
        print(
            '\nError! The Taste c code file ' + processlabel.lower() + '.c does not contain state structure definition\n')
        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
        logging.debug(
            'The TASTE c code file does not contain include statements File name: ' + processlabel.lower() + '.c does not contain state structure definition')
        exit(-1)
    else:
        pass


    # if flag1 == 0:
    #     print('\nError! The Taste c code file '+ processlabel.lower()+'.c does not contain include statements\n')
    #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
    #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
    #     logging.debug('The TASTE c code file does not contain include statements File name: ' + processlabel.lower()+'.c')
    #     exit(-1)
    # else:
    #     pass

    # In order to write the function prototypes taken from the tfileh file into the dolcfh file
    # Store the required lines from tfileh to a temporary temph list
    # Also store the function that corresponds to channel data in or channel data out into a temporary list termed temppm
    # temppm is used to define the channel input or output ports and the macro definitions

    temph = []
    temppm = []

    for linedata in tfileh:
        if '#' in linedata or ('extern' in linedata and '{' in linedata) or '}' in linedata:
            pass
        elif 'extern' in linedata and 'void' in linedata and processlabel.lower() + "_RI_" in linedata:
            temppm.append(linedata)
        else:
            temph.append(linedata)

    # This check is used only if the FPPN does not allow the absence of channels
    # if len(temppm) is 0:
    #     print(
    #         "\n ------- Taste h file does not include channel data in or out function declarations. The file name is: '{}' ---------\n".format(
    #             processlabel.lower() + ".h"))
    #     logging.error(
    #         'Taste h file does not include channel data in or out function declarations. The file name is: ' + processlabel.upper() + ".h")
    #     exit(-1)
    # else:
    #     pass

    # Create the port and macro definitions
    # dataport dictionary dataport = {port-direction-i:port-name-i,...}

    dataport = {}

    # port indexes for input and output
    in_i = 1
    out_i = 1

    dolcfh.write('\n//= For every input and output channel  a port and a macro-function are declared =//\n')
    for temppmdata in temppm:
        # print temppmdata
        # isolate the string from RI_ to (
        str1 = '_RI_'
        str2 = '('
        str3 = ')'

        i1 = temppmdata.find(str1)
        i2 = temppmdata.find(str2)
        i3 = temppmdata.find(str3)

        # function name
        portname = temppmdata[i1 + 4:i2]
        # function arguments
        stringarguments = temppmdata[i2:i3 + 1]
        # number of arguments
        numberofcommas = stringarguments.count(',')
        numberofconsts = stringarguments.count('const')

        booleansearch = stringarguments.lower()
        numberofbooleans = booleansearch.count('bool')

        # find the positions of the character "," in string with function arguments
        commapositions = []
        for iii, ltr in enumerate(stringarguments):
            if ltr == ',':
                commapositions.append(iii)

        if numberofcommas is 2 and 2 <= numberofbooleans <= 3 and numberofconsts is 2:
            dataport["PORT_OUT" + str(out_i)] = '"p_PROC_' + portname + '"'

            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator is creating port definition and macros dolc h file '{}'  ---------\n".format(
                    processlabel.lower() + '.h'))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator is creating port definition and macros dolc h file ' + processlabel.upper() + ".h")

            dolcfh.write("#define PORT_OUT" + str(out_i) + ' "p_PROC_' + portname + '"\n')


            # Take the first argument after const declaration
            posconst = stringarguments.find("const")
            tempdataint = stringarguments[posconst + 5:commapositions[0]]
            tempdataintind = tempdataint.find('*')

            dataint = tempdataint[:tempdataintind - 1]
            dolcfh.write(
                "\n#define " + processlabel.lower() + '_RI_' + portname + '(arg_data_ptr, arg_valid_ptr, arg_fail_ptr) \\')
            dolcfh.write('\n')
            dolcfh.write('if (1) { \\')
            dolcfh.write('\n')
            dolcfh.write('\tDOLCData dolcdat; \\')
            dolcfh.write('\n')
            dolcfh.write('\tdolcdat.data = (arg_data_ptr); \\')
            dolcfh.write('\n')
            dolcfh.write('\tdolcdat.valid = *(arg_valid_ptr); \\')
            dolcfh.write('\n')
            dolcfh.write('\tdolcdat.size = sizeof(' + dataint + '); \\')
            dolcfh.write('\n')
            dolcfh.write('\t*(arg_fail_ptr) = DOLC_write((void*) PORT_OUT' + str(out_i) + ', &(dolcdat), 1, p); \\')
            dolcfh.write('\n')
            dolcfh.write('}\n')
            out_i += 1

            # print dataint
            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator has created a port definition and the relevant macro in dolc h file '{}'  ---------\n".format(
                    processlabel.lower() + '.h'))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator has created a port definition and the relevant macro in dolc h file ' + processlabel.upper() + ".h")

        elif numberofcommas is 1 and 1 <= numberofbooleans <= 2:
            dataport["PORT_IN" + str(in_i)] = '"p_PROC_' + portname + '"'

            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator is creating port definition and macros dolc h file '{}'  ---------\n".format(
                    processlabel.lower() + '.h'))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator is creating port definition and macros dolc h file ' + processlabel.upper() + ".h")

            dolcfh.write("#define PORT_IN" + str(in_i) + ' "p_PROC_' + portname + '"\n')


            # Take the first argument after const declaration
            posconst = stringarguments.find("(")
            tempdatain = stringarguments[posconst + 1:commapositions[0]]
            tempdatainind = tempdatain.find('*')

            datainte = tempdatain[:tempdatainind - 1]
            dolcfh.write("\n#define " + processlabel.lower() + '_RI_' + portname + '(arg_data_ptr, arg_valid_ptr) \\')
            dolcfh.write('\n')
            dolcfh.write('if (1) { \\')
            dolcfh.write('\n')
            dolcfh.write('\tDOLCData dolcdat; \\')
            dolcfh.write('\n')
            dolcfh.write('\tdolcdat.data = (arg_data_ptr); \\')
            dolcfh.write('\n')
            dolcfh.write('\tdolcdat.size = sizeof(' + datainte + '); \\')
            dolcfh.write('\n')
            dolcfh.write('\tDOLC_read((void*) PORT_IN' + str(in_i) + ', &(dolcdat), 1, p); \\')
            dolcfh.write('\n')
            dolcfh.write('\t*(arg_valid_ptr) = dolcdat.valid; \\')
            dolcfh.write('\n')
            dolcfh.write('}\n')
            in_i += 1

            # print dataint
            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator has created a port definition and the relevant macro in dolc h file '{}'  ---------\n".format(
                    processlabel.lower() + '.h'))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator has created a port definition and the relevant macro in dolc h file ' + processlabel.upper() + ".h")
        else:
            print(
                "\n ------- Taste h file does not include the appropriate arguments in channel function declaration {}. The file name is: '{}' ---------\n".format(
                    processlabel.lower() + '_RI_' + portname, processlabel.lower() + ".h"))
            logging.error(
                'Taste h file does not include the appropriate arguments in channel function declaration: ' + processlabel.lower() + '_RI_' + portname)
            logging.error('The file name is: ' + processlabel.upper() + ".h")
            exit(-1)

    # temporary list that will store the lines with the struct definition of process's state
    # this tmp list is used for transforming the struct defintion to dolc equivalent struct
    # and write it to the process processlabel.lower().h file
    tmp = []
    # the struct ends with };
    # we call the method seek(0) which repositions the file object pointer at the starting point (line)
    tfilec.seek(0)
    # read lines of tfilec
    tflines = tfilec.readlines()
    while "};" not in tflines[pos]:
        tmp.append(tflines[pos])
        pos += 1
    # append the }; to the tmp list
    tmp.append(tflines[pos])
    # print(tmp)
    # Write the transformed structure definition to the dolc.h file
    for i in tmp:
        if processlabel.lower() + '_state_t' in i and 'struct' in i and 'static' not in i:
            dolcfh.write("\n//=== structure definition of the process state\n")
            dolcfh.write("typedef struct _dolc_local_states\n")
            dolcfh.write("{\n")
        elif "};" in i:
            dolcfh.write("} " + processlabel.lower() + '_State;\n\n')
        else:
            dolcfh.write(i)

    # remove state structure definition and state variable definition from tmp and create the dolcfc file

    ii = 0
    while ii < len(tmp):
        if tmp[ii] == dolctempc[postmp - 1]:
            dolctempc.pop(postmp - 1)
        else:
            pass
        ii += 1

    for data in dolctempc:
        if 'static' in data and 'struct' in data and processlabel.lower() + '_state_t' in str(data):
            dolcfc.write('')
        else:
            dolcfc.write(data)



    # create the terminate function to the dolc.c file
    dolcfc.write("\n// terminate function\n")
    dolcfc.write("void " + processlabel.lower() + "_terminate(DOLCProcess *p)\n")
    dolcfc.write("{\n")
    dolcfc.write("\tfflush(stdout);\n")
    dolcfc.write("}\n")
    print("\n ------- Taste-IV 2 FPPN-DOLC Generator has created the dolc c file '{}' ---------\n".format(
        processlabel.lower() + ".c"))
    logging.info('Taste-IV 2 FPPN-DOLC Generator has created the dolc c file ' + processlabel.upper() + ".c")




    # write the function definition for the termination procedure in dolcfh
    dolcfh.write("\n//Function Declarations\n")
    for temphdata in temph:
        if "void" in temphdata and processlabel.lower() + "_startup();" in temphdata:
            dolcfh.write("void " + processlabel.lower() + "_init(DOLCProcess *);\n")
        elif "void" in temphdata and processlabel.lower() + "_PI_" + pilabel + '()' in temphdata:
            dolcfh.write("void " + processlabel.lower() + "_fire(DOLCProcess *, int);\n")
        else:
            pass

    # write the function definition for the termination procedure in dolcfh
    dolcfh.write("void " + processlabel.lower() + "_terminate(DOLCProcess *);\n")
    dolcfh.write("\n")
    print(
        "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the function declarations in the dolc h file '{}' ---------\n".format(
            processlabel.lower() + ".h"))
    logging.info(
        'Taste-IV 2 FPPN-DOLC Generator has created the function declarations in dolc h file ' + processlabel.upper() + ".h")
    # Write the #endif in dolcfh file
    dolcfh.write("#endif")


    # close dolcfc dolcfh and tfileh files
    dolcfc.close()
    dolcfh.close()
    tfileh.close()
    print("\n ------- Taste-IV 2 FPPN-DOLC Generator has created the dolc h file '{}' ---------\n".format(
        processlabel.lower() + ".h"))
    logging.info('Taste-IV 2 FPPN-DOLC Generator has created the dolc h file ' + processlabel.upper() + ".h")

    return


##############################################################################
# the tempasnone function is called to generate the DOLC temp_asn1.h file    #
# input: (1) dolctempasn: it creates the base file for the temp_asn1.h       #
##############################################################################

def tempasnone(dolctempasn):
    # ------------------
    # Create the file temp_asn1.h
    # ------------------


    dolctempasn.write('#ifndef TEMP_ASN_H\n')
    dolctempasn.write('#define TEMP_ASN_H\n')
    dolctempasn.write('\n')
    dolctempasn.write(r'//data type definitions')
    dolctempasn.write('\n')
    dolctempasn.write('typedef int asn1SccMyInteger;\n')
    dolctempasn.write('typedef int asn1SccT_Boolean;\n')
    dolctempasn.write('\n')
    dolctempasn.write('#endif\n')
    print(
        "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the dolc c file '{}' ---------\n".format("temp_asn1.h"))
    logging.info('Taste-IV 2 FPPN-DOLC Generator has created the dolc header file "temp_asn1.h"')
    dolctempasn.close()

    return


############################################################################################################################
# the cprotocodegen function is called when sporadic protocol are present and it generates the associated  label.c and label.h files
# input: (1) tfilec: it is the taste protocol label.c file to be read
#        (2) tfileh: it is the taste protocol label.h file to be read
#        (3) protocolabel: it is the protocol label from Taste xml file
#        (4) dolcfc: it is the dolc file protocollabel.c to be created an writen
#        (5) dolcfh: it is the dolc file protocollabel.h to be created and writen
#        (6) pilabel: it is the label of the provided interface of the sporadic protocol. Please note that currently each
#            sporadic protocol has a single provided interface (PI)
#        (7) apeprocesslabel: it is the label of the aperiodic process to which the protocol is connected.
#        (8) rilabel: it is the label of the ri interface (RI)) of the sporadic protocol (each sporadic protocol has a single RI)
############################################################################################################################

def cprotocodegen(tfilec, tfileh, protocolabel, dolcfc, dolcfh, pilabel, apeprocesslabel, rilabel):
    print(
        "\n ------- Taste-IV 2 FPPN-DOLC Generator is processing the TASTE C files for the sporadic protocol ---------\n")
    print("\n ------- files: " + protocolabel + ".c" + protocolabel + ".h  ---------\n")
    logging.info('Taste-IV 2 FPPN-DOLC Generator is processing the TASTE C files for the sporadic protocol')
    logging.info('files: ' + protocolabel + ".c " + protocolabel + ".h")

    # 1st generate the default include statements in the dolcfh file
    dolcfh.write("#ifndef " + protocolabel.upper() + "_INCLUDED\n")
    dolcfh.write("#define " + protocolabel.upper() + "_INCLUDED\n")
    dolcfh.write("\n")

    print(
        "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc header file '{}' ---------\n".format(
            protocolabel.lower() + ".h"))
    logging.info(
        'Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc header file ' + protocolabel.upper() + ".h")

    # 2nd tranfer all include statements from tfilec to the dolcfh appart from the
    # #include "protocolabel.lower().h"

    # read lines from tfilec and store them

    # flag that counts the presence of #include statements in tfilec
    flag1 = 0
    pos = None

    # temporary list for storing the lines to be created in dolcfc file
    dolctempc = []
    for num, tfcline in enumerate(tfilec):
        if "#include" in tfcline and '#include "' + protocolabel.lower() + '.h"' not in tfcline:
            flag1 += 1

            # write the include statement in dolcfh file
            dolcfh.write(tfcline)
            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator has copied the include process declarations into dolc header file '{}' ---------\n".format(
                    protocolabel.lower() + ".h"))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator has copied the include process declarations into dolc header file ' + protocolabel.upper() + ".h")

        elif '#include "' + protocolabel.lower() + '.h"' in tfcline:
            # write the 1st line of the DOLC protocolabel.lower().c file
            dolctempc.append('#include "' + protocolabel.lower() + '.h"\n')

            print(
                "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc c file '{}' ---------\n".format(
                    protocolabel.lower() + ".c"))
            logging.info(
                'Taste-IV 2 FPPN-DOLC Generator has created the include declarations in dolc c file ' + protocolabel.upper() + ".c")

        else:
            # --------------------------------------
            # GENERATE the DOLC protocol label.c file
            # --------------------------------------
            # search for 'state.' string and replace with the dolc pointer
            if 'state.' in tfcline:
                su = tfcline.replace('state.', '(*local).')
                # copy the tfilec to dolcfc without the #include statements
                dolctempc.append(su)
                # dolcfc.write(su)
            # search for processlabel.lower()+'_startup() string and replace with the dolc init function
            elif protocolabel.lower() + '_startup()' in tfcline:
                startup = tfcline.replace(protocolabel.lower() + '_startup()',
                                          'C' + apeprocesslabel.lower() + '_init(void* local_state)')
                dolctempc.append(startup)

                # dolcfc.write(startup)
                print(
                    "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the startup function in dolc c file '{}' ---------\n".format(
                        protocolabel.lower() + ".c"))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator has created the startup function in dolc c file ' + protocolabel.upper() + ".c")

            # search for sporadic label.lower()+'_PI_pilabel() string and replace with the dolc fire function
            elif protocolabel.lower() + '_PI_' + pilabel in tfcline:

                dolctempc.append(
                    'int C' + apeprocesslabel.lower() + '_prefire(void* local_state, int interval, int m_max, int prefire_period) {\n')

                print(
                    "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the sporadic prefire function in dolc c file '{}' ---------\n".format(
                        protocolabel.lower() + ".c"))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator has created the sporadic prefire in dolc c file ' + protocolabel.upper() + ".c")

            # try to find the line position that contains the struct definition of process's state
            elif protocolabel.lower() + '_state_t' in tfcline and 'struct' in tfcline and 'static' not in tfcline:
                pos = num
                postmp = num
                dolctempc.append(tfcline)
                # dolcfc.write(tfcline)
                print(
                    "\n ------- Taste-IV 2 FPPN-DOLC Generator has identified the state stucture definition in TASTE c file '{}' at line '{}' ---------\n".format(
                        protocolabel.lower() + ".c", pos))
                logging.info(
                    'Taste-IV 2 FPPN-DOLC Generator has identified the state stucture definition in TASTE c file ' + protocolabel.upper() + ".c at line " + str(
                        pos))

            else:
                dolctempc.append(tfcline)
                # dolcfc.write(tfcline)
    if pos is None:
        print(
            '\nError! The Taste c code file ' + protocolabel.lower() + '.c does not contain state structure definition\n')
        print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
        logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
        logging.debug(
            'The TASTE c code file does not contain include statements File name: ' + protocolabel.lower() + '.c does not contain state structure definition')
        exit(-1)
    else:
        pass


    # if flag1 == 0:
    #     print('\nError! The Taste c code file '+ processlabel.lower()+'.c does not contain include statements\n')
    #     print("############# Taste-IV 2 FPPN-DOLC Generator terminates due to an error! ############\n")
    #     logging.error('Taste-IV 2 FPPN-DOLC Generator terminates due to an error!')
    #     logging.debug('The TASTE c code file does not contain include statements File name: ' + processlabel.lower()+'.c')
    #     exit(-1)
    # else:
    #     pass


    dolcfh.write(
        '\n/////==== Declaration of both counter and function that are used to call the sporadic process =====\n')
    dolcfh.write('\nstatic int C' + apeprocesslabel.lower() + '_call_counter = 0;\n')

    # We need to declare the function prototype for the single RI interface of the sporadic protocol.
    # Store the lines to temporary list called temppm

    temppm = []

    for linedata in tfileh:
        if '#' in linedata or ('extern' in linedata and '{' in linedata) or '}' in linedata:
            pass
        elif 'void' in linedata and protocolabel.lower() + "_RI_" + rilabel in linedata:
            temppm.append(linedata)
        else:
            pass

    if len(temppm) is 0 or len(temppm) > 1:
        print(
            "\n ------- Sporadic Taste h file does not include connection to sporadic process or it has more than one connections. The file name is: '{}' ---------\n".format(
                protocolabel.lower() + ".h"))
        logging.error(
            'Sporadic Taste h file does not include connection to sporadic process or it has more than one connections. The file name is: ' + protocolabel.upper() + ".h")
        exit(-1)
    else:
        pass

    # find if ; character is included in temppm and remove it and do it also for the character 'extern'
    rifunction = temppm[0]
    locsemicolon = rifunction.find(';')
    locextern = rifunction.find('extern')

    dolcfh.write('\n')
    # number 7 for removing extern plus the bar
    dolcfh.write(rifunction[locextern + 7:locsemicolon])
    dolcfh.write('\n{\n')
    dolcfh.write('  C' + apeprocesslabel.lower() + '_call_counter++;\n')
    dolcfh.write('}\n')
    dolcfh.write('\n')

    print(
        "\n ------- The function prototype for the RI interface of the sporadic protocol and its body has been declared in dolc h file. The file name is: '{}' ---------\n".format(
            protocolabel.lower() + ".h"))
    logging.info(
        'The function prototype for the RI interface of the sporadic protocol and its body has been declared in dolc h file. The file name is: ' + protocolabel.upper() + ".h")

    #
    # temporary list that will store the lines with the struct definition of process's state
    # this tmp list is used for transforming the struct definition to dolc equivalent struct
    # and write it to the process processlabel.lower().h file
    tmp = []
    # the struct ends with };
    # we call the method seek(0) which repositions the file object pointer at the starting point (line)
    tfilec.seek(0)
    # read lines of tfilec
    tflines = tfilec.readlines()
    while "};" not in tflines[pos]:
        tmp.append(tflines[pos])
        pos += 1
    # append the }; to the tmp list
    tmp.append(tflines[pos])

    # Write the transformed structure definition to the dolc.h file
    for i in tmp:
        if protocolabel.lower() + '_state_t' in i and 'struct' in i and 'static' not in i:
            dolcfh.write("\n//=== structure definition of the sporadic protocol state\n")
            dolcfh.write("typedef struct \n")
            dolcfh.write("{\n")
        elif "};" in i:
            dolcfh.write("} C" + apeprocesslabel.lower() + '_State;\n\n')
        else:
            dolcfh.write(i)
    # Write the #endif in dolcfh file
    dolcfh.write("#endif")

    # remove state structure definition and state variable definition from tmp and create the dolcfc file

    ii = 0
    while ii < len(tmp):
        if tmp[ii] == dolctempc[postmp - 1]:
            dolctempc.pop(postmp - 1)
        else:
            pass
        ii += 1


    # find the 'C'+apeprocesslabel.lower()+'_init(void* local_state)' in dolctempc
    posinit = [i for i, s in enumerate(dolctempc) if '_init(void* local_state)' in s]

    # check if '{' is located after the function
    if '{' in dolctempc[posinit[0] + 1]:

        # add the local state pointer within the body of the 'C'+apeprocesslabel.lower()+'_init(void* local_state)' function
        dolctempc.insert(int(posinit[0]) + 2,
                         '    C' + apeprocesslabel.lower() + '_State* local = (C' + apeprocesslabel.lower() + '_State*) local_state;\n')
    elif '{' in dolctempc[posinit[0]]:
        # add the local state pointer within the body of the 'C'+apeprocesslabel.lower()+'_init(void* local_state)' function
        dolctempc.insert(int(posinit[0]) + 1,
                         '    C' + apeprocesslabel.lower() + '_State* local = (C' + apeprocesslabel.lower() + '_State*) local_state;\n')
    else:
        pass

    # find int C'+apeprocesslabel.lower()+'_prefire(void* local_state, int interval, int m_max, int prefire_period) {\n
    posprefire = [i for i, s in enumerate(dolctempc) if
                  '_prefire(void* local_state, int interval, int m_max, int prefire_period)' in s]
    if '{' in dolctempc[posprefire[0] + 1]:
        dolctempc.pop(posprefire[0] + 1)
        # add the local state pointer within the body of the  function
        dolctempc.insert(int(posprefire[0]) + 1, '\n   /////===== standard beginning =============\n'
                                                 '    C' + apeprocesslabel.lower() + '_State* local = (C' + apeprocesslabel.lower() + '_State*) local_state;\n'
                                                                                                                                      '    C' + apeprocesslabel.lower() + '_call_counter = 0;\n\n')
    else:
        pass

    # find last } in dolctempc
    lastanc = [i for i, s in enumerate(dolctempc) if '}' in s]

    dolctempc.insert((int(lastanc[len(lastanc) - 1])), '\n////=============== Standard ending ==========\n'
                                                       '    return C' + apeprocesslabel.lower() + '_call_counter;\n')

    for data in dolctempc:
        if 'static' in data and 'struct' in data and protocolabel.lower() + '_state_t' in str(data):
            dolcfc.write('')
        else:
            dolcfc.write(data)

    # close dolcfc dolcfh and tfileh files
    dolcfc.close()
    dolcfh.close()
    tfileh.close()

    print(
        "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the sporadic protocol dolc c file '{}' ---------\n".format(
            protocolabel.lower() + ".c"))
    logging.info(
        'Taste-IV 2 FPPN-DOLC Generator has created the sporadic protocol dolc c file ' + protocolabel.lower() + ".c")

    print(
        "\n ------- Taste-IV 2 FPPN-DOLC Generator has created the sporadic protocoldolc h file '{}' ---------\n".format(
            protocolabel.lower() + ".h"))
    logging.info(
        'Taste-IV 2 FPPN-DOLC Generator has created the sporadic protocol dolc h file ' + protocolabel.upper() + ".h")

    return

################################################################################################################################
# the LCP function calculates the least common multiple of the periods of the function objects
# it is defined for periodic tasks at this momment
# input: (1) lstprd -> list with periods from periodic tasks.
# output:(1) lstcmltpl -> the least common multiple
#################################################################################################################################

def LCP(lstprd):
# DEBUG PURPOSES
#    for i in lstprd:
#        print("\n PERIODS = {} \n".format(i))


    values = set([abs(int(v)) for v in lstprd])
# DEBUG PURPOSES
#    print("\n values = {} \n".format(values))
    lstcmltpl = n0 = max(values)
# DEBUG PURPOSES
#    print ("lstcmltpl = {}\n".format(lstcmltpl))
    values.remove(lstcmltpl)
    while any( lstcmltpl % m for m in values ):
        lstcmltpl += n0
# DEBUG PURPOSES
#    print ("LCP = {}\n".format(lstcmltpl))
    return lstcmltpl


