#!/usr/bin/python
########################################################################################################
# Title: taste2dolc Generator
# Author: Fotios Gioulekas
# Date: 2016/01/16
# Project:MoSaTT-CMP
# Name: taste2dolc
# Version: 10.0
# Abstract:
# This Python generator that takes as input the exported by TASTE-IV XML file = InterfaceView.xml
# and the corresponding C code files, and generates an XML file according to Fixed Priority
# Process Network (FPPN) based on Distributed Operation Layer (DOL) framework
# The task graph arcs and jobs characteristics are also extracted and stored in two files to be used
# by the time-triggered scheduler
###############################################################################################

########################################################################################################
# USAGE:
# Please place the files taste2dolc.py, fppnxml.py and genmodules.py at the taste directory where
# the folders with the generated c code of the corresponding function folders exist and the InterfaceView.xml
# has been exported
#########################################################################################################

#########################################################################################################


# import the C based implementation of Python ElementTree API which manipulates XML files.
# the C implementation is faster and consumes significantly less memory than pure Python ElementTree API.
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET
# import the GenModules module
import genmodules
# Import the fppnxml module
import fppnxml
# Import minidom from xml.dom module
from xml.dom import minidom

# import the logging module for logging purposes
import logging

# find current directory path in debian os where assert is running
import os

path = os.getcwd()


# the filename of the exported xml file from Taste
file = '/InterfaceView.xml'
ivxml = path + file

# string with the file location for the generation of FPPN-DOLC xml file
pathforfppn = path + '/fppn'
# create the directory where the generated dolc based fppn model will be placed
# if the directory exists do nothing
if not os.path.exists(pathforfppn):
    os.makedirs(pathforfppn, 0o777)
else:
    pass
# string with the log file location and name
logging_file = pathforfppn + '/generator.out'
# strings with the location of the task graph files and their names
tgjobs = pathforfppn + '/fppn_tg.jobsprocs'
tg = pathforfppn + '/fppn_tg.jobs'
# string with the file location for the generation of FPPN-DOLC xml file
pathforfppnsrc = pathforfppn + '/src'

# string with the file that containts offsets and process labels
# the file is organized as follows
# each line includes label of process=offset value

offsetfile = path+'/offsets.dat'
# create the directory where the generated dolc based fppn model will be placed
# if the directory exists do nothing
if not os.path.exists(pathforfppnsrc):
    os.makedirs(pathforfppnsrc, 0o777)
else:
    pass

# string for the fppn based dolc xml file
pathforxml = pathforfppn + '/fppn-dolc.xml'

# Create the XML file for the fppndolc.xml
# print(pathforxml)
fxml = open(pathforxml, 'w')

# enter the name of the FPPN-XML file
# appname = raw_input("\n Please enter the name of the FPPN DOLC application and press enter:\n")
appname = 'TASTE2DOLC_XML'
# Create the element root of the FPPNDOLC.xml file tagged as app
app = ET.Element("app", name=appname)

# in the filename put the logging_file

logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s : %(levelname)s : %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename=logging_file,
                    filemode='w')

print("\n##########################################\n")
print("##### TASTE-IV 2 FPPN-DOLC GENERATOR #####")
print("\n##########################################\n")
logging.info("##########################################")
logging.info("WELCOME TO THE TASTE-IV 2 FPPN-DOLC GENERATOR")
logging.info("##########################################")
print("##### TASTE-IV XML file is being imported #####")
logging.info("TASTE-IV XML file is being imported")

try:
    os.path.isfile(ivxml)
    f = open(ivxml, 'r')
except IOError:
    print('\nERROR while importing the TASTE-IV XML file = InterfaceView.xml! Please check the file or the path to it!')
    logging.error(
        'ERROR while importing the TASTE-IV XML file = InterfaceView.xml! Please check the file or the path to it!')
    print("\n######################################################################\n")
    print("##### Taste-IV 2 FPPN-DOLC Generator terminates due to an error! #####")
    logging.debug('Taste-IV 2 FPPN-DOLC Generator terminates due to an error! '
                  'Check the TASTE-IV XML file location or name')
    print("\n######################################################################\n")
    exit(-1)
else:
    # Uncomment the next 3 lines of code when the generator runs on Linux OS
    # index=filename.rfind("InterfaceView.xml")
    # print(index)
    # print('\nThe XML File {} has been successfully imported'.format(filename[index+1:]))

    # Comment the next line when the Generator runs on Linux OS
    print('\nThe XML File {} has been successfully imported'.format(f.name))
    logging.info('The XML File {} has been successfully imported'.format(f.name))
# load the XML file in a tree object
tree = ET.ElementTree(file=f)

##########################
# Start parsing the file #
print("\n##### Taste-IV 2 FPPN-DOLC Generator is parsing the TASTE-IV XML file #####\n")
logging.info('Taste-IV 2 FPPN-DOLC Generator is parsing the TASTE-IV XML file')
##########################


# Get the XML root
xmlroot = tree.getroot()
# print the root element (Debugging purposes)
print('XML Root=%s\n' % xmlroot.tag)
logging.info('The XML root of the InterfaceView.xml file is {}'.format(xmlroot.tag))
# Call function that parses the xmlroot object and finds the children with tag FunctionObject
numOfFunctionObjects, functionlist = genmodules.functionObject(xmlroot)

# print the number of functionObjects (Debugging purposes)
print ("Number of elements tagged Taste I-V FunctionObjects = %d\n" % numOfFunctionObjects)
logging.info(
    'The Total Number of elements Taste I-V FunctionObjects identified in TASTE I-V xml file is {}\n'.format(numOfFunctionObjects))

for item in functionlist:
    print('The function object with label "{}" within InterfaceView.xml and attributes: {} has been identified\n'.format(
            item['label'], item))
    logging.info(
        'The function object with label "{}" within InterfaceView.xml and attributes: {} has been identified'.format(
            item['label'], item))

childrenlist = genmodules.functionchildren(xmlroot, functionlist)

for functobject in functionlist:
    print('\n --- The function object "{}" has the following children: {}\n'.format(functobject['label'],
                                                                                    childrenlist[functobject['label']]))
    logging.info('The function object "{}" has the following children:'.format(functobject['label']))
    logging.info(childrenlist[functobject['label']])

app, pervals, chains = fppnxml.createDolcSubElements(app, functionlist, ET, childrenlist, xmlroot, path, pathforfppnsrc)


XMLDOC = ET.ElementTree(app)

# XMLDOC.write(f, encoding='utf-8', xml_declaration=True)
print('\n')
# print(fppnxml.prettify(app, ET, minidom))

appxml = fppnxml.prettify(app, ET, minidom)
fxml.write(appxml)
fxml.close()
logging.info('The Generated FPPN-DOLC-XML file is located at:{}'.format(pathforxml))
print("\nThe Generated FPPN-DOLC-XML file is located at {}".format(pathforxml))

logging.info('The Generated FPPN-DOLC C code files are located at:{}'.format(pathforfppnsrc))
print("\nThe Generated FPPN-DOLC C code files are located at {}".format(pathforfppnsrc))

logging.info('Taste-IV 2 FPPN-DOLC Generator has successfully built the FPPN files!')
print('\n--------------------------------------------------------------')
print("\nTaste-IV 2 FPPN-DOLC Generator has successfully built the FPPN files!")
print('\n--------------------------------------------------------------')

logging.info('-------------------------------------------------------------------------------------------')
logging.info('###########################################################################################')
logging.info('##### THE TASTE-IV 2 FPPN-DOLC GENERATOR IS PROCEEDING WITH THE TASK GRAPH EXTRACTION #####')
logging.info('###########################################################################################')
logging.info('-------------------------------------------------------------------------------------------')

print('\n---------------------------------------------------------------------------------------------')
print('\n#############################################################################################')
print("\n#####  THE TASTE-IV 2 FPPN-DOLC GENERATOR IS PROCEEDING WITH THE TASK GRAPH EXTRACTION! #####")
print('\n#############################################################################################')
print('\n---------------------------------------------------------------------------------------------')

logging.info('-------------------------------------------------------------------------------------------')
logging.info("The unrolled schedule within the time frame of the LCP of the tasks' periods is calculated")
logging.info('-------------------------------------------------------------------------------------------')
print('\n-------------------------------------------------------------------------------------------')
print("The unrolled schedule within the time frame of the LCP of the tasks' periods is calculated")
print('-------------------------------------------------------------------------------------------\n')
#retrieve all periods from the periodic processes dictionary pervals
#charkeys = ["period","deadline", "WCET", "fpriority", "criticality"]
#dictionary with criticality values
criticality = {'A':5,'B':4, 'C':3, 'D':2, 'E':1}
# list with period values
lstprd = []
# list with criticality values
criticalitylist = []
for keyid in pervals:
    lstprd.append(pervals[keyid]['period'])
    criticalitylist.append(criticality[pervals[keyid]['criticality']])

#debuging purposes
#print(criticalitylist)
# Calculate the Least common multiple Hlcp of the tasks' periods - Hyperperiod
Hyper = genmodules.LCP(lstprd)
print("\n/******************************************/")
print("The hyperperiod of the tasks' periods = {}".format(Hyper))
print("/******************************************/\n")

logging.info("/--*--/ The hyperperiodof the tasks' periods = {} /--*--/".format(Hyper))

# DEBUG purposes
#print("\npervals = {} \n".format(pervals))
logging.info("/******************************************/")
logging.info("/*/ The tasks from where the task graph is extracted are = {} /*/".format(pervals))
logging.info("/******************************************/")
# Calculate the number of columns for WCET repetitions based on the criticality
maxWCETcolumns = max(2, max(criticalitylist))

print("\n/***************************************************************************************/")
print("The max number of columns for WCET repetitions in fppn_tg.jobs file is '{}'".format(maxWCETcolumns))
print("/***************************************************************************************/\n")
logging.info("/**************************************************************************************/")
logging.info(" The max number of columns for WCET repetitions in fppn_tg.jobs file is '{}'".format(maxWCETcolumns))
logging.info("/*************************************************************************************/")
#define keys for jobs data
if maxWCETcolumns is 2:
    jobskeys = ['id','interArrivalTime', 'absoluteDeadline', 'criticality', 'WCETcolumn1','WCETcolumn2']
elif maxWCETcolumns is 3:
    jobskeys = ['id','interArrivalTime', 'absoluteDeadline', 'criticality', 'WCETcolumn1', 'WCETcolumn2', 'WCETcolumn3']
elif maxWCETcolumns is 4:
    jobskeys = ['id','interArrivalTime', 'absoluteDeadline', 'criticality', 'WCETcolumn1', 'WCETcolumn2', 'WCETcolumn3',
                'WCETcolumn4' ]
else:
    jobskeys = ['id','interArrivalTime', 'absoluteDeadline', 'criticality', 'WCETcolumn1', 'WCETcolumn2', 'WCETcolumn3',
                'WCETcolumn4','WCETcolumn5' ]

# DEBUG purposes
#print('JOBKEYS = {}\n'.format(jobskeys))

# open the files for the task graph definition
# file with job and priorities within the hyperperiod
ftg = open(tg, 'w')
# file that provides the number of jobs executed by each process
ftgjobs = open(tgjobs, 'w')
print('---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
print("\nTaste-IV 2 FPPN-DOLC Generator has started working on the fppn_tg.jobs file!\n")
print("\nThe fppn_tg.jobs file provides the job ids, their inter-arrival times, their absolute deadlines, their criticality levels, their WCET per criticality level>2 and the Task Graph Arcs in a format of ordered pair of jobs e.g. (Ji,Jk)\n")
print('----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n')
logging.info('-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
logging.info("Taste-IV 2 FPPN-DOLC Generator has started working on the fppn_tg.jobs file!")
logging.info("The fppn_tg.jobs file provides the job ids, their inter-arrival times, their absolute deadlines, their criticality levels, their WCET per criticality level>2 and the Task Graph Arcs in a format of ordered pair of jobs e.g. (Ji,Jk)")
logging.info('--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')

#initialize job id
job_id = 1
#initialize the dictionary that keeps record of all jobs data
#e.g. {'1':{'interArrivalTime':#, 'absoluteDeadline':#, 'criticality':#, 'WCETcolumn1':#, 'WCETcolumn2':#}, '2':{...}, ...}
jobdict = {}
dictwithjobsperprocess ={}
dictwithnumofjobsperprocess ={}
# jobfp = {1:fp#, 2:fp#, ...} jobnumber:functional priority number
jobfp = {}
#list with iat values
iatlist = []

# open file with offsets
print("##### TASTE2DOLC generator checks if offsets.data exists. If exists it imports it. #####")
logging.info("TASTE2DOLC generator checks if offsets.data exists. If exists it imports it.")
# initialize the dictionary of offsets with {label:offset value, ...}
offsets = {}
if os.path.exists('offsets.dat'):
    foffset = open(offsetfile, 'r')
    # Comment the next line when the Generator runs on Linux OS
    print('\nThe offsets.dat has been successfully imported')
    logging.info('The offsets.dat has been successfully imported')
    for numofline, foffsetline in enumerate(foffset):
        eqindex = foffsetline.index('=')
        returnindex = foffsetline.index('\n')
        offsets[foffsetline[:eqindex]]=float(foffsetline[eqindex+1:returnindex])
else:
    offsets['noprocess']=0

# DEBUGING purposes
#print('OFFSETS = {}'.format(offsets))
#print("PERVALS = {}".format(pervals))
for process in pervals:
    #print("Process = {}".format(process))
    #time index (i.e. 0, 1, 2, till di<=Hyper and ndex*period' < Hyper
    index = 0
    #list tha contain the dictionaries of each job data for each process
    jobslist = []
    wcetvalues = []
    for iii  in range(1, maxWCETcolumns+1):
        wcetvalues.append(pervals[process]['WCET'])
    #DEBUG purposes
    #print('WCET values {}\n'.format(wcetvalues))
    di = 0
    if process in offsets:
        offset = offsets[process]
    else:
        offset = 0
    #DEBUGING purposes
    #print("OFFSET = {}\n".format(offset))
    while(di <= Hyper) and index*pervals[process]['period']+offset < Hyper:
        #jobs data list
        jobdata = []
        iat=index*pervals[process]['period']+int(offset)
        di=float(min(Hyper,iat+pervals[process]['deadline']))
        if iat not in iatlist:
            iatlist.append(iat)
        jobdata = [job_id,iat, di, criticality[pervals[process]['criticality']]]
        jobdata.extend(wcetvalues)
        jobdict [job_id] = dict(zip(jobskeys[1:], jobdata[1:]))
        print('------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
        print("Taste-IV 2 FPPN-DOLC Generator has generated the job = J{}, functional priority = {} and data = {}"
              . format(job_id, pervals[process]['fpriority'],dict(zip(jobskeys[1:], jobdata[1:]))))
        print('------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
        logging.info('-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
        logging.info("Taste-IV 2 FPPN-DOLC Generator has generated the job = J{}, functional priority = {} and data = {}".
                     format(job_id,pervals[process]['fpriority'], dict(zip(jobskeys[1:], jobdata[1:]))))
        logging.info('-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
        jobfp[job_id] = int(pervals[process]['fpriority'])
        jobslist.append(dict(zip(jobskeys, jobdata)))
        for eachvalue in jobdata[1:]:
            ftg.write(str(int(eachvalue)))
            ftg.write('\t')
        ftg.write('\n')
        index += 1
        #check if the process type is sporadic i.e. its type is imperiodic
        if pervals[process]['ptype'] == 'imperiodic':
            #if it is of type sporadic create extra 'm' jobs for each iat value
            #one is already placed, therefore add 'm-1' more jobs
            #DEBUG purposes
            #print(range(1,pervals[process]['m']-1))
            for mvar in range(1,int(pervals[process]['m'])):
                job_id += 1
                jobdatanew = []
                jobdatanew = [job_id,iat, di, criticality[pervals[process]['criticality']]]
                jobdatanew.extend(wcetvalues)
                jobdict [job_id] = dict(zip(jobskeys[1:], jobdatanew[1:]))
                jobfp[job_id] = int(pervals[process]['fpriority'])
                jobslist.append(dict(zip(jobskeys, jobdatanew)))
                print('-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
                print("Taste-IV 2 FPPN-DOLC Generator has generated the job = J{}, functional priority = {} and data = {}"
                    . format(job_id, pervals[process]['fpriority'],dict(zip(jobskeys[1:], jobdatanew[1:]))))
                print('-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
                logging.info('----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
                logging.info("Taste-IV 2 FPPN-DOLC Generator has generated the job = J{}, functional priority = {} and data = {}".
                             format(job_id,pervals[process]['fpriority'], dict(zip(jobskeys[1:], jobdatanew[1:]))))
                logging.info('----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
                for eachvalue in jobdatanew[1:]:
                    ftg.write(str(int(eachvalue)))
                    ftg.write('\t')
                ftg.write('\n')
            job_id += 1
        else:
            job_id += 1
    dictwithjobsperprocess[process]=jobslist
    dictwithnumofjobsperprocess[process]=len(jobslist)
    #print("JOBS LIST {} FOR PROCESS = {}\n".format(jobslist, process))

logging.info('----------------------------------------------------------------------------------------------------------------------------------')
logging.info("JOB features for all processes = {}".format(dictwithjobsperprocess))
logging.info('----------------------------------------------------------------------------------------------------------------------------------')
#DEBUG purposes
print("JOB features for all processes = {}\n".format(dictwithjobsperprocess))

print('----------------------------------------------------------------------------------------------------------------------')
print("\nTaste-IV 2 FPPN-DOLC Generator is proceeding with the task graph arcs descriptions used in Time-Triggered Scheduling!\n")
print('----------------------------------------------------------------------------------------------------------------------\n')
logging.info('---------------------------------------------------------------------------------------------------------------------------------------------------')
logging.info("Taste-IV 2 FPPN-DOLC Generator Taste-IV 2 FPPN-DOLC Generator is proceeding with the task graph arcs descriptions used in Time-Triggered Scheduling!")
logging.info('----------------------------------------------------------------------------------------------------------------------------------------------------')


# Create dictionary that associated the job number with its interarrival time
# jobiat = {1:iat, 2:iat, 3: iat, ...} jobnumber:interarrival-time
jobiat = {}


for dataitem in jobdict:
    jobiat[dataitem] = jobdict[dataitem]['interArrivalTime']

sortedjobsiat = {}
# DEBUG purposes
#print('JOBIAT = {}\n'.format(jobiat))
# DEBUG purposes
#print('iatlist = {}\n'.format(iatlist))
# sort jobiat in iat ascending order and create a list of tuples [(job#, iat),...]
jobiatsortedtuples= sorted(jobiat.items(), key = lambda x:x[1])

# DEBUG purposes
#print('SORTED JOBIAT in tuples = {}\n'.format(jobiatsortedtuples))

# DEBUG purposes
#print('JOBFP = {}\n'.format(jobfp))

#temporary list to store the job names
listtemp = []

for item in jobiatsortedtuples:
    listtemp.append('J'+str(item[0]))
#DEBUG purposes
#print('LISTTEMP = {}\n'.format(listtemp))

############################################################################
# Per precedence chain order define all the job pairs for each interarrival time
# In case that the precedence chain includes a sporadic process ensure that it is
# recreated by setting the sporadic in the higher priority use the fps
# ############################################################################

# retrieve precedence chains and create a list of tuples that contain the chains
prechain = []
for  chainstring in chains:
    commapos = chainstring.index(',')
    name1 = chainstring[:commapos]
    name2 = chainstring[commapos+1:]

    if pervals[name1]['ptype'] == 'periodic' and pervals[name2]['ptype'] == 'periodic':
        tuplechain = (name1,name2)
        #DEBUG purposes
        #print("TUPLE CHAIN1 = {}\n".format(tuplechain))
        prechain.append(tuplechain)
    elif pervals[name1]['ptype'] == 'periodic' and pervals[name2]['ptype'] == 'imperiodic':
        tuplechain = (name2,name1)
        #DEBUG purposes
        #print("TUPLE CHAIN2 = {}\n".format(tuplechain))
        prechain.append(tuplechain)
    elif pervals[name1]['ptype'] == 'imperiodic' and pervals[name2]['ptype'] == 'periodic':
        tuplechain = (name1,name2)
        #DEBUG purposes
        #print("TUPLE CHAIN3 = {}\n".format(tuplechain))
        prechain.append(tuplechain)
    else:
        print('---------------------------------------------------------------------------------------------------------------------------------------')
        print("\nERROR! Taste-IV 2 FPPN-DOLC Generator identified that sporadic processes '{}' and '{}' are connected!\n".format(name1,name2))
        print('----------------------------------------------------------------------------------------------------------------------------------------\n')
        logging.info('----------------------------------------------------------------------------------------------------------------------------------')
        logging.error('Taste-IV 2 FPPN-DOLC Generator identified that 2 sporadic processes are connected!')
        logging.debug("The labels of the sporadic processes are '{}' and '{}'".format(name1,name2))
        logging.info('----------------------------------------------------------------------------------------------------------------------------------')
        exit(-1)
    #DEBUG purposes
    #print("TUPLE CHAIN = {}\n".format(tuplechain))

#DEBUG purposes
print('/----------------------------------------------------------------------------------/')
print("PRECENDECE CHAIN LIST = {}".format(prechain))
print('/----------------------------------------------------------------------------------/')
print("DICT OF PROCESSES = {}".format(dictwithjobsperprocess))
print('/----------------------------------------------------------------------------------/')
# for each ordered pair (tuple) of processes defined in prechain, create the job graph arcs per interarrival time

#initialize list0 with jobids per interarrival time and per predence chains
list0 = []

for chainitem in prechain:
    process1 = chainitem[0]
    process2 = chainitem[1]
    #DEBUG purposes
    #print("PROCESS1 = {}\n".format(process1))
    #print("PROCESS2 = {}\n".format(process2))

    #retrieve the jobs for each process
    jobsofprocess1 = dictwithjobsperprocess[process1]
    jobsofprocess2 = dictwithjobsperprocess[process2]
    #DEBUG purposes
    #print("jobsofprocess1 = {}\n".format(jobsofprocess1))
    #print("jobsofprocess2 = {}\n".format(jobsofprocess2))
    #define one dict per process that contain jobsid:iat and a list with process jobids
    process1jobsiat = {}
    process2jobsiat = {}
    process1jobids = []
    process2jobids = []
    processeswithjobids = {}
    for eachjobproc1 in jobsofprocess1:
        job1 = eachjobproc1['id']
        iat1 = eachjobproc1['interArrivalTime']
        process1jobsiat[job1] = iat1
        process1jobids.append(job1)
    for eachjobproc2 in jobsofprocess2:
        job2 = eachjobproc2['id']
        iat2 = eachjobproc2['interArrivalTime']
        process2jobsiat[job2] = iat2
        process2jobids.append(job2)

    processeswithjobids[process1] = process1jobids
    processeswithjobids[process2] = process2jobids
    #DEBUG purposes
    #print("PROCESSESWITHJOBIDS = {}".format(processeswithjobids))
    #print("PROCESS1JOBSIAT = {}".format(process1jobsiat))
    #print("PROCESS2JOBSIAT = {}".format(process2jobsiat))
    # since process2 is always periodic check if process1 is sporadic in order to deal with multiple copies of m consecutive
    # jobs that take place in the same interarrival time
    if pervals[process1]['ptype'] is 'imperiodic':
        #DEBUG purposes
        #print('imperiodic\n')
        # define a list with jobs [J#,J#, ...]
        list1 = []
        # for each interarrival time in the time ascending list of interarrival times
        for iatvalue in sorted(iatlist):
            # check if the current interarrival time belongs to a job of the first process of the chain which has the highest priority)
            if iatvalue in process1jobsiat.values():
                #if yes then append each job at this iat value to list1. Note that jobids are already in ascending form in the dictionary
                for jobid, iat in process1jobsiat.items():
                    if iat == iatvalue:
                        list1.append(jobid)
                #if the current interarrival time also belongs to a job from the second periodic process of lowest priority in the chain
                if iatvalue in process2jobsiat.values():
                    # find the jobid that has this iat value and append it to the list of jobs list1
                    for jobid, iat in process2jobsiat.items():
                        if iat == iatvalue:
                            list1.append(jobid)
            # the current interarrival time does not belong to a job of the first process. Proceed with process2 in the chain
            elif iatvalue in process2jobsiat.values():
                #if the current interarrival time belongs to a job, find it and append it to the list1
                for jobid, iat in process2jobsiat.items():
                    if iat == iatvalue:
                        list1.append(jobid)
        #DEBUG purposes
        print('list1 = {}\n'.format(list1))

    else:
        #DEBUG purposes
        #print('periodic\n')
        # define a list with jobs [J#,J#, ...]
        list1 = []
        # for each interarrival time in the time ascending list of interarrival times
        for iatvalue in sorted(iatlist):
            # check if the current interarrival time belongs to a job of the first process of the chain which has the highest priority)
            if iatvalue in process1jobsiat.values():
                #if yes then append each job at this iat value to list1. Note that jobids are already in ascending form in the dictionary
                for jobid, iat in process1jobsiat.items():
                    if iat == iatvalue:
                        list1.append(jobid)
                #if the current interarrival time also belongs to a job from the second periodic process of lowest priority in the chain
                if iatvalue in process2jobsiat.values():
                    # find the jobid that has this iat value and append it to the list of jobs list1
                    for jobid, iat in process2jobsiat.items():
                        if iat == iatvalue:
                            list1.append(jobid)
            # the current interarrival time does not belong to a job of the first process. Proceed with process2 in the chain
            elif iatvalue in process2jobsiat.values():
                #if the current interarrival time belongs to a job, find it and append it to the list1
                for jobid, iat in process2jobsiat.items():
                    if iat == iatvalue:
                        list1.append(jobid)
        #DEBUG purposes
        #print('list1 = {}\n'.format(list1))
        # append tuples with task graph arcs for the jobs sorted in list1 to list2
    if len(list1) > 1:
        ind1 = 0
        for item in list1:
            ind1 = ind1 + 1
            #if index exceeds the length of list1 terminate the appending procedure
            if ind1 < len(list1):
                tuple = ('J'+str(item),'J'+str(list1[ind1]))
                if tuple not in list0:
                    list0.append(tuple)


#DEBUG purposes
#print('list0 = {}\n'.format(list0))

#2nd pass create lists for the identification of the consecutive jobs performed by each one of the processes not included
#in the list2 list appended above

listkak=[]
consecutivejobs = []
for processlabel in dictwithjobsperprocess:
    #DEBUG purposes
    #print(dictwithjobsperprocess[processlabel])
    if pervals[processlabel]['ptype'] == 'imperiodic':
        for dataprops in dictwithjobsperprocess[processlabel]:
            listkak.append('J'+str(dataprops['id']))
            #DEBUG purposes
            #print("DATAPROPS = {}\n".format(dataprops))
            #DEBUG purposes
            #print(listkak)
        if len(listkak) > 1:
            ind = 0
            for i in listkak:
                ind=ind+1
                if ind <len(listkak):
                    #DEBUG purposes
                    ## print(i,listkak[ind])
                    tuplejobs = (i,listkak[ind])
                    consecutivejobs.append(tuplejobs)
        listkak = []

    else:
        for dataprops in dictwithjobsperprocess[processlabel]:
            listkak.append('J'+str(dataprops['id']))
            #DEBUG purposes
            #print(listkak)
        if len(listkak) > 1:
            ind = 0
            for i in listkak:
                ind=ind+1
                if ind <len(listkak):
                    #DEBUG purposes
                    ## print(i,listkak[ind])
                    tuplejobs = (i,listkak[ind])
                    consecutivejobs.append(tuplejobs)
        listkak = []
#DEBUG purposes
#print("CONSECUTIVE JOBS = {}\n".format(consecutivejobs))

# append the missed task graph arcs to list0
for job in consecutivejobs:
    #DEBUG purposes
    #print("JOB = {}\n".format(job))
    if job not in list0:
        list0.append(job)
#DEBUG purposes
# print("TASK GRAPH JOB = {}\n".format(list2))
# print('JOBDICT = {}\n'.format(jobdata))
# print('JOBDATA = {}\n'.format(jobdict))
# print('Process dictionary = {}\n'.format(dictwithjobsperprocess))
# print('DATA = {}\n'.format(dictwithnumofjobsperprocess))
print('\n###########################')
print('The TASK GRAPH ARCS are:')
logging.info('###########################')
logging.info('The TASK GRAPH ARCS are:')
for listdata in list0:
    ftg.write('('+listdata[0]+', '+listdata[1]+')\n')
    print('('+listdata[0]+', '+listdata[1]+')')
    logging.info('('+listdata[0]+', '+listdata[1]+')')

print('###########################\n')
logging.info('###########################')
# Counter for total executed jobs
totalnumber_of_jobs = 0

print('----------------------------------------------------------------------------------')
print("\nTaste-IV 2 FPPN-DOLC Generator has started working on the fppn_tg.jobsprocs file!\n")
print("The fppn_tg.jobsprocs file provides the number of jobs executed by each process\n")
print('----------------------------------------------------------------------------------\n')
logging.info('---------------------------------------------------------------------------------')
logging.info("Taste-IV 2 FPPN-DOLC Generator has started working on the fppn_tg.jobsprocs file!")
logging.info("The fppn_tg.jobsprocs file provides the number of jobs executed by each process")
logging.info('---------------------------------------------------------------------------------')

for process in pervals:
    ftgjobs.write(str(process) +'\t'+ str(dictwithnumofjobsperprocess[process]) +'\n')
    print('The process "'+str(process) +'" executes '+ str(dictwithnumofjobsperprocess[process]) +' jobs within the '
                                                                                                  'Hyperperiod of '+str(Hyper)+'ms\n')
    logging.info('The process "'+str(process) +'" executes '+ str(dictwithnumofjobsperprocess[process]) +' jobs within '
                                                                                                         'the Hyperperiod of '+str(Hyper)+'ms')
    totalnumber_of_jobs += dictwithnumofjobsperprocess[process]

print('The total number of executed jobs is "{}"\n'.format(totalnumber_of_jobs))
logging.info('The total number of executed jobs is "{}"'.format(totalnumber_of_jobs))


# close the files
ftg.close()
ftgjobs.close()

print('#####################################################################################################################################################################\n')
print("Taste-IV 2 FPPN-DOLC Generator has successfully created the 'fppn_tg.jobsprocs' and 'fppn_tg.jobs' files located at "
      "the FPPN directory ={}!\n".format(pathforfppn))
print('#####################################################################################################################################################################\n')
logging.info('#############################################################################################################################################################')
logging.info("Taste-IV 2 FPPN-DOLC Generator has has successfully created the 'fppn_tg.jobsprocs' and 'fppn_tg.jobs' "
             "files located at the FPPN directory ={}!".format(pathforfppn))
logging.info('#############################################################################################################################################################')

